/*
 * sisctrl - SiS Display Control Panel
 * for the SiS XFree86 driver
 *
 * (C) 2003-2004 Thomas Winischhofer <thomas@winischhofer.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include "sisctrl.h"

/* libc etc wrappers */
static int
get_pid(void)
{
#ifdef HAVE_GETPID
   return(getpid());
#else
   return 0;
#endif
}

static void
setseed(void)
{
#if defined(HAVE_GETTIMEOFDAY) || defined(X_GETTIMEOFDAY)
   struct timeval mytimeval;
#if defined(HAVE_GETTIMEOFDAY)
   struct timezone mytimezone;
   gettimeofday(&mytimeval, &mytimezone);
#else
   X_GETTIMEOFDAY(&mytimeval);
#endif
#if defined(HAVE_SRAND)
   srand(mytimeval.tv_usec);
#elif defined(HAVE_SRANDOM)
   srandom(mytimeval.tv_usec);
#else
#warning no srandom() or srand()!
#endif
#else
#if defined(HAVE_SRAND)
   srand(time(0) + get_pid());
#elif defined(HAVE_SRANDOM)
   srandom(mytimeval.tv_usec);
#else
#warning no srandom() or srand()!
#endif
#endif
}

static int
getrandom(void)
{
#if defined(HAVE_RAND)
   return(rand());
#elif defined(HAVE_RANDOM)
   return(random());
#else
   return(7653176 + get_pid());
#endif
}

static void
waitrandom(void)
{
   if(getrandom() & 1) sleep(1);
}

/* Our exit function */
static void
cleanupandexit(int errcode)
{
    int i;

    /* Free mode name strings */
    for(i = 0; i < MYMAXMODES; i++) {
       if(sc.modetextptr[i]) {
          free(sc.modetextptr[i]);
          sc.modetextptr[i] = NULL;
       }
#ifdef HAVERANDR
       if(sc.sizetextptr[i]) {
          free(sc.sizetextptr[i]);
	  sc.sizetextptr[i] = NULL;
       }
#endif
    }

    /* Free modeinfo array returned from GetAllModeLines */
    if(sc.vmmodecount) {
       for(i = 0; i < sc.vmmodecount; i++) {
          if(sc.vmmodelines[i]->privsize > 0)
             XFree(sc.vmmodelines[i]->private);
       }
       XFree(sc.vmmodelines);
       sc.vmmodecount = 0;
    }

#ifdef HAVE_RANDR
    if(sc.sconf) XRRFreeScreenConfigInfo(sc.sconf);
#endif

#ifdef USEXV
    if(!sc.noxvdemo) {
       destroyxvimage();
    }
#endif

    /* Close display */
    if(sc.dpy) {
       XCloseDisplay(sc.dpy);
       sc.dpy = NULL;
    }

    /* Close and remove our lockfile */
    if(sc.lockfile != -1) {
       close(sc.lockfile);
       remove(sc.filenamebuffer);
    }

    if(errcode) exit(errcode);
}

RETSIGTYPE signal_handler(int sig)
{
   switch(sig) {
   case SIGHUP:
   case SIGTERM:
   case SIGINT:
      cleanupandexit(0);
      if(sc.window) gtk_main_quit();
      else exit(0);
   }
}

static int
calcrefresh(XF86VidModeModeInfo *modeinfo)
{
   float hsync, refresh = 0;

   if(modeinfo->htotal > 0)
       	hsync = (float)modeinfo->dotclock / (float)modeinfo->htotal;
   else return 0;

   if(modeinfo->vtotal > 0)
       	refresh = hsync * 1000.0 / modeinfo->vtotal;
   else return 0;

   if(modeinfo->flags & V_INTERLACE)
       	refresh *= 2.0;

   if(modeinfo->flags & V_DBLSCAN)
       	refresh /= 2.0;

   return(int)(refresh);
}

static Bool
UnlockSiSDirect(int unlock)
{
   Bool ret = TRUE;
   int key = unlock ? SIS_DIRECTKEY : 0;
   int result;

   if(XvSetPortAttribute(sc.dpy, sc.xv_port_nr, sc.atom_usd, key) != Success) {
      fprintf(stderr, "Failed to unlock/lock SiS direct access\n");
      ret = FALSE;
   }
   XFlush(sc.dpy);

   return ret;
}

static int
getxvval(Atom atom)
{
   int val = 0;

   if(UnlockSiSDirect(1)) {
      if(XvGetPortAttribute(sc.dpy, sc.xv_port_nr, atom, &val) != Success) {
         fprintf(stderr, xverrorstring);
      }
      UnlockSiSDirect(0);
   }
   return val;
}

static void
setxvval(Atom atom, int val)
{
   if(UnlockSiSDirect(1)) {
      if(XvSetPortAttribute(sc.dpy, sc.xv_port_nr, atom, val) != Success) {
         fprintf(stderr, xverrorstring);
      }
      UnlockSiSDirect(0);
   }
 }

Atom
searchandmakeatom(char *name, int num_attrs, XvAttribute *attributes)
{
    int i;
    Bool found = FALSE;

    for(i = 0; i < num_attrs; i++) {
       if(strcmp(attributes[i].name, name) == 0) {
          found = TRUE;
	  break;
       }
    }

    if(found) return(XInternAtom(sc.dpy, name, False));
    else {
       fprintf(stderr, "Could not find %s attribute, update your X driver\n", name);
       showerror_pregui("Required Xv attribute not found, update your X driver\n", FALSE, FALSE);
       return 0;
    }
}

static char *
getchiptype(void)
{
   char *chip;
   unsigned short chipflags = sc.hwinfo & 0xffff;

   switch((sc.hwinfo >> 16) & 0xff) {
   case SIS_300:    chip = "300";    break;
   case SIS_540:    chip = "540";    break;
   case SIS_630:    chip = "630";    break;
   case SIS_730:    chip = "730";    break;
   case SIS_315H:   chip = "315H";   break;
   case SIS_315:    chip = "315";    break;
   case SIS_315PRO: chip = "315PRO"; break;
   case SIS_550:    chip = "550";    break;
   case SIS_650:    if(chipflags & SiSCF_IsM650)        chip = "M650";
   	            else if(chipflags & SiSCF_Is651)    chip = "651";
		    else if(chipflags & SiSCF_Is652)    chip = "652";
		    else if(chipflags & SiSCF_IsM652)   chip = "M652";
		    else if(chipflags & SiSCF_IsM653)   chip = "M653";
                    else                                chip = "650";
		    break;
   case SIS_740:    chip = "740";    break;
   case SIS_330:    chip = "Xabre";  break;
   case SIS_661:    if(chipflags & SiSCF_IsM661)        chip = "M661FX";
   		    else if(chipflags & SiSCF_IsM661M)  chip = "M661MX";
                    else                                chip = "661FX";
		    break;
   case SIS_741:    if(chipflags & SiSCF_IsM741) chip = "M741";
    		    else 			 chip = "741/741GX";
		    break;
   case SIS_660:    chip = "660";    break;
   case SIS_760:    if(chipflags & SiSCF_IsM760) chip = "M760";
                    else                         chip = "760";
		    break;
   default:         chip = "UNKNOWN";
   }
   return chip;
}

static Bool
vbsupportsvga2(void)
{
   if(sc.sdflags & SiS_SD_ADDLSUPFLAG) {
      if(sc.sdflags & SiS_SD_SUPPORTVGA2) return TRUE;
   } else {
      if(sc.currentvbflags & (VB_301|VB_301B|VB_301C|VB_302B)) return TRUE;
   }
   return FALSE;
}

static Bool
vbsupportsscart(void)
{
   if(sc.sdflags & SiS_SD_ADDLSUPFLAG) {
      if(sc.sdflags & SiS_SD_SUPPORTSCART) return TRUE;
   } else {
      if(sc.currentvbflags & (VB_301|VB_301B|VB_301C|VB_302B)) return TRUE;
   }
   return FALSE;
}

static Bool
vbsupportstv(void)
{
   if(sc.sdflags & SiS_SD_ADDLSUPFLAG) {
      if(sc.sdflags & SiS_SD_SUPPORTTV) return TRUE;
   } else {
      if(sc.currentvbflags & (VB_SISTVBRIDGE | VB_CHRONTEL)) return TRUE;
   }
   return FALSE;
}

static Bool
vbsupportsoverscan(void)
{
   if(sc.sdflags & SiS_SD_ADDLSUPFLAG) {
      if(sc.sdflags & SiS_SD_SUPPORTOVERSCAN) return TRUE;
   } else {
      if(sc.currentvbflags & VB_CHRONTEL) return TRUE;
   }
   return FALSE;
}

int finddmhistory(void)
{
    gint index = 0;
    int i;
    int vmdotclock;
    XF86VidModeModeLine vmmodeline;

    if(XF86VidModeGetModeLine(sc.dpy, sc.myscreen, &vmdotclock, &vmmodeline)) {
       for(i = 0; (i < sc.vmmodecount) && (i < 64); i++) {
          if((vmdotclock == sc.vmmodelines[i]->dotclock)          &&
	     (vmmodeline.hdisplay == sc.vmmodelines[i]->hdisplay) &&
	     (vmmodeline.hsyncstart == sc.vmmodelines[i]->hsyncstart) &&
	     (vmmodeline.hsyncend == sc.vmmodelines[i]->hsyncend) &&
	     (vmmodeline.htotal == sc.vmmodelines[i]->htotal) &&
	     (vmmodeline.vdisplay == sc.vmmodelines[i]->vdisplay) &&
	     (vmmodeline.vsyncstart == sc.vmmodelines[i]->vsyncstart) &&
	     (vmmodeline.vsyncend == sc.vmmodelines[i]->vsyncend) &&
	     (vmmodeline.vtotal == sc.vmmodelines[i]->vtotal) &&
	     (vmmodeline.flags == sc.vmmodelines[i]->flags)) {
	     index = i;
	     break;
	  }
       }
    } else {
       fprintf(stderr, "Failed to query current modeline\n");
    }
    return index;
}

static Bool
checkmodeforscreen(dm_index)
{
   if(sc.HaveGss) {
      int sizeX, sizeY;
      unsigned long val;
      val = getxvval(sc.atom_gss);
      sizeX = val >> 16;
      sizeY = val & 0xffff;
      if((sc.vmmodelines[dm_index]->hdisplay > sizeX) ||
         (sc.vmmodelines[dm_index]->vdisplay > sizeY))
	 return FALSE;

   }
   return TRUE;
}

/* The API for mode checking is not entirely bullet-proof. If
 * more than one program is querying the server, this might
 * lead to incorrect results. I tried to circumvent this by
 * using a (hopefully) unique ID and checking this ID with
 * the result, but congestions might happen anyway.
 */

static int
checkmodeforgivencrt2(int index, unsigned short cond)
{
   unsigned long myid, i = 25;
   int myidr;
   Bool warned = FALSE;

   if(index > 255) {
        fprintf(stderr, "Cannot check mode, index too high (%d)\n", index);
   	return 0x80;
   }

   do {

      if(!UnlockSiSDirect(1)) return 0x80;

      setseed();
      myid = getrandom();
      myid = ((((myid & 0xffff0000) >> 16) ^ (myid & 0xffff)) + get_pid()) & 0xffff;

      myid = (myid << 16) | ((index & 0xff) << 8) | (cond & 0xff);

      if(XvSetPortAttribute(sc.dpy, sc.xv_port_nr, sc.atom_cmd, myid) != Success) {
         fprintf(stderr, xverrorstring);
      }
      XFlush(sc.dpy);
      if(XvGetPortAttribute(sc.dpy, sc.xv_port_nr, sc.atom_cmdr, &myidr) != Success) {
         fprintf(stderr, xverrorstring);
      }
      UnlockSiSDirect(0);

      if((myid & 0xffffff00) == (myidr & 0xffffff00)) {
         return(myidr & 0xff);
      } else {
         if(!warned) fprintf(stderr, "SiS SD API congestion - is there another session running?\n");
	 warned = TRUE;
         waitrandom();
      }

   } while(i--);

   fprintf(stderr, "25 attempts to check mode failed, giving up\n");
   return 0x80;

}

static int
checkmodeforcrt2(int index)
{
   return(checkmodeforgivencrt2(index, 0));
}

static void
rebuildmodefieldlist(void)
{
   int i;

   for(i = 0; (i < sc.vmmodecount) && (i < MYMAXMODES); i++) {
      if(sc.modetextptr[i]) {
         sc.modecheckfield[i] = 0;
	 if((sc.currentvbflags & VB_VIDEOBRIDGE) && (!(sc.sdflags & SiS_SD_ISDUALHEAD))) {
            if(!(checkmodeforgivencrt2(i, SiS_CF2_LCD) & 0x01)) {
	       sc.modecheckfield[i] |= SiS_CRT2_LCD;
	    }
	    if(sc.sdflags & SiS_SD_SUPPORTLCDA) {
	       if(!(checkmodeforgivencrt2(i, SiS_CF2_CRT1LCDA) & 0x01)) {
	          sc.modecheckfield[i] |= SiS_CRT2_CRT1LCDA;
	       }
	    }
	    if(!(checkmodeforgivencrt2(i, (SiS_CF2_TV | SiS_CF2_TVPAL)) & 0x01)) {
	       sc.modecheckfield[i] |= (SiS_CRT2_TV | SiS_CRT2_TVPAL);
	    }
	    if(!(checkmodeforgivencrt2(i, (SiS_CF2_TV | SiS_CF2_TVNTSC)) & 0x01)) {
	       sc.modecheckfield[i] |= (SiS_CRT2_TV | SiS_CRT2_TVNTSC);
	    }
	    if(sc.sdflags & SiS_SD_SUPPORTPALMN) {
	       if(!(checkmodeforgivencrt2(i, (SiS_CF2_TV | SiS_CF2_TVPALM)) & 0x01)) {
	          sc.modecheckfield[i] |= (SiS_CRT2_TV | SiS_CRT2_TVPALM);
	       }
	       if(!(checkmodeforgivencrt2(i, (SiS_CF2_TV | SiS_CF2_TVPALN)) & 0x01)) {
	          sc.modecheckfield[i] |= (SiS_CRT2_TV | SiS_CRT2_TVPALN);
	       }
	    }
	    if(sc.sdflags & SiS_SD_SUPPORTHIVISION) {
	       if(!(checkmodeforgivencrt2(i, (SiS_CF2_TVSPECIAL | SiS_CF2_TVHIVISION)) & 0x01)) {
	          sc.modecheckfield[i] |= SiS_CRT2_HIVISION;
	       }
	    }
	    if(sc.sdflags & SiS_SD_SUPPORTYPBPR) {
	       if(!(checkmodeforgivencrt2(i, (SiS_CF2_TVSPECIAL | SiS_CF2_TVYPBPR525I)) & 0x01)) {
	          sc.modecheckfield[i] |= SiS_CRT2_YPBPR525I;
	       }
	       if(!(checkmodeforgivencrt2(i, (SiS_CF2_TVSPECIAL | SiS_CF2_TVYPBPR525P)) & 0x01)) {
	          sc.modecheckfield[i] |= SiS_CRT2_YPBPR525P;
	       }
	       if(!(checkmodeforgivencrt2(i, (SiS_CF2_TVSPECIAL | SiS_CF2_TVYPBPR750P)) & 0x01)) {
	          sc.modecheckfield[i] |= SiS_CRT2_YPBPR750P;
	       }
	       if(!(checkmodeforgivencrt2(i, (SiS_CF2_TVSPECIAL | SiS_CF2_TVYPBPR1080I)) & 0x01)) {
	          sc.modecheckfield[i] |= SiS_CRT2_YPBPR1080I;
	       }
	    }
	    if(vbsupportsvga2()) {
	       if(!(checkmodeforgivencrt2(i, SiS_CF2_VGA2) & 0x01)) {
	          sc.modecheckfield[i] |= SiS_CRT2_VGA2;
	       }
	    }
	 }
      }
   }
}

#ifdef HAVE_RANDR
static void
build_randr_sizes(void)
{
    int i;
    char buffer[80];

    sc.rrsizes = XRRConfigSizes(sc.sconf, &sc.rrnsize);
    if(sc.rrnsize > MYMAXMODES) sc.rrnsize = MYMAXMODES;
    for(i=0; i < sc.rrnsize; i++) {
        sprintf(buffer, "%dx%d (ID %d)",
		sc.rrsizes[i].width,sc.rrsizes[i].height, i);
	if(!sc.sizetextptr[i]) {
	   sc.sizetextptr[i] = malloc(strlen(buffer)+1);
	}
	if(sc.sizetextptr[i]) {
           strcpy(sc.sizetextptr[i], buffer);
	}
    }
    strcpy(sc.leaveunchangedtext, "Leave unchanged");
}
#endif

static Bool
updatevbandcrt1(void)
{
    int val;

    if(XvGetPortAttribute(sc.dpy, sc.xv_port_nr, sc.atom_qvf, &val) != Success) {
       fprintf(stderr, "Couldn't read vbflags\n");
       return FALSE;
    }
    sc.currentvbflags = (unsigned long)val;

    if(XvGetPortAttribute(sc.dpy, sc.xv_port_nr, sc.atom_ct1, &val) != Success) {
       fprintf(stderr, "Couldn't read CRT1 status\n");
       return FALSE;
    }
    sc.CRT1isoff = val ? 0 : 1;
    return TRUE;
}

static void
getgamma(void)
{
   if(!XF86VidModeGetGamma(sc.dpy, sc.myscreen, &sc.currentgamma)) {
      fprintf(stderr, "Unable to get gamma correction\n");
   }
}

static void
setgamma(void)
{
#ifdef HAVE_MATH_H
   int i, j, k;
#endif

   if(!XF86VidModeSetGamma(sc.dpy, sc.myscreen, &sc.newgamma)) {
      fprintf(stderr, "Unable to set gamma correction\n");
      return;
   }
   XFlush(sc.dpy);

#ifndef HAVE_MATH_H
   getgamma();
#else

   if(!XF86VidModeGetGammaRampSize(sc.dpy, sc.myscreen, &sc.nramp)) {
      fprintf(stderr, "Unable to get gamma ramp size\n");
      return;
   }

   for(i=0; i<3; i++) {
      sc.ramp[i] = (unsigned short *)malloc(sc.nramp * sizeof(unsigned short));
      if(!sc.ramp[i]) {
         if(sc.ramp[0]) { free(sc.ramp[0]); sc.ramp[0] = NULL; }
	 if(sc.ramp[1]) { free(sc.ramp[1]); sc.ramp[1] = NULL; }
	 fprintf(stderr, "Memory allocation error\n");
         return;
      }
   }

   for(i = 0; i < 3; i++) {
      float dramp = 1. / (sc.nramp - 1);
      int fullscale = 65535 * sc.gamma_max[i];
      float invgamma, v;

      switch(i) {
      case 0: invgamma = 1. / sc.newgamma.red; break;
      case 1: invgamma = 1. / sc.newgamma.green; break;
      case 2: invgamma = 1. / sc.newgamma.blue; break;
      }

      for(j = 0; j < sc.nramp; j++) {
         sc.framp = pow(sc.gamma_prescale[i] * j * dramp, invgamma);

         v = (fullscale < 0) ? (65535 + fullscale * sc.framp) :
	 		       fullscale * sc.framp;
	 if(v < 0) v = 0;
	 else if(v > 65535) v = 65535;
	 sc.ramp[i][j] = (unsigned short)v;
      }
   }

   /* border */
   for(i = 0; i < GACW*3; i++) sc.rgbbuf[i] = 0xff;
   for(i = 1; i < GACH-1; i++) {
      sc.rgbbuf[i * GACW * 3] = 0xff;
      sc.rgbbuf[(i * GACW * 3) + 1] = 0xff;
      sc.rgbbuf[(i * GACW * 3) + 2] = 0xff;
      for(j = 3; j < (GACW-1) * 3; j++) sc.rgbbuf[(i * GACW * 3) + j] = 0;
      sc.rgbbuf[(i * GACW * 3) + ((GACW - 1) * 3)] = 0x44;
      sc.rgbbuf[(i * GACW * 3) + ((GACW - 1) * 3) + 1] = 0x44;
      sc.rgbbuf[(i * GACW * 3) + ((GACW - 1) * 3) + 2] = 0x44;
   }
   for(i = 0; i < GACW*3; i++) sc.rgbbuf[((GACH-1) * GACW * 3) + i] = 0x44;

   /* curve */
   for(i = 0; i < 3; i++) {
      for(k = 1; k < GACW-1; k++) {
         j = (k - 1) * (sc.nramp / (GACW-2));
         sc.rgbbuf[(((GACH-3) - (sc.ramp[i][j] / (65535 / (GACH-3)))) * (GACW * 3)) + (GACW * 3) + (k * 3) + i] = 0xff;
      }
   }

   update_gamma_gfx();

   if(!XF86VidModeSetGammaRamp(sc.dpy, sc.myscreen, sc.nramp, sc.ramp[0], sc.ramp[1], sc.ramp[2])) {
      fprintf(stderr, "Unable to set gamma correction\n");
   }
   XFlush(sc.dpy);

   for(i = 0; i < 3; i++) {
       if(sc.ramp[i]) free(sc.ramp[i]);
       sc.ramp[i] = NULL;
   }

   memcpy(&sc.currentgamma, &sc.newgamma, sizeof(XF86VidModeGamma));
   if(sc.HaveBri) {
      setxvval(sc.atom_brr, (int)(sc.gamma_max[0] * 1000));
      setxvval(sc.atom_brg, (int)(sc.gamma_max[1] * 1000));
      setxvval(sc.atom_brb, (int)(sc.gamma_max[2] * 1000));
      setxvval(sc.atom_pbr, (int)(sc.gamma_prescale[0] * 1000));
      setxvval(sc.atom_pbg, (int)(sc.gamma_prescale[1] * 1000));
      setxvval(sc.atom_pbb, (int)(sc.gamma_prescale[2] * 1000));
   }
#endif
}

static void
resetgamma(void)
{
   if(!XF86VidModeGetGamma(sc.dpy, sc.myscreen, &sc.currentgamma)) {
      fprintf(stderr, "Unable to get gamma correction\n");
      return;
   }
   XFlush(sc.dpy);
#ifdef HAVE_MATH_H
   setgamma();
#else
   if(!XF86VidModeSetGamma(sc.dpy, sc.myscreen, &sc.currentgamma)) {
      fprintf(stderr, "Unable to set gamma correction\n");
   }
   XFlush(sc.dpy);
#endif
}

static void
endisablegamma1(Bool enable)
{
   sc.gammaflags = getxvval(sc.atom_sga) & ~0x01;
   if(enable) sc.gammaflags |= 0x01;
   if(sc.sdflags & SiS_SD_ISDUALHEAD) {
      sc.backupgammaflags &= 0x01;
      sc.backupgammaflags |= sc.gammaflags & ~0x01;
   }
   setxvval(sc.atom_sga, sc.gammaflags);
   resetgamma();
   setxvgammastatus();
}

static void
endisablegamma2(Bool enable)
{
   sc.gammaflags = getxvval(sc.atom_sga) & ~0x02;
   if(enable) sc.gammaflags |= 0x02;
   if(sc.sdflags & SiS_SD_ISDUALHEAD) {
      sc.backupgammaflags &= 0x02;
      sc.backupgammaflags |= sc.gammaflags & ~0x02;
   }
   setxvval(sc.atom_sga, sc.gammaflags);
   resetgamma();
}

static void
endisablexvgamma(Bool enable)
{
   sc.gammaflags = getxvval(sc.atom_sga) & ~0x04;
   if(enable) sc.gammaflags |= 0x04;
   if(sc.sdflags & SiS_SD_ISDUALHEAD) {
      sc.backupgammaflags &= 0x04;
      sc.backupgammaflags |= sc.gammaflags & ~0x04;
   }
   setxvval(sc.atom_sga, sc.gammaflags);
   setxvgammasensitive();
   if(!(sc.gammaflags & 0x04)) resetgamma();
}

static void
setxvgamma(Atom myatom, int value)
{
   if(!(sc.sdflags & SiS_SD_SUPPORTXVGAMMA1)) return;
   if(value < 100) value = 100;
   if(value > 10000) value = 10000;
   setxvval(myatom, value);
}

static void
setxvgammared(void)
{
   setxvgamma(sc.atom_vgr, (int)(sc.xvgammared * 1000));
}

static void
setxvgammagreen(void)
{
   setxvgamma(sc.atom_vgg, (int)(sc.xvgammagreen * 1000));
}

static void
setxvgammablue(void)
{
   setxvgamma(sc.atom_vgb, (int)(sc.xvgammablue * 1000));
}

static void
dotvposscale(Atom myatom, Bool Up, int limit)
{
   int val;

   if(XvGetPortAttribute(sc.dpy, sc.xv_port_nr, myatom, &val) != Success) {
      fprintf(stderr, xverrorstring);
      return;
   }
   if(Up) {
      if(val >= limit) return;
      val++;
   } else {
      if(val <= limit) return;
      val--;
   }
   if(UnlockSiSDirect(1)) {
      if(XvSetPortAttribute(sc.dpy, sc.xv_port_nr, myatom, val) != Success) {
         fprintf(stderr, xverrorstring);
      }
      XFlush(sc.dpy);
      UnlockSiSDirect(0);
   }
}

static void
tvdefaultposscale(void)
{
   if(UnlockSiSDirect(1)) {
      if(XvSetPortAttribute(sc.dpy, sc.xv_port_nr, sc.atom_tvx, 0) != Success) {
         fprintf(stderr, xverrorstring);
      }
      XFlush(sc.dpy);
      if(XvSetPortAttribute(sc.dpy, sc.xv_port_nr, sc.atom_tvy, 0) != Success) {
         fprintf(stderr, xverrorstring);
      }
      if(sc.currentvbflags & VB_SISBRIDGE) {
         if(XvSetPortAttribute(sc.dpy, sc.xv_port_nr, sc.atom_txs, 0) != Success) {
            fprintf(stderr, xverrorstring);
	 }
	 if(XvSetPortAttribute(sc.dpy, sc.xv_port_nr, sc.atom_tys, 0) != Success) {
            fprintf(stderr, xverrorstring);
	 }
      }
      XFlush(sc.dpy);
      UnlockSiSDirect(0);
   }
}

/* Get CRT2 type menu index from vbflags */
static int
getcrt2typeindex(unsigned long vbflags)
{
  int index;

  if(vbflags & CRT2_LCD)             index = MI_TV_LCD;
  else if(vbflags & CRT2_VGA)        index = MI_TV_VGA;
  else if(vbflags & CRT2_TV) {
     if(vbflags & VB_SISBRIDGE) {
        if((vbflags & (TV_SVIDEO | TV_AVIDEO)) == (TV_SVIDEO | TV_AVIDEO)) index = MI_TV_CVSV;
	else if(vbflags & TV_SVIDEO)   index = MI_TV_SVHS;
        else if(vbflags & TV_AVIDEO)   index = MI_TV_CVBS;
        else if(vbflags & TV_SCART)    index = MI_TV_SCART;
	else if(vbflags & TV_HIVISION) index = MI_TV_HIVI;
	else if(vbflags & TV_YPBPR) {
	   if(vbflags & TV_YPBPR525I)	    index = MI_TV_525I;
           else if(vbflags & TV_YPBPR525P)  index = MI_TV_525P;
           else if(vbflags & TV_YPBPR750P)  index = MI_TV_750P;
	   else if(vbflags & TV_YPBPR1080I) index = MI_TV_1080I;
	} else		             index = MI_TV_SVHS;
     } else                          index = MI_TV_TV;
  } else                             index = MI_TV_OFF;

  return index;
}

/* Get TV standard menu index from vbflags */
static int
gettvtypeindex(unsigned long vbflags)
{
    int index = 0;

    if(vbflags & TV_NTSC) {
       if(vbflags & TV_NTSCJ) index = 4;
       else index = 1;
    } else {
       if(vbflags & TV_PALM) index = 2;
       else if(vbflags & TV_PALN) index = 3;
       else index = 0;
    }
    return index;
}

/**************************************************/
/*             GUI toolkit specific               */
/**************************************************/

void
showerror(gchar *message)
{
    GtkWidget *dialog, *label;

    dialog = gtk_dialog_new_with_buttons("Error", GTK_WINDOW(sc.window),
              		GTK_DIALOG_DESTROY_WITH_PARENT | GTK_DIALOG_MODAL,
	      		GTK_STOCK_OK,
	      		GTK_RESPONSE_NONE,
			NULL);

    label = gtk_label_new(message);

    g_signal_connect_swapped(GTK_OBJECT(dialog), "response", G_CALLBACK(gtk_widget_destroy), GTK_OBJECT(dialog));

    gtk_container_add(GTK_CONTAINER(GTK_DIALOG(dialog)->vbox), label);

    gtk_widget_show_all(dialog);
}

Bool
showerror_pregui(gchar *message, Bool yesno, Bool warning)
{
   GtkWidget *dialog;
   gint result;

   dialog = gtk_message_dialog_new(NULL,
              		   GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT,
			   warning ? GTK_MESSAGE_WARNING : GTK_MESSAGE_ERROR,
			   yesno ? GTK_BUTTONS_YES_NO : GTK_BUTTONS_OK,
	      		   message);

   result = gtk_dialog_run(GTK_DIALOG(dialog));

   gtk_widget_destroy(dialog);

   switch(result) {
   case GTK_RESPONSE_YES:
      return TRUE;
   default:
      return FALSE;
   }
}

static GtkWidget *xpm_label_box(char **xpmdata, gchar *labeltext, GtkWidget *toactivate, int space)
{
   GtkWidget *box, *label = NULL, *image, *align;
   GdkPixmap *icon = NULL;
   GdkBitmap *mask;

   box = gtk_hbox_new(FALSE, 0);
   gtk_container_set_border_width(GTK_CONTAINER(box), 1);

   align = gtk_alignment_new(0.5, 0.5, 0, 0);
   gtk_container_add(GTK_CONTAINER(align), box);

   icon = gdk_pixmap_create_from_xpm_d(sc.window->window, &mask, NULL, xpmdata);
   image = gtk_pixmap_new(icon, mask);

   gtk_box_pack_start(GTK_BOX(box), image, FALSE, FALSE, space);

   if(labeltext) {
      label = gtk_label_new_with_mnemonic(labeltext);
      gtk_label_set_justify(GTK_LABEL(label), GTK_JUSTIFY_LEFT);
      gtk_box_pack_start(GTK_BOX(box), label, FALSE, FALSE, space);
      gtk_widget_show(label);
   }

   gtk_widget_show(box);
   gtk_widget_show(image);

   if((label) && (toactivate)) {
      gtk_label_set_mnemonic_widget(GTK_LABEL(label), toactivate);
   }

   return align;
}

/* Update CRT1 gfx */
static void
updatecrt1gfx(void)
{
   int c1_index = gtk_option_menu_get_history(GTK_OPTION_MENU(sc.c1_omenu));

   if(c1_index) {
      gtk_widget_hide(sc.c1_imageoff);
      if((sc.sdflags & SiS_SD_SUPPORTLCDA) && (c1_index == 2)) {
        gtk_widget_hide(sc.c1_imageon);
	gtk_widget_show(sc.c1_imagelcda);
      } else {
        gtk_widget_hide(sc.c1_imagelcda);
	gtk_widget_show(sc.c1_imageon);
      }
   } else {
      gtk_widget_hide(sc.c1_imageon);
      gtk_widget_hide(sc.c1_imagelcda);
      gtk_widget_show(sc.c1_imageoff);
   }
}

/* Update CRT2 gfx */
static void
updatecrt2gfx(void)
{
   int om_index = gtk_option_menu_get_history(GTK_OPTION_MENU(sc.om_omenu));

   gtk_widget_hide(sc.om_lcd);
   gtk_widget_hide(sc.om_tv);
   gtk_widget_hide(sc.om_tvs);
   gtk_widget_hide(sc.om_tvc);
   gtk_widget_hide(sc.om_ypbpr);
   gtk_widget_hide(sc.om_vga);

   switch(om_index) {
   case MI_TV_LCD:
      gtk_widget_show(sc.om_lcd);
      break;
   case MI_TV_TV:
   case MI_TV_SCART:
   case MI_TV_CVSV:
      gtk_widget_show(sc.om_tv);
      break;
   case MI_TV_HIVI:
   case MI_TV_525I:
   case MI_TV_525P:
   case MI_TV_750P:
   case MI_TV_1080I:
      gtk_widget_show(sc.om_ypbpr);
      break;
   case MI_TV_SVHS:
      gtk_widget_show(sc.om_tvs);
      break;
   case MI_TV_CVBS:
      gtk_widget_show(sc.om_tvc);
      break;
   case MI_TV_VGA:
      gtk_widget_show(sc.om_vga);
      break;
   }
}

/* Set Displaymode history (currently selected item) */
static int
setdmhistory(void)
{
    gint index = (gint)finddmhistory();

    gtk_tree_selection_unselect_all(sc.dm_select);
    gtk_tree_selection_select_iter(sc.dm_select, &sc.dm_iter[index]);
    return index;
}

#ifdef HAVE_RANDR
static void
updateleaveunchanged(void)
{
   SizeID current_size;
   Rotation current_rot;
   GtkWidget *label;
   gint  current_menu;

   current_size = XRRConfigCurrentConfiguration(sc.sconf, &current_rot);
   sprintf(sc.leaveunchangedtext, "Leave unchanged at %dx%d (ID %d)",
       sc.rrsizes[current_size].width,
       sc.rrsizes[current_size].height,
       current_size);

   current_menu = gtk_option_menu_get_history(GTK_OPTION_MENU(sc.dm_omenu));
   if(current_menu == 0) {
      gtk_option_menu_set_history(GTK_OPTION_MENU(sc.dm_omenu), 1);
   }
   label = gtk_bin_get_child(GTK_BIN(sc.dm_menuitem[0]));
   gtk_label_set_text(GTK_LABEL(label), sc.leaveunchangedtext);
   if(current_menu == 0) {
      gtk_option_menu_set_history(GTK_OPTION_MENU(sc.dm_omenu), 0);
   }
}
#endif

/* Add crt2 type icons to display mode field list */
static void
setcrt2typesformodelist(void)
{
    int i;

    if(!sc.pixbuf_lcd)     sc.pixbuf_lcd = gdk_pixbuf_new_from_xpm_data((const char **)icon_lcd);
    if(!sc.pixbuf_lcda)    sc.pixbuf_lcda = gdk_pixbuf_new_from_xpm_data((const char **)icon_lcda);
    if(!sc.pixbuf_tvpal)   sc.pixbuf_tvpal = gdk_pixbuf_new_from_xpm_data((const char **)icon_tvpal);
    if(!sc.pixbuf_tvpalm)  sc.pixbuf_tvpalm = gdk_pixbuf_new_from_xpm_data((const char **)icon_tvpalm);
    if(!sc.pixbuf_tvpaln)  sc.pixbuf_tvpaln = gdk_pixbuf_new_from_xpm_data((const char **)icon_tvpaln);
    if(!sc.pixbuf_tvntsc)  sc.pixbuf_tvntsc = gdk_pixbuf_new_from_xpm_data((const char **)icon_tvntsc);
    if(!sc.pixbuf_tvhivi)  sc.pixbuf_tvhivi = gdk_pixbuf_new_from_xpm_data((const char **)icon_tvhivi);
    if(!sc.pixbuf_tv525i)  sc.pixbuf_tv525i = gdk_pixbuf_new_from_xpm_data((const char **)icon_tv525i);
    if(!sc.pixbuf_tv525p)  sc.pixbuf_tv525p = gdk_pixbuf_new_from_xpm_data((const char **)icon_tv525p);
    if(!sc.pixbuf_tv750p)  sc.pixbuf_tv750p = gdk_pixbuf_new_from_xpm_data((const char **)icon_tv750p);
    if(!sc.pixbuf_tv1080i) sc.pixbuf_tv1080i = gdk_pixbuf_new_from_xpm_data((const char **)icon_tv1080i);
    if(!sc.pixbuf_vga2)    sc.pixbuf_vga2 = gdk_pixbuf_new_from_xpm_data((const char **)icon_vga2);

    for(i = 0; i < MYMAXMODES; i++) {
       if(sc.modetextptr[i]) {
          if(sc.modecheckfield[i] & SiS_CRT2_LCD) {
             gtk_list_store_set(sc.dm_liststore, &sc.dm_iter[i], 1, sc.pixbuf_lcd, -1);
	  } else {
	     gtk_list_store_set(sc.dm_liststore, &sc.dm_iter[i], 1, NULL, -1);
	  }
	  if((sc.sdflags & SiS_SD_SUPPORTLCDA) && (sc.modecheckfield[i] & SiS_CRT2_CRT1LCDA)) {
	     gtk_list_store_set(sc.dm_liststore, &sc.dm_iter[i], 2, sc.pixbuf_lcda, -1);
	  } else {
	     gtk_list_store_set(sc.dm_liststore, &sc.dm_iter[i], 2, NULL, -1);
	  }
	  if(sc.modecheckfield[i] & SiS_CRT2_TV) {
             gtk_list_store_set(sc.dm_liststore, &sc.dm_iter[i], 3,
	     			(sc.modecheckfield[i] & SiS_CRT2_TVPAL) ? sc.pixbuf_tvpal : NULL, -1);
	     gtk_list_store_set(sc.dm_liststore, &sc.dm_iter[i], 4,
	     			(sc.modecheckfield[i] & SiS_CRT2_TVNTSC) ? sc.pixbuf_tvntsc : NULL, -1);
	     if(sc.sdflags & SiS_SD_SUPPORTPALMN) {
	        gtk_list_store_set(sc.dm_liststore, &sc.dm_iter[i], 5,
	     			(sc.modecheckfield[i] & SiS_CRT2_TVPALM) ? sc.pixbuf_tvpalm : NULL, -1);
	        gtk_list_store_set(sc.dm_liststore, &sc.dm_iter[i], 6,
	     			(sc.modecheckfield[i] & SiS_CRT2_TVPALN) ? sc.pixbuf_tvpaln : NULL, -1);
	     } else {
	        gtk_list_store_set(sc.dm_liststore, &sc.dm_iter[i], 5, NULL, -1);
		gtk_list_store_set(sc.dm_liststore, &sc.dm_iter[i], 6, NULL, -1);
	     }
	  }
	  if(sc.modecheckfield[i] & SiS_CRT2_HIVISION) {
	     gtk_list_store_set(sc.dm_liststore, &sc.dm_iter[i], 7, sc.pixbuf_tvhivi, -1);
	  } else {
	     gtk_list_store_set(sc.dm_liststore, &sc.dm_iter[i], 7, NULL, -1);
	  }
	  if(sc.modecheckfield[i] & SiS_CRT2_YPBPR525I) {
	     gtk_list_store_set(sc.dm_liststore, &sc.dm_iter[i], 8, sc.pixbuf_tv525i, -1);
	  } else {
	     gtk_list_store_set(sc.dm_liststore, &sc.dm_iter[i], 8, NULL, -1);
	  }
	  if(sc.modecheckfield[i] & SiS_CRT2_YPBPR525P) {
	     gtk_list_store_set(sc.dm_liststore, &sc.dm_iter[i], 9, sc.pixbuf_tv525p, -1);
	  } else {
	     gtk_list_store_set(sc.dm_liststore, &sc.dm_iter[i], 9, NULL, -1);
	  }
	  if(sc.modecheckfield[i] & SiS_CRT2_YPBPR750P) {
	     gtk_list_store_set(sc.dm_liststore, &sc.dm_iter[i], 10, sc.pixbuf_tv750p, -1);
	  } else {
	     gtk_list_store_set(sc.dm_liststore, &sc.dm_iter[i], 10, NULL, -1);
	  }
	  if(sc.modecheckfield[i] & SiS_CRT2_YPBPR1080I) {
	     gtk_list_store_set(sc.dm_liststore, &sc.dm_iter[i], 11, sc.pixbuf_tv1080i, -1);
	  } else {
	     gtk_list_store_set(sc.dm_liststore, &sc.dm_iter[i], 11, NULL, -1);
	  }
	  if((vbsupportsvga2()) && (sc.modecheckfield[i] & SiS_CRT2_VGA2)) {
             gtk_list_store_set(sc.dm_liststore, &sc.dm_iter[i], 12, sc.pixbuf_vga2, -1);
	  } else {
	     gtk_list_store_set(sc.dm_liststore, &sc.dm_iter[i], 12, NULL, -1);
	  }
       }
    }

}

/* Set CRT1 on/off history (currently selected item) */
static void
setcrt1history()
{
    if((sc.sdflags & SiS_SD_ISDUALHEAD) || (!(sc.currentvbflags & VB_VIDEOBRIDGE))) return;

    if(sc.sdflags & SiS_SD_SUPPORTLCDA) {
       int index = 1;
       if(sc.CRT1isoff) index = 0;
       else if(sc.currentvbflags & CRT1_LCDA) index = 2;
       gtk_option_menu_set_history(GTK_OPTION_MENU(sc.c1_omenu), index);
    } else {
       gtk_option_menu_set_history(GTK_OPTION_MENU(sc.c1_omenu), sc.CRT1isoff ? 0 : 1);
    }
    updatecrt1gfx();
}

/* Hide all CRT2 types in menu */
static void
hidecrt2typemenu(void)
{
    if((sc.sdflags & SiS_SD_ISDUALHEAD) || (!(sc.currentvbflags & VB_VIDEOBRIDGE))) return;

    gtk_widget_hide(sc.om_menuitemlcd);
    gtk_widget_hide(sc.om_menuitemsvhs);
    gtk_widget_hide(sc.om_menuitemcvbs);
    gtk_widget_hide(sc.om_menuitemcvsv);
    gtk_widget_hide(sc.om_menuitemtv);
    gtk_widget_hide(sc.om_menuitemhivi);
    gtk_widget_hide(sc.om_menuitemy525i);
    gtk_widget_hide(sc.om_menuitemy525p);
    gtk_widget_hide(sc.om_menuitemy750p);
    gtk_widget_hide(sc.om_menuitemy1080i);
    gtk_widget_hide(sc.om_menuitemvga2);
    gtk_widget_hide(sc.om_menuitemscart);
    gtk_widget_hide(sc.om_menuitemoff);
}

/* Show CRT2 type items in menu (depending on hw and detection) */
static void
showcrt2typemenu(unsigned long vbflags, Bool detectedonly)
{
    Bool showsometv = FALSE, showsvid = FALSE, showcvbs = FALSE;

    if((sc.sdflags & SiS_SD_ISDUALHEAD) || (!(sc.currentvbflags & VB_VIDEOBRIDGE))) return;

    if((vbflags & CRT2_LCD) || (sc.currentvbflags & CRT2_LCD)) {
       gtk_widget_show(sc.om_menuitemlcd);
    }
    if(vbsupportstv()) {
       if((vbflags & CRT2_TV) || (sc.currentvbflags & CRT2_TV) || (!detectedonly)) {
          if(sc.currentvbflags & VB_SISBRIDGE) {
	     if((vbflags & TV_SVIDEO) || (sc.currentvbflags & TV_SVIDEO) || (!detectedonly)) {
	        gtk_widget_show(sc.om_menuitemsvhs);
	        showsometv = TRUE;
	        showsvid = TRUE;
	     }
	     if((vbflags & TV_AVIDEO) || (sc.currentvbflags & TV_AVIDEO) || (!detectedonly)) {
                gtk_widget_show(sc.om_menuitemcvbs);
	        showsometv = TRUE;
	        showcvbs = TRUE;
	     }
	     if(showsvid && showcvbs) {
	        gtk_widget_show(sc.om_menuitemcvsv);
	     }
	     if(sc.sdflags & SiS_SD_SUPPORTHIVISION) {
	        if((vbflags & TV_HIVISION) || (sc.currentvbflags & TV_HIVISION) || (!detectedonly)) {
		   gtk_widget_show(sc.om_menuitemhivi);
		   showsometv = TRUE;
		}
	     }
	     if(sc.sdflags & SiS_SD_SUPPORTYPBPR) {
	        if((vbflags & TV_YPBPR) || (sc.currentvbflags & TV_YPBPR) || (!detectedonly)) {
		   gtk_widget_show(sc.om_menuitemy525i);
		   gtk_widget_show(sc.om_menuitemy525p);
		   gtk_widget_show(sc.om_menuitemy750p);
		   gtk_widget_show(sc.om_menuitemy1080i);
		   showsometv = TRUE;
		}
	     }
	     if((vbsupportsscart()) && (sc.sdflags & SiS_SD_VBHASSCART)) {
                if((vbflags & TV_SCART) || (sc.currentvbflags & TV_SCART) || (!detectedonly)) {
	           gtk_widget_show(sc.om_menuitemscart);
		   showsometv = TRUE;
	        }
             }
	     if(detectedonly && (!showsometv)) {  /* TV set, but neither SVIDEO, CVBS or SCART -> show all */
	        gtk_widget_show(sc.om_menuitemsvhs);
	        gtk_widget_show(sc.om_menuitemcvbs);
	        gtk_widget_show(sc.om_menuitemcvsv);
	        if((vbsupportsscart()) &&
	           (sc.sdflags & SiS_SD_VBHASSCART)) {
	           gtk_widget_show(sc.om_menuitemscart);
	        }
		if(sc.sdflags & SiS_SD_SUPPORTHIVISION) {
		   gtk_widget_show(sc.om_menuitemhivi);
		}
		if(sc.sdflags & SiS_SD_SUPPORTYPBPR) {
		   gtk_widget_show(sc.om_menuitemy525i);
		   gtk_widget_show(sc.om_menuitemy525p);
		   gtk_widget_show(sc.om_menuitemy750p);
		   gtk_widget_show(sc.om_menuitemy1080i);
		}
	     }
          } else if(sc.currentvbflags & VB_CHRONTEL) {
	     gtk_widget_show(sc.om_menuitemtv);
          }
       }
    }
    if((vbsupportsvga2()) && (!(sc.sdflags & SiS_SD_VBHASSCART))) {
       if((vbflags & CRT2_VGA) || (sc.currentvbflags & CRT2_VGA) || (!detectedonly)) {
	  gtk_widget_show(sc.om_menuitemvga2);
       }
    }

    if(!sc.CRT1isoff) {
       gtk_widget_show(sc.om_menuitemoff);
    }
}

/* Set CRT2 type history (currently selected item) */
static void
setcrt2typehistory(unsigned long vbflags)
{
    int index = 0;

    if((sc.sdflags & SiS_SD_ISDUALHEAD) || (!(sc.currentvbflags & VB_VIDEOBRIDGE))) return;

    index = getcrt2typeindex(vbflags);
    gtk_option_menu_set_history(GTK_OPTION_MENU(sc.om_omenu), index);
    updatecrt2gfx();
}

/* Show LCD scaling menu(s) */
static void
showlcdmenu(int menunum)
{
   gint index1, index2, index3;

   if(sc.sdflags & SiS_SD_ISDUALHEAD) return;
   if(!(sc.sdflags & SiS_SD_SUPPORTSCALE)) return;

   if(menunum == 2) {
      index1 = gtk_option_menu_get_history(GTK_OPTION_MENU(sc.om_omenu));
      index2 = MI_TV_LCD;
      index3 = gtk_option_menu_get_history(GTK_OPTION_MENU(sc.om_omenuls));
   } else {
      if(!(sc.sdflags & SiS_SD_SUPPORTLCDA)) return;
      index1 = gtk_option_menu_get_history(GTK_OPTION_MENU(sc.c1_omenu));
      index2 = 2;
      index3 = gtk_option_menu_get_history(GTK_OPTION_MENU(sc.c1_omenuls));
   }

   if(index1 == index2) {
      if(menunum == 2) {
         gtk_widget_show(sc.om_omenuls);
	 gtk_widget_show(sc.om_label_lcds);
      } else {
         gtk_widget_show(sc.c1_omenuls);
	 gtk_widget_show(sc.c1_label_lcds);
      }
      if(sc.sdflags & SiS_SD_SUPPORTCENTER) {
         if(index3 == 1) {
	    if(menunum == 2) gtk_widget_hide(sc.om_omenulcc);
	    else	     gtk_widget_hide(sc.c1_omenulcc);
	 } else {
            if(menunum == 2) gtk_widget_show(sc.om_omenulcc);
	    else	     gtk_widget_show(sc.c1_omenulcc);
	 }
      }
   } else {
      if(menunum == 2) {
         gtk_widget_hide(sc.om_omenuls);
	 gtk_widget_hide(sc.om_label_lcds);
      } else {
         gtk_widget_hide(sc.c1_omenuls);
	 gtk_widget_hide(sc.c1_label_lcds);
      }
      if(sc.sdflags & SiS_SD_SUPPORTCENTER) {
         if(menunum == 2) gtk_widget_hide(sc.om_omenulcc);
	 else	          gtk_widget_hide(sc.c1_omenulcc);
      }
   }
}

/* Set LCD menu history (currently selected item) */
static void
setlcdhistory(void)
{
    gint index1=1, index2=1;

    if(sc.sdflags & SiS_SD_ISDUALHEAD) return;
    if(!(sc.sdflags & SiS_SD_SUPPORTSCALE)) return;

    if(sc.lcdscaler & 0x01)      index1 = 0;
    else if(sc.lcdscaler & 0x02) index1 = 2;
    if(sc.om_omenuls) {
       gtk_option_menu_set_history(GTK_OPTION_MENU(sc.om_omenuls), index1);
    }
    if(sc.sdflags & SiS_SD_SUPPORTLCDA) {
       if(sc.c1_omenuls) {
          gtk_option_menu_set_history(GTK_OPTION_MENU(sc.c1_omenuls), index1);
       }
    }
    if(sc.sdflags & SiS_SD_SUPPORTCENTER) {
       if(sc.lcdscaler & 0x04)      index2 = 0;
       else if(sc.lcdscaler & 0x08) index2 = 2;
       if(sc.om_omenulcc) {
          gtk_option_menu_set_history(GTK_OPTION_MENU(sc.om_omenulcc), index2);
       }
       if(sc.sdflags & SiS_SD_SUPPORTLCDA) {
          if(sc.c1_omenulcc) {
             gtk_option_menu_set_history(GTK_OPTION_MENU(sc.c1_omenulcc), index2);
	  }
       }
    }
}

/* Show aspect ratio menu */
static void
showtvasmenu(unsigned long vbflags)
{
   gint index;

   if(sc.sdflags & SiS_SD_ISDUALHEAD) return;
   if(!(sc.sdflags & SiS_SD_SUPPORTYPBPRAR)) return;

   index = gtk_option_menu_get_history(GTK_OPTION_MENU(sc.om_omenu));

   if(index >= MI_TV_525I && index <= MI_TV_1080I) {
      gtk_widget_show(sc.om_omenuas);
      gtk_widget_show(sc.om_label_tvas);
   } else {
      gtk_widget_hide(sc.om_omenuas);
      gtk_widget_hide(sc.om_label_tvas);
   }
}

/* Set Aspect Ratio menu history (currently selected item) */
static void
settvashistory(unsigned long vbflags)
{
    gint index = 0;

    if(sc.sdflags & SiS_SD_ISDUALHEAD) return;
    if(!(sc.sdflags & SiS_SD_SUPPORTYPBPRAR)) return;

    if(sc.currentvbflags & TV_YPBPR) {
       if((sc.currentvbflags & TV_YPBPRAR) == TV_YPBPR43)        index = 0;
       else if((sc.currentvbflags & TV_YPBPRAR) == TV_YPBPR43LB) index = 1;
       else if((sc.currentvbflags & TV_YPBPRAR) == TV_YPBPR169)  index = 2;
    }

    gtk_option_menu_set_history(GTK_OPTION_MENU(sc.om_omenuas), index);
}

/* Show Chrontel-Overscan menu items */
static void
showtvovermenu(unsigned long vbflags)
{
   gint index1, index2;

   if(!(sc.currentvbflags & VB_CHRONTEL)) return;
   if(sc.sdflags & SiS_SD_ISDUALHEAD) return;

   index1 = gtk_option_menu_get_history(GTK_OPTION_MENU(sc.om_omenu));
   index2 = gtk_option_menu_get_history(GTK_OPTION_MENU(sc.om_omenut));

   if(index1 >= MI_TV_SVHS && index1 <= MI_TV_TV) {
      gtk_widget_show(sc.om_omenuo);
      if((index2 == 0) && (sc.sdflags & SiS_SD_SUPPORTSOVER)) {
         gtk_widget_show(sc.om_menuitemsover);
      } else {
	 if(gtk_option_menu_get_history(GTK_OPTION_MENU(sc.om_omenuo)) == 3) {
	    gtk_option_menu_set_history(GTK_OPTION_MENU(sc.om_omenuo), 2);
	 }
	 gtk_widget_hide(sc.om_menuitemsover);
      }
   } else {
      gtk_widget_hide(sc.om_omenuo);
   }
}

/* Set Chrontel-Overscan menu history (currently selected item) */
static void
settvoverhistory(void)
{
    if(!(sc.currentvbflags & VB_CHRONTEL)) return;
    if(sc.sdflags & SiS_SD_ISDUALHEAD) return;

    if(!(sc.sdflags & SiS_SD_SUPPORTSOVER)) {
       if(sc.choverscanstatus == 3) sc.choverscanstatus = 2;
    }

    gtk_option_menu_set_history(GTK_OPTION_MENU(sc.om_omenuo), sc.choverscanstatus);
}

/* Show TV standard menu items, and overscan menu */
static void
showtvtypemenu(unsigned long vbflags)
{
   gint index;

   if((sc.sdflags & SiS_SD_ISDUALHEAD) || (!(sc.currentvbflags & VB_VIDEOBRIDGE))) return;

   index = gtk_option_menu_get_history(GTK_OPTION_MENU(sc.om_omenu));

   if(index >= MI_TV_SVHS && index <= MI_TV_TV) {
      gtk_widget_show(sc.om_omenut);
      gtk_widget_show(sc.om_label_tvtype);
      showtvovermenu(sc.currentvbflags);
   } else {
      gtk_widget_hide(sc.om_omenut);
      gtk_widget_hide(sc.om_label_tvtype);
      if(sc.currentvbflags & VB_CHRONTEL) {
         gtk_widget_hide(sc.om_omenuo);
      }
   }
}

static void
settvtypehistory(unsigned long vbflags)
{
    gint index = 0;

    if((sc.sdflags & SiS_SD_ISDUALHEAD) || (!(sc.currentvbflags & VB_VIDEOBRIDGE))) return;

    index = gettvtypeindex(vbflags);
    gtk_option_menu_set_history(GTK_OPTION_MENU(sc.om_omenut), index);

    if(sc.currentvbflags & VB_CHRONTEL) {
       if(gtk_option_menu_get_history(GTK_OPTION_MENU(sc.om_omenuo)) == 3) {
          if((index != 0) || (!(sc.sdflags & SiS_SD_SUPPORTSOVER))) {
	     gtk_option_menu_set_history(GTK_OPTION_MENU(sc.om_omenuo), 2);
	  }
       }
    }
}

/* update crt1 and crt2 pages */
static void
updatecrt1crt2()
{
    setcrt1history();
    hidecrt2typemenu();
    if((!(sc.sdflags & SiS_SD_ISDUALHEAD)) && (sc.currentvbflags & VB_VIDEOBRIDGE)) {
       if(gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(sc.om_check))) {
          showcrt2typemenu(sc.detecteddevices, TRUE);
       } else {
          showcrt2typemenu(sc.detecteddevices, FALSE);
       }
    }
    setcrt2typehistory(sc.currentvbflags);
    showtvtypemenu(sc.currentvbflags);
    settvtypehistory(sc.currentvbflags);
    showtvasmenu(sc.currentvbflags);
    settvashistory(sc.currentvbflags);
    showtvovermenu(sc.currentvbflags);
    settvoverhistory();
    showlcdmenu(1);
    showlcdmenu(2);
    setlcdhistory();
    sc.currentmodeindex = setdmhistory();
#ifdef HAVE_RANDR
    if(sc.haverandr) {
       updateleaveunchanged();
    }
#endif
}

/* Gamma correction */

/* Set gamma-enabled button status */
static void
setgammastatus(void)
{
    if(sc.sdflags & SiS_SD_ISDEPTH8) {
       gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(sc.ga_check1), TRUE);
       gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(sc.ga_check2), TRUE);
    } else {
       gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(sc.ga_check1), (sc.gammaflags & 0x01) ? TRUE : FALSE);
       if(sc.currentvbflags & VB_SISBRIDGE) {
          gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(sc.ga_check2), (sc.gammaflags & 0x02) ? TRUE : FALSE);
       } else {
          gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(sc.ga_check2), FALSE);
       }
    }
}

/* Set current gamma values in spinbuttons */
static void
setgammaspinvalues(void)
{
    gtk_spin_button_set_value(GTK_SPIN_BUTTON(sc.ga_spin1), sc.currentgamma.red);
    gtk_spin_button_set_value(GTK_SPIN_BUTTON(sc.ga_spin2), sc.currentgamma.green);
    gtk_spin_button_set_value(GTK_SPIN_BUTTON(sc.ga_spin3), sc.currentgamma.blue);
#ifdef HAVE_MATH_H
    gtk_spin_button_set_value(GTK_SPIN_BUTTON(sc.ga_spin4), sc.gamma_max[0]);
    gtk_spin_button_set_value(GTK_SPIN_BUTTON(sc.ga_spin5), sc.gamma_max[1]);
    gtk_spin_button_set_value(GTK_SPIN_BUTTON(sc.ga_spin6), sc.gamma_max[2]);
    gtk_spin_button_set_value(GTK_SPIN_BUTTON(sc.ga_spin7), sc.gamma_prescale[0]);
    gtk_spin_button_set_value(GTK_SPIN_BUTTON(sc.ga_spin8), sc.gamma_prescale[1]);
    gtk_spin_button_set_value(GTK_SPIN_BUTTON(sc.ga_spin9), sc.gamma_prescale[2]);
#endif
}

/* Update the gamma curve gfx */
void update_gamma_gfx(void)
{
   if((sc.ga_drawable) && (sc.notebook_page == 3)) {
      gdk_draw_rgb_image(sc.ga_drawable->window, sc.ga_drawable->style->fg_gc[GTK_STATE_NORMAL],
   	                 0, 0, GACW, GACH, GDK_RGB_DITHER_MAX, sc.rgbbuf, GACW * 3);
   }
}

/* TV */

static void
setantiflickerhistory(void)
{
    gint index = 0;
    int val;

    if(sc.tp_type != 1) return;

    if(!(sc.currentvbflags & VB_SISBRIDGE)) {
       gtk_widget_set_sensitive(sc.tp_omenu, FALSE);
       gtk_option_menu_set_history(GTK_OPTION_MENU(sc.tp_omenu), 0);
    } else {
       gtk_widget_set_sensitive(sc.tp_omenu, TRUE);

       if(UnlockSiSDirect(1)) {
          if(XvGetPortAttribute(sc.dpy, sc.xv_port_nr, sc.atom_taf, &val) != Success) {
             fprintf(stderr, xverrorstring);
          }
          UnlockSiSDirect(0);
          if(val == -1) {
            val = 0;
	    gtk_widget_set_sensitive(sc.tp_omenu, FALSE);
          }
          index = val;
      }
      gtk_option_menu_set_history(GTK_OPTION_MENU(sc.tp_omenu), index);
   }
}

static void
setsistvvalue(GtkWidget *widget, Atom myatom)
{
   int val = 0;

   if(sc.tp_type == 1) {

      GtkAdjustment *myadj = gtk_range_get_adjustment(GTK_RANGE(widget));

      if(!(sc.currentvbflags & VB_SISBRIDGE)) {
         gtk_widget_set_sensitive(widget, FALSE);
      } else {
         gtk_widget_set_sensitive(widget, TRUE);

	 if(UnlockSiSDirect(1)) {
            if(XvGetPortAttribute(sc.dpy, sc.xv_port_nr, myatom, &val) != Success) {
               fprintf(stderr, xverrorstring);
            }
            UnlockSiSDirect(0);
            if(val == -1) {
               val = 0;
	       gtk_widget_set_sensitive(widget, FALSE);
            }
         }
         gtk_adjustment_set_value(myadj, (double)val);
         g_signal_emit_by_name(G_OBJECT(myadj), "changed");
      }
   }
}

static void
setsaturationvalue(void)
{
   int val = 0;

   if(sc.tp_type == 1) {

      GtkAdjustment *myadj = gtk_range_get_adjustment(GTK_RANGE(sc.tp_range_sa));

      if(sc.currentvbflags & VB_301) {
         gtk_widget_set_sensitive(sc.tp_range_sa, FALSE);
      } else {
         gtk_widget_set_sensitive(sc.tp_range_sa, TRUE);

         if(UnlockSiSDirect(1)) {
            if(XvGetPortAttribute(sc.dpy, sc.xv_port_nr, sc.atom_tsa, &val) != Success) {
               fprintf(stderr, xverrorstring);
            }
            UnlockSiSDirect(0);
            if(val == -1) {
               val = 0;
               gtk_widget_set_sensitive(sc.tp_range_sa, FALSE);
            }
         }
         gtk_adjustment_set_value(myadj, (double)val);
         g_signal_emit_by_name(G_OBJECT(myadj), "changed");
      }
   }
}

static void
setedgeenhancevalue(void)
{
   int val = 0;

   if(sc.tp_type == 1) {

      GtkAdjustment *myadj = gtk_range_get_adjustment(GTK_RANGE(sc.tp_range_ee));

      if(!(sc.currentvbflags & VB_301)) {
         gtk_widget_set_sensitive(sc.tp_range_ee, FALSE);
      } else {
         gtk_widget_set_sensitive(sc.tp_range_ee, TRUE);

         if(UnlockSiSDirect(1)) {
            if(XvGetPortAttribute(sc.dpy, sc.xv_port_nr, sc.atom_tee, &val) != Success) {
               fprintf(stderr, xverrorstring);
            }
            UnlockSiSDirect(0);
            if(val == -1) {
               val = 0;
               gtk_widget_set_sensitive(sc.tp_range_ee, FALSE);
            }
         }
         gtk_adjustment_set_value(myadj, (double)val);
         g_signal_emit_by_name(G_OBJECT(myadj), "changed");
      }
   }
}

static void
setcolcalibcvalue(void)
{
   setsistvvalue(sc.tp_range_coc, sc.atom_coc);
}

static void
setcolcalibfvalue(void)
{
   setsistvvalue(sc.tp_range_cof, sc.atom_cof);
}

static void
setcfiltervalue(void)
{
   int val = 1;

   if(sc.tp_type != 1) return;

   if(!(sc.currentvbflags & VB_SISBRIDGE)) {
      gtk_widget_set_sensitive(sc.tp_check_cfi, FALSE);
   } else {
      gtk_widget_set_sensitive(sc.tp_check_cfi, TRUE);

      if(UnlockSiSDirect(1)) {
         if(XvGetPortAttribute(sc.dpy, sc.xv_port_nr, sc.atom_cfi, &val) != Success) {
            fprintf(stderr, xverrorstring);
         }
         UnlockSiSDirect(0);
      }

      gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(sc.tp_check_cfi), val ? TRUE : FALSE);
   }
}

static void
setyfilterhistory(void)
{
    gint index = 0;
    int val;

    if(sc.tp_type != 1) return;

    if((!(sc.currentvbflags & VB_SISBRIDGE)) || (sc.currentvbflags & (TV_PALM | TV_PALN))) {
       gtk_widget_set_sensitive(sc.tp_omenuy, FALSE);
       gtk_option_menu_set_history(GTK_OPTION_MENU(sc.tp_omenuy), 1);
    } else {
       gtk_widget_set_sensitive(sc.tp_omenuy, TRUE);

       if(UnlockSiSDirect(1)) {
          if(XvGetPortAttribute(sc.dpy, sc.xv_port_nr, sc.atom_yfi, &val) != Success) {
             fprintf(stderr, xverrorstring);
          }
          UnlockSiSDirect(0);
          index = val;
      }
      gtk_option_menu_set_history(GTK_OPTION_MENU(sc.tp_omenuy), index);
   }
}

static void
setchtvvalue(GtkWidget *widget, Atom myatom)
{
   int val = 0;

   if(sc.tp_type == 2) {

      GtkAdjustment *myadj = gtk_range_get_adjustment(GTK_RANGE(widget));

      if(!(sc.currentvbflags & VB_CHRONTEL)) {
         gtk_widget_set_sensitive(widget, FALSE);
      } else {
         gtk_widget_set_sensitive(widget, TRUE);

	 if(UnlockSiSDirect(1)) {
            if(XvGetPortAttribute(sc.dpy, sc.xv_port_nr, myatom, &val) != Success) {
               fprintf(stderr, xverrorstring);
            }
            UnlockSiSDirect(0);
            if(val == -1) {
               val = 0;
	       gtk_widget_set_sensitive(widget, FALSE);
            }
         }
         gtk_adjustment_set_value(myadj, (double)val);
         g_signal_emit_by_name(G_OBJECT(myadj), "changed");

      }
   }
}

static void
setcontrastvalue(void)
{
   setchtvvalue(sc.tp_range_co, sc.atom_tco);
}

static void
settextenhancevalue(void)
{
   setchtvvalue(sc.tp_range_te, sc.atom_tte);
}

static void
setchromaffvalue(void)
{
   setchtvvalue(sc.tp_range_cf, sc.atom_tcf);
}

static void
setlumaffvalue(void)
{
   setchtvvalue(sc.tp_range_lf, sc.atom_tlf);
}

static void
setcvbscolorvalue(void)
{
    int val = 1;

    if(sc.tp_type != 2) return;

    if(!(sc.currentvbflags & VB_CHRONTEL)) {
      gtk_widget_set_sensitive(sc.tp_check_col, FALSE);
   } else {
      gtk_widget_set_sensitive(sc.tp_check_col, TRUE);

      val = getxvval(sc.atom_ccc);
      gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(sc.tp_check_col), val ? TRUE : FALSE);
   }
}

/* Update the entire TV page */
static void
updatealltv(void)
{
   setantiflickerhistory();
   setsaturationvalue();
   setedgeenhancevalue();
   setcolcalibcvalue();
   setcolcalibfvalue();
   setcfiltervalue();
   setyfilterhistory();
   setcontrastvalue();
   settextenhancevalue();
   setchromaffvalue();
   setlumaffvalue();
   setcvbscolorvalue();
}

/* Xv */

#ifdef USEXV
static void
makexvwindow(void)
{
   char *ptr;
   int i, j;

   if(sc.noxvdemo) return;

   if(sc.xvwin) return;

   sc.xvwin = XCreateSimpleWindow (sc.dpy,
				   GDK_WINDOW_XWINDOW(sc.window->window),
				   0, 0,
				   1, 1,
				   1, 0, 0);
   XMapWindow(sc.dpy,sc.xvwin);
   XFlush(sc.dpy);
   sc.xvraised = FALSE;

   if(!sc.xvcontext) {
      sc.xvcontext = XCreateGC(sc.dpy, sc.xvwin, 0, 0);
   }
   if(!sc.xvcontext) return;

   if(!sc.myxvimage) {
      sc.myxvimage = XvShmCreateImage(sc.dpy, sc.xv_port_nr,
             		0x59565955, 0, XVWIDTH, XVHEIGHT,
			&sc.xvshminfo);
   }
   if(!sc.myxvimage) {
      fprintf(stderr, "XvShmCreateImage failed\n");
      return;
   }
   sc.xvshminfo.shmid = shmget(IPC_PRIVATE, sc.myxvimage->data_size, IPC_CREAT|0777);
   sc.xvshminfo.shmaddr = sc.myxvimage->data = shmat(sc.xvshminfo.shmid, 0, 0);
   sc.xvshminfo.readOnly = False;
   sc.xvattached = XShmAttach(sc.dpy, &sc.xvshminfo);
   if(sc.xvattached) {
      ptr = sc.myxvimage->data;
      if(!ptr) {
         fprintf(stderr, "XvImage->data is NULL\n");
         return;
      }
      for(i = 0; i < XVHEIGHT; i++, ptr += sc.myxvimage->pitches[0]) {
         for(j = 0; j < XVWIDTH; j++) {
	    ptr[(j << 1)] = xviconbuffer[(i * XVWIDTH * 2) + (j*2)];
	    ptr[(j << 1) + 1] = xviconbuffer[(i * XVWIDTH * 2) + (j*2) + 1];
         }
      }
   } else {
      fprintf(stderr, "XShmAttach failed\n");
   }
}

static void
destroyxvimage(void)
{
   if(sc.xvshown) {
      XvStopVideo(sc.dpy, sc.xv_port_nr, sc.xvwin);
      sc.xvshown = FALSE;
   }
   if(sc.xvattached) {
      XShmDetach(sc.dpy, &sc.xvshminfo);
      shmdt(sc.xvshminfo.shmaddr);
      shmctl(sc.xvshminfo.shmid, IPC_RMID, 0);
      XFree(sc.myxvimage);
      sc.myxvimage = NULL;
      sc.xvattached = FALSE;
   }
   if(sc.xvcontext) {
      XFreeGC(sc.dpy, sc.xvcontext);
      sc.xvcontext = NULL;
   }
   if(sc.xvwin) {
      XDestroyWindow(sc.dpy, sc.xvwin);
      sc.xvwin = 0;
   }
}

static void
resizexvwindow(void)
{
   XWindowAttributes wa;

   if(!sc.xvinit) return;
   if(!sc.xvwin) return;

   if(!GTK_WIDGET_REALIZED(sc.vi_layout)) {
      sc.xvx = sc.xvy = 0;
      sc.xvwidth = sc.xvheight = 1;
      return;
   }

   sc.xvx = sc.vi_layout->allocation.x;
   sc.xvy = sc.vi_layout->allocation.y;
   sc.xvwidth = sc.vi_layout->allocation.width;
   sc.xvheight = sc.vi_layout->allocation.height;

   XGetWindowAttributes(sc.dpy, sc.xvwin, &wa);

   if( (sc.xvx != wa.x) ||
       (sc.xvy != wa.y) ||
       (sc.xvwidth != wa.width) ||
       (sc.xvheight != wa.height) ) {
      XMoveResizeWindow(sc.dpy, sc.xvwin,
   		sc.vi_layout->allocation.x,
		sc.vi_layout->allocation.y,
		sc.vi_layout->allocation.width,
   		sc.vi_layout->allocation.height);
   }
}

static void
hidexvimage(void)
{
   if(!sc.xvinit) return;
   if(sc.xvshown) {
      XvStopVideo(sc.dpy, sc.xv_port_nr, sc.xvwin);
      XSync(sc.dpy, False);
      sc.xvshown = FALSE;
   }
   if(sc.xvraised) {
      XUnmapWindow(sc.dpy,sc.xvwin);
      XFlush(sc.dpy);
      sc.xvraised = FALSE;
   }
}

static void
showxvimage(void)
{
   sc.xvshown = FALSE;

   if(!sc.xvinit) return;

   if(!sc.xvwin) return;
   if(!sc.xvattached) return;
   if((sc.xvwidth == 0) || (sc.xvheight == 0)) return;

   if(XvGrabPort(sc.dpy, sc.xv_port_nr, CurrentTime) == Success) {
      XvUngrabPort(sc.dpy, sc.xv_port_nr, CurrentTime);
   } else {
      fprintf(stderr, "Xv busy\n");
      hidexvimage();
      sc.xvwasbusy = TRUE;
      return;
   }

   if((!sc.xvraised) || (sc.xvwasbusy)) {
      XMapWindow(sc.dpy,sc.xvwin);
      XRaiseWindow(sc.dpy, sc.xvwin);
      XFlush(sc.dpy);
      resizexvwindow();
      XSync(sc.dpy, False);
      sc.xvraised = TRUE;
      sc.xvwasbusy = FALSE;
   }

   XvShmPutImage(sc.dpy, sc.xv_port_nr,
                 sc.xvwin,
                 sc.xvcontext,
		 sc.myxvimage,
		 0, 0,
		 XVWIDTH, XVHEIGHT,
		 0, 0,
		 sc.xvwidth, sc.xvheight,
		 False);

   XSync(sc.dpy, False);

   sc.xvshown = TRUE;
}
#endif

static void
setvideorangevalue(GtkWidget *myrange, Atom viatom)
{
   GtkAdjustment *myadj = gtk_range_get_adjustment(GTK_RANGE(myrange));
   int val = getxvval(viatom);

   gtk_adjustment_set_value(myadj, (double)val);
   g_signal_emit_by_name(G_OBJECT(myadj), "changed");
}

static void
setvicontrastvalue(void)
{
   setvideorangevalue(sc.vi_range_co, sc.atom_con);
}

static void
setvibrightnessvalue(void)
{
   setvideorangevalue(sc.vi_range_br, sc.atom_bri);
}

static void
setvisaturationvalue(void)
{
   setvideorangevalue(sc.vi_range_sa, sc.atom_sat);
}

static void
setvihuevalue(void)
{
   setvideorangevalue(sc.vi_range_hu, sc.atom_hue);
}

static void
setxvgammastatus(void)
{
    if(!(sc.sdflags & SiS_SD_SUPPORTXVGAMMA1)) return;

    gtk_widget_set_sensitive(sc.vi_check_gamma,
    	((sc.gammaflags & 0x01) && (!(sc.sdflags & SiS_SD_ISDEPTH8))) ? TRUE : FALSE);
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(sc.vi_check_gamma),
    	((sc.gammaflags & 0x04) && (!(sc.sdflags & SiS_SD_ISDEPTH8))) ? TRUE : FALSE);
    setxvgammasensitive();
}

static void
setxvgammaspinvalues(void)
{
    gtk_spin_button_set_value(GTK_SPIN_BUTTON(sc.vi_spin1), sc.xvgammared);
    gtk_spin_button_set_value(GTK_SPIN_BUTTON(sc.vi_spin2), sc.xvgammagreen);
    gtk_spin_button_set_value(GTK_SPIN_BUTTON(sc.vi_spin3), sc.xvgammablue);
}

static void
setviswitchcrt(void)
{
   int val = getxvval(sc.atom_swc);

   gtk_toggle_button_set_active((val ? GTK_TOGGLE_BUTTON(sc.vi_radio_crt2) :
   			   	       GTK_TOGGLE_BUTTON(sc.vi_radio_crt1)), TRUE);
}

/*******************************************/
/*            GTK Signal handlers          */
/*******************************************/

/* CRT1 page */

void c1_menu_changed(GtkWidget *widget, gpointer data)
{
    updatecrt1gfx();
    showlcdmenu(1);
}

void c1_omenuls_changed(GtkWidget *widget, gpointer data)
{
    gint index = gtk_option_menu_get_history(GTK_OPTION_MENU(sc.c1_omenuls));

    if(!(sc.sdflags & SiS_SD_SUPPORTCENTER)) return;

    if(index == 1) {
       gtk_widget_hide(sc.c1_omenulcc);
    } else {
       gtk_widget_show(sc.c1_omenulcc);
    }
}

/* CRT2 page */

void om_check_toggled(GtkWidget *widget, gpointer data)
{
    updatecrt1crt2();
}

void om_menu_changed(GtkWidget *widget, gpointer data)
{
   showtvtypemenu(sc.currentvbflags);
   showtvasmenu(sc.currentvbflags);
   showlcdmenu(2);
   updatecrt2gfx();
}

void om_menut_changed(GtkWidget *widget, gpointer data)
{
   showtvovermenu(sc.currentvbflags);
}

void om_omenuls_changed(GtkWidget *widget, gpointer data)
{
    gint index = gtk_option_menu_get_history(GTK_OPTION_MENU(sc.om_omenuls));

    if(!(sc.sdflags & SiS_SD_SUPPORTCENTER)) return;

    if(index == 1) {
       gtk_widget_hide(sc.om_omenulcc);
    } else {
       gtk_widget_show(sc.om_omenulcc);
    }
}

/* Gamma page */

/* CRT1-gamma toggled */
void ga_check1_toggled(GtkWidget *widget, gpointer data)
{
    endisablegamma1(gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(sc.ga_check1)) ? TRUE : FALSE);
}

/* CRT2-gamma toggled */
void ga_check2_toggled(GtkWidget *widget, gpointer data)
{
    endisablegamma2(gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(sc.ga_check2)) ? TRUE : FALSE);
}

void ga_spin1_changed(GtkWidget *widget, gpointer data)
{
   memcpy(&sc.newgamma, &sc.currentgamma, sizeof(XF86VidModeGamma));
   sc.newgamma.red = gtk_spin_button_get_value(GTK_SPIN_BUTTON(sc.ga_spin1));
   if(sc.newgamma.red >= 0.1 && sc.newgamma.red <= 10.0) setgamma();
}

void ga_spin2_changed(GtkWidget *widget, gpointer data)
{
   memcpy(&sc.newgamma, &sc.currentgamma, sizeof(XF86VidModeGamma));
   sc.newgamma.green = gtk_spin_button_get_value(GTK_SPIN_BUTTON(sc.ga_spin2));
   if(sc.newgamma.green >= 0.1 && sc.newgamma.green <= 10.0) setgamma();
}

void ga_spin3_changed(GtkWidget *widget, gpointer data)
{
   memcpy(&sc.newgamma, &sc.currentgamma, sizeof(XF86VidModeGamma));
   sc.newgamma.blue = gtk_spin_button_get_value(GTK_SPIN_BUTTON(sc.ga_spin3));
   if(sc.newgamma.blue >= 0.1 && sc.newgamma.blue <= 10.0) setgamma();
}

#ifdef HAVE_MATH_H
void ga_spin4_changed(GtkWidget *widget, gpointer data)
{
   sc.gamma_max[0] = gtk_spin_button_get_value(GTK_SPIN_BUTTON(sc.ga_spin4));
   setgamma();
}

void ga_spin5_changed(GtkWidget *widget, gpointer data)
{
   sc.gamma_max[1] = gtk_spin_button_get_value(GTK_SPIN_BUTTON(sc.ga_spin5));
   setgamma();
}

void ga_spin6_changed(GtkWidget *widget, gpointer data)
{
   sc.gamma_max[2] = gtk_spin_button_get_value(GTK_SPIN_BUTTON(sc.ga_spin6));
   setgamma();
}
void ga_spin7_changed(GtkWidget *widget, gpointer data)
{
   sc.gamma_prescale[0] = gtk_spin_button_get_value(GTK_SPIN_BUTTON(sc.ga_spin7));
   setgamma();
}

void ga_spin8_changed(GtkWidget *widget, gpointer data)
{
   sc.gamma_prescale[1] = gtk_spin_button_get_value(GTK_SPIN_BUTTON(sc.ga_spin8));
   setgamma();
}

void ga_spin9_changed(GtkWidget *widget, gpointer data)
{
   sc.gamma_prescale[2] = gtk_spin_button_get_value(GTK_SPIN_BUTTON(sc.ga_spin9));
   setgamma();
}

gboolean ga_picture_exposed(GtkWidget *widget, GdkEventExpose *event, gpointer data)
{
   gdk_draw_rgb_image(widget->window, widget->style->fg_gc[GTK_STATE_NORMAL],
   	0, 0, GACW, GACH, GDK_RGB_DITHER_MAX, sc.rgbbuf, GACW * 3);
}
#endif

/* TV page */

void tv_de_clicked(GtkWidget *widget, gpointer data)
{
   tvdefaultposscale();
}

void tv_up_clicked(GtkWidget *widget, gpointer data)
{
   dotvposscale(sc.atom_tvy, FALSE, -32);
}

void tv_dn_clicked(GtkWidget *widget, gpointer data)
{
   dotvposscale(sc.atom_tvy, TRUE, 32);
}

void tv_le_clicked(GtkWidget *widget, gpointer data)
{
   dotvposscale(sc.atom_tvx, FALSE, -32);
}

void tv_ri_clicked(GtkWidget *widget, gpointer data)
{
   dotvposscale(sc.atom_tvx, TRUE, 32);
}

void tv_xm_clicked(GtkWidget *widget, gpointer data)
{
   dotvposscale(sc.atom_txs, FALSE, -16);
}

void tv_xp_clicked(GtkWidget *widget, gpointer data)
{
   dotvposscale(sc.atom_txs, TRUE, 16);
}

void tv_ym_clicked(GtkWidget *widget, gpointer data)
{
   dotvposscale(sc.atom_tys, FALSE, -4);
}

void tv_yp_clicked(GtkWidget *widget, gpointer data)
{
   dotvposscale(sc.atom_tys, TRUE, 3);
}

void tp_af_changed(GtkWidget *widget, gpointer data)
{
   gint index = gtk_option_menu_get_history(GTK_OPTION_MENU(sc.tp_omenu));

   if(index >= 0 && index <= 4) {
      if(UnlockSiSDirect(1)) {
         if(XvSetPortAttribute(sc.dpy, sc.xv_port_nr, sc.atom_taf, index) != Success) {
            fprintf(stderr, xverrorstring);
         }
         XFlush(sc.dpy);
         UnlockSiSDirect(0);
      }
   }
}

void some_range_changed(GtkWidget *widget, gpointer data, Atom myatom, gdouble uplimit, gdouble downlimit)
{
   GtkAdjustment *myadj = gtk_range_get_adjustment(GTK_RANGE(widget));
   gdouble myval = gtk_adjustment_get_value(myadj);

   if((myval >= downlimit) && (myval <= uplimit)) {
      if(UnlockSiSDirect(1)) {
         if(XvSetPortAttribute(sc.dpy, sc.xv_port_nr, myatom, (int)myval) != Success) {
            fprintf(stderr, xverrorstring);
         }
         XFlush(sc.dpy);
         UnlockSiSDirect(0);
      }
   }
}

void tp_sa_changed(GtkWidget *widget, gpointer data)
{
   some_range_changed(sc.tp_range_sa, data, sc.atom_tsa, 15.0, 0.0);
}

void tp_ee_changed(GtkWidget *widget, gpointer data)
{
   some_range_changed(sc.tp_range_ee, data, sc.atom_tee, 15.0, 0.0);
}

void tp_coc_changed(GtkWidget *widget, gpointer data)
{
   some_range_changed(sc.tp_range_coc, data, sc.atom_coc, 120.0, -120.0);
}

void tp_cof_changed(GtkWidget *widget, gpointer data)
{
   some_range_changed(sc.tp_range_cof, data, sc.atom_cof, 127.0, -128.0);
}

void tp_check_cfi_toggled(GtkWidget *widget, gpointer data)
{
    int myval = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget)) ? 1 : 0;

    if(UnlockSiSDirect(1)) {
       if(XvSetPortAttribute(sc.dpy, sc.xv_port_nr, sc.atom_cfi, (int)myval) != Success) {
         fprintf(stderr, xverrorstring);
       }
       XFlush(sc.dpy);
       UnlockSiSDirect(0);
    }
}

void tp_yfi_changed(GtkWidget *widget, gpointer data)
{
   gint index = gtk_option_menu_get_history(GTK_OPTION_MENU(sc.tp_omenuy));

   if(index >= 0 && index <= 8) {
      if(UnlockSiSDirect(1)) {
         if(XvSetPortAttribute(sc.dpy, sc.xv_port_nr, sc.atom_yfi, index) != Success) {
            fprintf(stderr, xverrorstring);
         }
         XFlush(sc.dpy);
         UnlockSiSDirect(0);
      }
   }
}

void tp_co_changed(GtkWidget *widget, gpointer data)
{
   some_range_changed(sc.tp_range_co, data, sc.atom_tco, 15.0, 0.0);
}

void tp_te_changed(GtkWidget *widget, gpointer data)
{
   some_range_changed(sc.tp_range_te, data, sc.atom_tte, 15.0, 0.0);
}

void tp_cf_changed(GtkWidget *widget, gpointer data)
{
   some_range_changed(sc.tp_range_cf, data, sc.atom_tcf, 15.0, 0.0);
}

void tp_lf_changed(GtkWidget *widget, gpointer data)
{
   some_range_changed(sc.tp_range_lf, data, sc.atom_tlf, 15.0, 0.0);
}

void tp_check_col_toggled(GtkWidget *widget, gpointer data)
{
    int myval = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget)) ? 1 : 0;

    if(UnlockSiSDirect(1)) {
       if(XvSetPortAttribute(sc.dpy, sc.xv_port_nr, sc.atom_ccc, (int)myval) != Success) {
         fprintf(stderr, xverrorstring);
       }
       XFlush(sc.dpy);
       UnlockSiSDirect(0);
    }
}

/* Xv page */

void vi_co_changed(GtkWidget *widget, gpointer data)
{
   some_range_changed(sc.vi_range_co, data, sc.atom_con, 7.0, 0.0);
#ifdef USEXV
   if(!sc.noxvdemo) {
      showxvimage();
   }
#endif
}

void vi_br_changed(GtkWidget *widget, gpointer data)
{
   some_range_changed(sc.vi_range_br, data, sc.atom_bri, 127.0, -128.0);
#ifdef USEXV
   if(!sc.noxvdemo) {
      showxvimage();
   }
#endif
}

void vi_hu_changed(GtkWidget *widget, gpointer data)
{
   some_range_changed(sc.vi_range_hu, data, sc.atom_hue, 7.0, -8.0);
#ifdef USEXV
   if(!sc.noxvdemo) {
      showxvimage();
   }
#endif
}

void vi_sa_changed(GtkWidget *widget, gpointer data)
{
   some_range_changed(sc.vi_range_sa, data, sc.atom_sat, 7.0, -7.0);
#ifdef USEXV
   if(!sc.noxvdemo) {
      showxvimage();
   }
#endif
}

void vi_radio_changed(GtkWidget *widget, gpointer data)
{
   if(XvSetPortAttribute(sc.dpy, sc.xv_port_nr, sc.atom_swc,
              (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget)) ? 0 : 1)) != Success) {
         fprintf(stderr, xverrorstring);
   }
   setviswitchcrt();
#ifdef USEXV
   if(!sc.noxvdemo) {
      showxvimage();
   }
#endif
}

void vi_check_gamma_toggled(GtkWidget *widget, gpointer data)
{
   endisablexvgamma(gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(sc.vi_check_gamma)) ? TRUE : FALSE);
#ifdef USEXV
   if(!sc.noxvdemo) {
      showxvimage();
   }
#endif
}

void vi_spin1_changed(GtkWidget *widget, gpointer data)
{
   sc.xvgammared = gtk_spin_button_get_value(GTK_SPIN_BUTTON(sc.vi_spin1));
   if(sc.xvgammared >= 0.1 && sc.xvgammared <= 10.0) setxvgammared();
}

void vi_spin2_changed(GtkWidget *widget, gpointer data)
{
   sc.xvgammagreen = gtk_spin_button_get_value(GTK_SPIN_BUTTON(sc.vi_spin2));
   if(sc.xvgammagreen >= 0.1 && sc.xvgammagreen <= 10.0) setxvgammagreen();
}

void vi_spin3_changed(GtkWidget *widget, gpointer data)
{
   sc.xvgammablue = gtk_spin_button_get_value(GTK_SPIN_BUTTON(sc.vi_spin3));
   if(sc.xvgammablue >= 0.1 && sc.xvgammablue <= 10.0) setxvgammablue();
}

static void
setxvgammasensitive(void)
{
   Bool enable = ((sc.gammaflags & 0x05) == 0x05) ? TRUE : FALSE;

   if(!(sc.sdflags & SiS_SD_SUPPORTXVGAMMA1)) return;

   if(sc.sdflags & SiS_SD_ISDEPTH8) enable = FALSE;

   gtk_widget_set_sensitive(sc.vi_spin1, enable);
   gtk_widget_set_sensitive(sc.vi_spin2, enable);
   gtk_widget_set_sensitive(sc.vi_spin3, enable);
}

#ifdef USEXV
gboolean vi_layout_expose(GtkWidget *widget, GdkEventExpose *event, gpointer data)
{
   /* fprintf(stderr, "layout expose\n"); */
   if(sc.noxvdemo) return;
   if(sc.notebook_page != 5) return TRUE;
   if(!sc.xvwin) return TRUE;
   if(!sc.xvinit) return TRUE;
   resizexvwindow();
   showxvimage();
   return TRUE;
}

gboolean vi_layout_configure(GtkWidget *widget, GdkEventExpose *event, gpointer data)
{
   /* fprintf(stderr, "layout config\n"); */
   if(sc.noxvdemo) return;
   if(sc.notebook_page != 5) return;
   if(!sc.xvwin) return;
   if(!sc.xvinit) return TRUE;
   resizexvwindow();
   showxvimage();
   return TRUE;
}

gboolean window_configure(GtkWidget *widget, GdkEventExpose *event, gpointer data)
{
   /* fprintf(stderr, "window config\n"); */
   if(sc.noxvdemo) return;
   if(sc.notebook_page != 5) return FALSE;
   if(!sc.xvwin) return FALSE;
   if(!sc.xvinit) return TRUE;
   resizexvwindow();
   showxvimage();
   return FALSE;
}

gboolean window_expose(GtkWidget *widget, GdkEventExpose *event, gpointer data)
{
   /* fprintf(stderr, "window expose\n"); */
   if(sc.noxvdemo) return;
   if(sc.notebook_page != 5) return FALSE;
   if(!sc.xvwin) return FALSE;
   if(!sc.xvinit) return TRUE;
   resizexvwindow();
   showxvimage();
   return FALSE;
}

gboolean window_visibility(GtkWidget *widget, GdkEventExpose *event, gpointer data)
{
   /* fprintf(stderr, "window visibility\n"); */
   if(sc.noxvdemo) return;
   if(sc.notebook_page != 5) return FALSE;
   if(!sc.xvwin) return FALSE;
   if(!sc.xvinit) return TRUE;
   hidexvimage();
   resizexvwindow();
   showxvimage();
   return FALSE;
}
#endif

/* Current page */

static void
cf_insert_option(char *oname, char *ovalue, GtkTextTag *tag)
{
   GtkTextIter myiter, start, end;
   char options[] = "\tOption \"";
   int curlen;

   curlen = gtk_text_buffer_get_char_count(sc.cf_buffer);

   gtk_text_buffer_get_end_iter(sc.cf_buffer, &myiter);
   gtk_text_buffer_insert(sc.cf_buffer, &myiter, options, -1);
   gtk_text_buffer_get_end_iter(sc.cf_buffer, &myiter);
   gtk_text_buffer_insert(sc.cf_buffer, &myiter, oname, -1);
   gtk_text_buffer_get_end_iter(sc.cf_buffer, &myiter);
   gtk_text_buffer_insert(sc.cf_buffer, &myiter, "\" \"", -1);
   gtk_text_buffer_get_end_iter(sc.cf_buffer, &myiter);
   gtk_text_buffer_insert(sc.cf_buffer, &myiter, ovalue, -1);
   gtk_text_buffer_get_end_iter(sc.cf_buffer, &myiter);
   gtk_text_buffer_insert(sc.cf_buffer, &myiter, "\"\n", -1);

   gtk_text_buffer_get_iter_at_offset(sc.cf_buffer, &start, curlen);
   gtk_text_buffer_get_end_iter(sc.cf_buffer, &end);
   if(tag) gtk_text_buffer_apply_tag(sc.cf_buffer, tag, &start, &end);
}

static void
cf_append_text(char *mytext, GtkTextTag *tag)
{
   GtkTextIter myiter, start, end;
   int curlen;

   curlen = gtk_text_buffer_get_char_count(sc.cf_buffer);
   gtk_text_buffer_get_end_iter(sc.cf_buffer, &myiter);
   gtk_text_buffer_insert(sc.cf_buffer, &myiter, mytext, -1);
   gtk_text_buffer_get_iter_at_offset(sc.cf_buffer, &start, curlen);
   gtk_text_buffer_get_end_iter(sc.cf_buffer, &end);
   if(tag) gtk_text_buffer_apply_tag(sc.cf_buffer, tag, &start, &end);
}

static void
appendcrt1gamma(void)
{
   if(!(sc.gammaflags & 0x01)) {
      cf_append_text("\n\t# [sisctrl] Enable/disable gamma correction for CRT1\n", sc.red);
      cf_insert_option("CRT1Gamma", "false", sc.red);
   }
}
static void
appendbrightness(void)
{
   char numbuffer[64];
   int i;

   cf_append_text("\n\t# [sisctrl] Gamma correction: Brightness values\n", sc.red);
   cf_append_text("\t# The driver does NOT write these to the hardware. Instead, for\n", sc.grey);
   cf_append_text("\t# using gamma brightness, you need to execute sisctrl with the\n", sc.grey);
   cf_append_text("\t# -setgammabrightness (or -sg) switch at server start. It is\n", sc.grey);
   cf_append_text("\t# most convenient to do this by adding this to your ~/.xsession\n", sc.grey);
   cf_append_text("\t# or ~/.xinitrc file:\n", sc.grey);
   cf_append_text("\t#    sisctrl -sg\n", sc.grey);
   cf_append_text("\t# If you run Xinerama, do that twice for both screens. Add this:\n", sc.grey);
   cf_append_text("\t#    sisctrl -screen 1 -sg\n", sc.grey);
   sprintf(numbuffer, "%.3f %.3f %.3f",
	   (float)sc.gamma_max[0], (float)sc.gamma_max[1], (float)sc.gamma_max[2]);
   for(i=0; i<strlen(numbuffer); i++) {
      if(numbuffer[i] == ',') numbuffer[i] = '.';
   }
   cf_insert_option("StoredGammaBrightness", numbuffer, sc.red);
   sprintf(numbuffer, "%.3f %.3f %.3f",
       (float)sc.gamma_prescale[0], (float)sc.gamma_prescale[1], (float)sc.gamma_prescale[2]);
   for(i=0; i<strlen(numbuffer); i++) {
      if(numbuffer[i] == ',') numbuffer[i] = '.';
   }
   cf_insert_option("StoredGammaPreBrightness", numbuffer, sc.red);
}

static void
appendxvstuff(void)
{
   char numbuffer[64];
   int i;

   sprintf(numbuffer, "%d", getxvval(sc.atom_con));
   cf_insert_option("XvDefaultContrast", numbuffer, sc.red);
   sprintf(numbuffer, "%d", getxvval(sc.atom_bri));
   cf_insert_option("XvDefaultBrightness", numbuffer, sc.red);
   if(sc.sdflags & (SiS_SD_IS315SERIES | SiS_SD_IS330SERIES)) {
      sprintf(numbuffer, "%d", getxvval(sc.atom_hue));
      cf_insert_option("XvDefaultHue", numbuffer, sc.red);
      sprintf(numbuffer, "%d", getxvval(sc.atom_sat));
      cf_insert_option("XvDefaultSaturation", numbuffer, sc.red);
   }
   if(sc.sdflags & SiS_SD_SUPPORTXVGAMMA1) {
      if(!(sc.gammaflags & 0x04)) {
	 cf_insert_option("XvGamma", "false", sc.red);
      } else {
	 sprintf(numbuffer, "%.3f %.3f %.3f",
	     (float)sc.xvgammared,(float)sc.xvgammagreen,(float)sc.xvgammablue);
	 for(i=0; i<strlen(numbuffer); i++) {
	    if(numbuffer[i] == ',') numbuffer[i] = '.';
         }
         cf_insert_option("XvGamma", numbuffer, sc.red);
      }
   }
}

void cf_generate_current(void)
{
   GtkTextIter myiter;
   int val, i;
   char numbuffer[64];
   char forcecrt2type[] = "ForceCRT2Type";
   char chtvoverscan[] = "CHTVOverscan";
   char chtvsoverscan[] = "CHTVSuperOverscan";
   char sistvantiflicker[] = "SISTVAntiFlicker";
   char header[] = "For the current configuration - except the mode and desktop size settings - "
   		   "to be permanent, the \"Monitor\" "
   		   "and \"Device\" sections of your XF86Config(-4) should contain at least "
		   "the following: (Note that this fragment does not suffice to fully configure "
		   "dual head or MergedFB mode, nor any other special setup. The comments "
		   "for dual head / Xinerama mode assume that mode on a single SiS card "
		   "and do not generally apply if you use two separate graphics cards. "
		   "Remember that dual head / Xinerama and MergedFB mode are mutually "
		   "exclusive. For detailed documentation, see "
		   "http://www.winischhofer.net/linuxsisvga.shtml)\n\n";

   gtk_text_buffer_set_text(sc.cf_buffer, header, -1);

   cf_append_text("black = Required keywords; section delimiters\n", sc.black);
   cf_append_text("grey", sc.grey);
   cf_append_text(" = Generic settings; configuration hints\n", sc.black);
   cf_append_text("red", sc.red);
   cf_append_text(" = Settings changed/added by sisctrl\n", sc.black);
   cf_append_text("Lines starting with \"#\" are comments and ignored by XFree86\n\n", sc.black);

   cf_append_text("Section \"Monitor\"\n", sc.black);
   cf_append_text("\t# Identifier, VendorName, ModelName,\n", sc.grey);
   cf_append_text("\t# VertRefresh, HorizSync, DPMS, ...\n", sc.grey);
   if( (!(sc.sdflags & SiS_SD_ISDUALHEAD)) ||
       ((!(sc.sdflags & SiS_SD_ISDHSECONDHEAD)) && (!(sc.sdflags & SiS_SD_ISDHXINERAMA))) ||
       ((sc.sdflags & SiS_SD_ISDHXINERAMA) && (sc.myscreen == DefaultScreen(sc.dpy))) ) {
      cf_append_text("\t# [sisctrl] Gamma correction\n", sc.red);
      sprintf(numbuffer, "\tGamma %.3f %.3f %.3f\n", sc.currentgamma.red, sc.currentgamma.green,
      					sc.currentgamma.blue);
      for(i=0; i<strlen(numbuffer); i++) {
	  if(numbuffer[i] == ',') numbuffer[i] = '.';
      }
      cf_append_text(numbuffer, sc.red);
   }  else {
      cf_append_text("\t# Gamma, ...\n", sc.grey);
   }
   cf_append_text("EndSection\n\n", sc.black);

   cf_append_text("Section \"Device\"\n", sc.black);
   cf_append_text("\t# Replace IDENTIFIER with the real identifier from the \"Device\"\n", sc.grey);
   cf_append_text("\t# tag in your \"Screen\" section.\n", sc.grey);
   cf_append_text("\tIdentifier \"IDENTIFIER\"\n", sc.black);

   cf_append_text("\tVendorName \"SiS\"\n", sc.grey);
   sprintf(numbuffer, "\tBoardName \"%s\"\n", getchiptype());
   cf_append_text(numbuffer, sc.grey);

   cf_append_text("\tDriver \"sis\"\n", sc.black);

   cf_append_text("\n\t# \"BusID\" is mandatory for dual head / Xinerama mode.\n", sc.grey);
   sprintf(numbuffer, "\tBusID \"PCI:%d:%d:%d\"\n", (sc.busid >> 16), (sc.busid >> 8) & 0xff, sc.busid & 0xff);
   cf_append_text(numbuffer, sc.grey);

   cf_append_text("\n\t# \"Screen\" is mandatory for dual head / Xinerama mode.\n", sc.grey);
   cf_append_text("\t# Usually, \"Screen 0\" is CRT2 (master), \"Screen 1\" is CRT1\n", sc.grey);
   cf_append_text("\t# (slave)\n", sc.grey);
   cf_append_text("\tScreen 0\n", sc.grey);

   cf_append_text("\n\t# EnableSiSCtrl must be set to use SiSCtrl\n", sc.grey);
   cf_append_text("\tOption \"EnableSiSCtrl\" \"true\"\n", sc.grey);

   if( (!(sc.sdflags & SiS_SD_ISDUALHEAD)) ||
       ((!(sc.sdflags & SiS_SD_ISDHSECONDHEAD)) && (!(sc.sdflags & SiS_SD_ISDHXINERAMA))) ||
       ((sc.sdflags & SiS_SD_ISDHXINERAMA) && (sc.myscreen == DefaultScreen(sc.dpy))) ) {
      appendcrt1gamma();
   }

   cf_append_text("\n\t# [sisctrl] Set CRT1 status (Note: overrides auto-detection)\n", sc.red);
   cf_insert_option("ForceCRT1", (sc.CRT1isoff ? "false" : "true"), sc.red);

   if(sc.sdflags & SiS_SD_SUPPORTLCDA) {
      cf_append_text("\n\t# [sisctrl] Set CRT1 type\n", sc.red);
      if(sc.currentvbflags & CRT1_LCDA) {
         cf_insert_option("ForceCRT1Type", "LCD", sc.red);
      } else {
         cf_insert_option("ForceCRT1Type", "VGA", sc.red);
      }
   }

   if(sc.currentvbflags & VB_VIDEOBRIDGE) {
      cf_append_text("\n\t# [sisctrl] Set CRT2 type (Note: overrides auto-detection)\n", sc.red);
      switch(sc.currentvbflags & CRT2_ENABLE) {
      case CRT2_LCD:
         cf_insert_option(forcecrt2type, "LCD", sc.red);
	 break;
      case CRT2_VGA:
         cf_insert_option(forcecrt2type, "VGA", sc.red);
	 break;
      case CRT2_TV:
         if(sc.currentvbflags & VB_SISBRIDGE) {
	    if((sc.currentvbflags & (TV_SVIDEO | TV_AVIDEO)) == (TV_SVIDEO | TV_AVIDEO))
	       cf_insert_option(forcecrt2type, "SVIDEO+COMPOSITE", sc.red);
	    else if(sc.currentvbflags & TV_SVIDEO)
	       cf_insert_option(forcecrt2type, "SVIDEO", sc.red);
	    else if(sc.currentvbflags & TV_AVIDEO)
	       cf_insert_option(forcecrt2type, "COMPOSITE", sc.red);
	    else if(sc.currentvbflags & TV_SCART)
	       cf_insert_option(forcecrt2type, "SCART", sc.red);
	    else if(sc.currentvbflags & TV_HIVISION)
	       cf_insert_option(forcecrt2type, "HIVISION", sc.red);
	    else if(sc.currentvbflags & TV_YPBPR) {
	       if(sc.currentvbflags & TV_YPBPR525I)
	          cf_insert_option(forcecrt2type, "YPBPR480I", sc.red);
	       else if(sc.currentvbflags & TV_YPBPR525P)
	          cf_insert_option(forcecrt2type, "YPBPR480P", sc.red);
	       else if(sc.currentvbflags & TV_YPBPR750P)
	          cf_insert_option(forcecrt2type, "YPBPR720P", sc.red);
	       else if(sc.currentvbflags & TV_YPBPR1080I)
	          cf_insert_option(forcecrt2type, "YPBPR1080I", sc.red);
	       else
	          cf_insert_option(forcecrt2type, "TV", sc.red);
	    } else
	       cf_insert_option(forcecrt2type, "TV", sc.red);
	 } else {
	    cf_insert_option(forcecrt2type, "TV", sc.red);
	 }
	 break;
      default:
         cf_insert_option(forcecrt2type, "NONE", sc.red);
      }

      if(sc.currentvbflags & (CRT2_LCD | CRT1_LCDA)) {
         if(sc.sdflags & SiS_SD_SUPPORTSCALE) {
	    cf_append_text("\n\t# [sisctrl] LCD related options\n", sc.red);
	    if(!(sc.lcdscaler & 0x01)) {
	       cf_insert_option("ScaleLCD", ((sc.lcdscaler & 0x02) ? "false" : "true"), sc.red);
	    }
	    if(sc.sdflags & SiS_SD_SUPPORTCENTER) {
	       if((sc.lcdscaler & 0x03) && (!(sc.lcdscaler & 0x04))) {
	          cf_insert_option("CenterLCD", ((sc.lcdscaler & 0x08) ? "true" : "false"), sc.red);
	       }
	    }
	 }
      }

      if(sc.currentvbflags & CRT2_TV) {
         cf_append_text("\n\t# [sisctrl] TV related options\n", sc.red);
	 if(!(sc.currentvbflags & (TV_HIVISION | TV_YPBPR))) {
            if(sc.currentvbflags & TV_NTSC) {
	       if(sc.currentvbflags & TV_NTSCJ)
	          cf_insert_option("TVStandard", "NTSCJ", sc.red);
	       else
	          cf_insert_option("TVStandard", "NTSC", sc.red);
	    } else {
	       if(sc.currentvbflags & TV_PALM)
	          cf_insert_option("TVStandard", "PAL-M", sc.red);
	       else if(sc.currentvbflags & TV_PALN)
	          cf_insert_option("TVStandard", "PAL-N", sc.red);
	       else
	          cf_insert_option("TVStandard", "PAL", sc.red);
	    }
	 }
	 if(sc.currentvbflags & TV_YPBPR) {
	    if(sc.sdflags & SiS_SD_SUPPORTYPBPRAR) {
	       if((sc.currentvbflags & TV_YPBPRAR) == TV_YPBPR43)
		  cf_insert_option("YPbPrAspectRatio", "4:3", sc.red);
	       else if((sc.currentvbflags & TV_YPBPRAR) == TV_YPBPR43LB)
		  cf_insert_option("YPbPrAspectRatio", "4:3LB", sc.red);
	       else if((sc.currentvbflags & TV_YPBPRAR) == TV_YPBPR169)
		  cf_insert_option("YPbPrAspectRatio", "16:9", sc.red);
	    }
	 }
	 if(sc.currentvbflags & VB_CHRONTEL) {
	    switch(sc.choverscanstatus) {
	    case 1:
	       cf_insert_option(chtvoverscan, "false", sc.red);
	       cf_insert_option(chtvsoverscan, "false", sc.red);
	       break;
	    case 2:
	       cf_insert_option(chtvoverscan, "true", sc.red);
	       cf_insert_option(chtvsoverscan, "false", sc.red);
	       break;
	    case 3:
	       cf_insert_option(chtvoverscan, "true", sc.red);
	       cf_insert_option(chtvsoverscan, "true", sc.red);
	       break;
	    }
	    sprintf(numbuffer, "%d", getxvval(sc.atom_tco));
	    cf_insert_option("CHTVContrast", numbuffer, sc.red);
	    sprintf(numbuffer, "%d", getxvval(sc.atom_tte));
	    cf_insert_option("CHTVTextEnhance", numbuffer, sc.red);
	    sprintf(numbuffer, "%d", getxvval(sc.atom_tlf));
	    cf_insert_option("CHTVLumaFlickerfilter", numbuffer, sc.red);
	    sprintf(numbuffer, "%d", getxvval(sc.atom_tcf));
	    cf_insert_option("CHTVChromaFlickerfilter", numbuffer, sc.red);
	    cf_insert_option("CHTVCVBSColor", getxvval(sc.atom_ccc) ? "true" : "false", sc.red);
	 }
	 if(sc.currentvbflags & VB_SISBRIDGE) {
	    switch(getxvval(sc.atom_taf)) {
	    case 0:  cf_insert_option(sistvantiflicker, "OFF", sc.red); break;
	    case 1:  cf_insert_option(sistvantiflicker, "LOW", sc.red); break;
	    case 2:  cf_insert_option(sistvantiflicker, "MED", sc.red); break;
	    case 3:  cf_insert_option(sistvantiflicker, "HIGH", sc.red); break;
	    case 4:  cf_insert_option(sistvantiflicker, "ADAPTIVE", sc.red); break;
	    }
	    if(!(sc.currentvbflags & VB_301)) {
	       sprintf(numbuffer, "%d", getxvval(sc.atom_tsa));
	       cf_insert_option("SISTVSaturation", numbuffer, sc.red);
	    }
	    if(sc.currentvbflags & VB_301) {
	       sprintf(numbuffer, "%d", getxvval(sc.atom_tee));
	       cf_insert_option("SISTVEdgeEnhance", numbuffer, sc.red);
	    }
	    sprintf(numbuffer, "%d", getxvval(sc.atom_coc));
	    cf_insert_option("SISTVColorCalibCoarse", numbuffer, sc.red);
	    sprintf(numbuffer, "%d", getxvval(sc.atom_cof));
	    cf_insert_option("SISTVColorCalibFine", numbuffer, sc.red);
	    cf_insert_option("SISTVCFilter", getxvval(sc.atom_cfi) ? "true" : "false", sc.red);
	    sprintf(numbuffer, "%d", getxvval(sc.atom_yfi));
	    cf_insert_option("SISTVYFilter", numbuffer, sc.red);

	    sprintf(numbuffer, "%d", getxvval(sc.atom_txs));
	    cf_insert_option("SISTVXScale", numbuffer, sc.red);
	    sprintf(numbuffer, "%d", getxvval(sc.atom_tys));
	    cf_insert_option("SISTVYScale", numbuffer, sc.red);
	 }
	 if(sc.sdflags & SiS_SD_SUPPORTTVPOS) {
	    sprintf(numbuffer, "%d", getxvval(sc.atom_tvx));
	    cf_insert_option("TVXPosOffset", numbuffer, sc.red);
	    sprintf(numbuffer, "%d", getxvval(sc.atom_tvy));
	    cf_insert_option("TVYPosOffset", numbuffer, sc.red);
	 }

      }
      if((sc.currentvbflags & VB_SISBRIDGE) && (!(sc.gammaflags & 0x02))) {
         cf_append_text("\n\t# [sisctrl] Enable/disable gamma correction for CRT2\n", sc.red);
         cf_insert_option("CRT2Gamma", "false", sc.red);
      }
   }  /* VB_VIDEOBRIDGE */

   if( (!(sc.sdflags & SiS_SD_ISDUALHEAD)) ||
       ((!(sc.sdflags & SiS_SD_ISDHSECONDHEAD)) && (!(sc.sdflags & SiS_SD_ISDHXINERAMA))) ||
       ((sc.sdflags & SiS_SD_ISDHXINERAMA) && (sc.myscreen == DefaultScreen(sc.dpy))) ) {
      appendbrightness();
      if(sc.sdflags & SiS_SD_ISDUALHEAD) {
         cf_append_text("\n\t# [sisctrl] Xv related options for CRT2\n", sc.red);
      } else {
         cf_append_text("\n\t# [sisctrl] Xv related options\n", sc.red);
      }
      appendxvstuff();
   }

   if(!(sc.sdflags & SiS_SD_SUPPORT2OVL)) {
      cf_insert_option("XvOnCRT2", getxvval(sc.atom_swc) ? "true" : "false", sc.red);
   }

   if(!(sc.sdflags & SiS_SD_ISMERGEDFB)) {
      cf_append_text("\n\t# For MergedFB mode, see the following options:\n\n", sc.grey);

      cf_append_text("\t# To enable MergedFB mode regardless of CRT1 detection,\n", sc.grey);
      cf_append_text("\t# uncomment this:\n", sc.grey);
      cf_append_text("\t# Option \"MergedFB\" \"true\"\n\n", sc.grey);

      cf_append_text("\t# To enable MergedFB mode automatically if both CRT1 and CRT2\n", sc.grey);
      cf_append_text("\t# are detected, uncomment this:\n", sc.grey);
      cf_append_text("\t# Option \"MergedFBAuto\" \"true\"\n\n", sc.grey);

      cf_append_text("\t# \"CRT2Position\" selects the relative position of CRT2 to CRT1\n", sc.grey);
      cf_append_text("\t# Valid choices are \"LeftOf\", \"RightOf\", \"Above\", \"Below\" and\n", sc.grey);
      cf_append_text("\t# \"Clone\"\n", sc.grey);
      cf_append_text("\t# Option \"CRT2Position\" \"LeftOf\"\n\n", sc.grey);

      cf_append_text("\t# Fill in the supported frequency ranges for your CRT2 device\n", sc.grey);
      cf_append_text("\t# below. For LCD and TV, the values here don't matter, but\n", sc.grey);
      cf_append_text("\t# they must be high enough to keep the server from deleting\n", sc.grey);
      cf_append_text("\t# some modes. HSync is in kHz, VRefresh in Hz. For example:\n", sc.grey);
      cf_append_text("\t# Option \"CRT2HSync\" \"31-64\"\n", sc.grey);
      cf_append_text("\t# Option \"CRT2VRefresh\" \"50-75\"\n\n", sc.grey);

      cf_append_text("\t# The \"MetaModes\" select the (real) display modes to be used\n", sc.grey);
      cf_append_text("\t# for CRT1 (left of \"-\") and CRT2 (right of \"-\"). Eg:\n", sc.grey);
      cf_append_text("\t# Option \"MetaModes\" \"1280x1024-1024x768 1024x768-1024x768\"\n\n", sc.grey);

      cf_append_text("\t# To force the screen's DPI for MergedFB mode to a fixed value,\n", sc.grey);
      cf_append_text("\t# use this: (first parameter X, second parameter Y)\n", sc.grey);
      cf_append_text("\t# Option \"MergedDPI\" \"100 100\"\n\n", sc.grey);

      cf_append_text("\t# Set this option to disable Pseudo-Xinerama support\n", sc.grey);
      cf_append_text("\t# Option \"NoMergedXinerama\" \"true\"\n\n", sc.grey);

      cf_append_text("\t# To reverse the screen order in Pseudo-Xinerama mode:\n", sc.grey);
      cf_append_text("\t# Option \"MergedXineramaCRT2IsScreen0\" \"true\"\n", sc.grey);
      cf_append_text("\t# or\n", sc.grey);
      cf_append_text("\t# Option \"MergedXineramaCRT2IsScreen0\" \"false\"\n", sc.grey);
   } else {
      cf_append_text("\n\t# ...\n", sc.grey);
      cf_append_text("\t# [your current options for MergedFB mode]\n", sc.grey);
      cf_append_text("\t# ...\n", sc.grey);
   }

   cf_append_text("EndSection\n", sc.black);

   cf_append_text("\n# For dual head / Xinerama, we need a second \"Monitor\" and\n", sc.grey);
   cf_append_text("# \"Device\" section for CRT1 (slave). This is not required for\n", sc.grey);
   cf_append_text("# MergedFB mode (which is recommended over Xinerama).\n\n", sc.grey);

   cf_append_text("# Monitor for CRT1\n", sc.grey);
   cf_append_text("Section \"Monitor\"\n", sc.black);
   cf_append_text("\t# Identifier, VendorName, ModelName,\n", sc.grey);
   cf_append_text("\t# VertRefresh, HorizSync, DPMS, ...\n", sc.grey);
   if((sc.sdflags & SiS_SD_ISDUALHEAD) && (sc.sdflags & (SiS_SD_ISDHSECONDHEAD | SiS_SD_ISDHXINERAMA))) {
      if((!(sc.sdflags & SiS_SD_ISDHXINERAMA)) || (sc.myscreen != DefaultScreen(sc.dpy))) {
         cf_append_text("\t# [sisctrl] Gamma correction\n", sc.red);
         sprintf(numbuffer, "\tGamma %.3f %.3f %.3f\n", sc.currentgamma.red, sc.currentgamma.green,
	 				sc.currentgamma.blue);
	 for(i=0; i<strlen(numbuffer); i++) {
	    if(numbuffer[i] == ',') numbuffer[i] = '.';
	 }
	 cf_append_text(numbuffer, sc.red);
      } else {
         cf_append_text("\t# Gamma, ...\n", sc.grey);
      }
   }
   cf_append_text("EndSection\n\n", sc.black);

   cf_append_text("# This short \"Device\" section should suffice as most options can\n", sc.grey);
   cf_append_text("# only be given in the \"Device\" section for CRT2 (master).\n", sc.grey);
   cf_append_text("# Options that must be given in this second \"Device\" section are\n", sc.grey);
   cf_append_text("# all Xv and gamma correction related options for CRT1, such as\n", sc.grey);
   cf_append_text("# all \"XvDefault..\" options, \"XvGamma\", \"CRT1Gamma\",\n", sc.grey);
   cf_append_text("# \"StoredGammaBrightness\", \"StoredGammaPreBrightness\".\n\n", sc.grey);

   cf_append_text("# Device for CRT1\n", sc.grey);
   cf_append_text("Section \"Device\"\n", sc.black);
   cf_append_text("\t# Replace IDENTIFIER2 with the real identifier from the \"Device\"\n", sc.grey);
   cf_append_text("\t# tag in your second \"Screen\" section\n", sc.grey);
   cf_append_text("\tIdentifier \"IDENTIFIER2\"\n", sc.black);

   cf_append_text("\tVendorName \"SiS\"\n", sc.grey);
   sprintf(numbuffer, "\tBoardName \"%s\"\n", getchiptype());
   cf_append_text(numbuffer, sc.grey);

   cf_append_text("\tDriver \"sis\"\n", sc.black);

   sprintf(numbuffer, "\tBusID \"PCI:%d:%d:%d\"\n", (sc.busid >> 16), (sc.busid >> 8) & 0xff, sc.busid & 0xff);
   cf_append_text(numbuffer, sc.grey);

   cf_append_text("\tScreen 1\n", sc.grey);

   if((sc.sdflags & SiS_SD_ISDUALHEAD) && (sc.sdflags & (SiS_SD_ISDHSECONDHEAD | SiS_SD_ISDHXINERAMA))) {
      if((!(sc.sdflags & SiS_SD_ISDHXINERAMA)) || (sc.myscreen != DefaultScreen(sc.dpy))) {
         appendcrt1gamma();
         appendbrightness();
	 cf_append_text("\n\t# [sisctrl] Xv related options for CRT1\n", sc.red);
	 appendxvstuff();
      }
   }

   cf_append_text("EndSection\n", sc.black);
}

/* Info page */

gboolean ab_layout_expose(GtkWidget *widget, GdkEventExpose *event, gpointer data)
{
   int w, h, i, hstepwidth, vstepwidth;
   gint lwidth = widget->allocation.width;
   gint lheight = widget->allocation.height;
   PangoLayout *playout;
   GtkWidget *labels[] = { sc.ab_label_ab1,
   			   sc.ab_label_ab2,
			   sc.ab_label_ab4,
			   sc.ab_label_ab3,
			   sc.ab_label_ab5};

#define BW 4   /* border width */
#define BWT 8  /* border width for text area */

   gdk_draw_rectangle(GTK_LAYOUT(widget)->bin_window,
   		      widget->style->fg_gc[GTK_WIDGET_STATE(widget)],
		      TRUE,
		      BW, BW,
		      lwidth - (BW * 2), lheight - (BW * 2));

   gdk_draw_drawable(GTK_LAYOUT(widget)->bin_window,
   		     widget->style->fg_gc[GTK_WIDGET_STATE(widget)],
		     sc.ab_iconul,
		     0, 0,
		     BW, BW,
		     92, 130);

   gdk_draw_drawable(GTK_LAYOUT(widget)->bin_window,
   		     widget->style->fg_gc[GTK_WIDGET_STATE(widget)],
		     sc.ab_iconlr,
		     0, 0,
		     lwidth - 71 - BW, lheight - 90 - BW,
		     71, 90);

   hstepwidth = ((lwidth - (BWT * 2)) / 2);
   vstepwidth = ((lheight - (BWT * 2)) / 6);

   gdk_draw_drawable(GTK_LAYOUT(widget)->bin_window,
   		     widget->style->fg_gc[GTK_WIDGET_STATE(widget)],
		     sc.ab_icon,
		     0, 0,
		     hstepwidth - (92 / 2) + BWT,
		     ((vstepwidth - (20 / 2)) / 2) + BWT,
		     92, 20);

   for(i = 0; i < 5; i++) {
      playout = gtk_label_get_layout(GTK_LABEL(labels[i]));
      pango_layout_get_pixel_size(playout, &w, &h);
      gdk_draw_layout(GTK_LAYOUT(widget)->bin_window,
                   labels[i]->style->fg_gc[GTK_WIDGET_STATE(labels[i])],
		   hstepwidth - (w / 2) + BWT,
		   (vstepwidth * (i + 1)) + ((vstepwidth - (h / 2)) / 2) + BWT,
		   playout);
   }

   return TRUE;
}

/* Notebook signal handler */

gboolean notebook_changed(GtkNotebook *nb, GtkNotebookPage *page, guint page_num, gpointer data)
{
   int i;
   gint temp;

   sc.notebook_page = page_num;

#ifdef USEXV
   if(!sc.noxvdemo) {
      if(page_num == 5) {
         showxvimage();
      } else {
         hidexvimage();
      }
   }
#endif

   if(page_num > 2)  gtk_widget_set_sensitive(sc.button2, FALSE);
   else gtk_widget_set_sensitive(sc.button2, TRUE);

   switch(page_num) {
   case 0:
      sc.backupmodeindex = sc.currentmodeindex;
#ifdef HAVE_RANDR
      if(sc.haverandr) {
         sc.rrrotation_backup = sc.rrcurrent_rotation;
         sc.rrsize_backup = sc.rrcurrent_size;
      }
#endif
      gtk_widget_set_sensitive(sc.button4, FALSE);
      setdmhistory();
#ifdef HAVE_RANDR
      if(sc.haverandr) {
         updateleaveunchanged();
	 temp = gtk_option_menu_get_history(GTK_OPTION_MENU(sc.dm_omenu));
	 if(temp >= 2) {
	    gtk_option_menu_set_history(GTK_OPTION_MENU(sc.dm_omenu), 0);
	 }
      }
#endif
      break;
   case 1:
      sc.backupvbflags = sc.currentvbflags;
      sc.CRT1isoffbackup = sc.CRT1isoff;
      sc.backuplcdscaler = sc.lcdscaler;
      gtk_widget_set_sensitive(sc.button4, FALSE);
      break;
   case 2:
      sc.backupvbflags = sc.currentvbflags;
      sc.backupoverscan = sc.choverscanstatus;
      sc.backuplcdscaler = sc.lcdscaler;
      gtk_widget_set_sensitive(sc.button4, FALSE);
      break;
   case 3:
      sc.backupgammaflags = sc.gammaflags;
      sc.backup_gred = sc.newgamma.red;
      sc.backup_ggreen = sc.newgamma.green;
      sc.backup_gblue = sc.newgamma.blue;
      for(i=0; i<3; i++) {
         sc.gamma_max_b[i] = sc.gamma_max[i];
	 sc.gamma_pre_b[i] = sc.gamma_prescale[i];
      }
      gtk_widget_set_sensitive(sc.button4, TRUE);
      break;
   case 6:
      cf_generate_current();
      /* fall through */
   default:
      gtk_widget_set_sensitive(sc.button4, FALSE);
   }

   return FALSE;
}

/* "OK", "Apply" handlers */

static gint
getcrt2lsmenuindex(int menunum)
{
   gint mode=0, index;

   if(!(sc.sdflags & SiS_SD_SUPPORTSCALE)) return((0x01 | 0x04));

   if(menunum == 1) {
      if(!sc.c1_omenuls) return ((0x01 | 0x04));
   } else {
      if(!sc.om_omenuls) return ((0x01 | 0x04));
   }

   if(sc.guimode) {
      if(menunum == 2) index = gtk_option_menu_get_history(GTK_OPTION_MENU(sc.om_omenuls));
      else             index = gtk_option_menu_get_history(GTK_OPTION_MENU(sc.c1_omenuls));
      switch(index) {
      case 0: mode |= 0x01; break;
      case 2: mode |= 0x02; break;
      }
      if(sc.sdflags & SiS_SD_SUPPORTCENTER) {
         if(menunum == 2) index = gtk_option_menu_get_history(GTK_OPTION_MENU(sc.om_omenulcc));
	 else             index = gtk_option_menu_get_history(GTK_OPTION_MENU(sc.c1_omenulcc));
         switch(index) {
         case 0: mode |= 0x04; break;
         case 2: mode |= 0x08; break;
         }
      }
      return(mode);
   } else
      return((0x01 | 0x04));  /* both default */
}

static Bool
handle_apply_CRT1(int c1_index)
{
   int val;
   gint newpanelmode;
   unsigned long myvbflags = sc.currentvbflags;

   /* Read CRT1 status here again; user might have changed the mode without us
    * (Slave mode modes switch CRT1 on!)
    */
   if(XvGetPortAttribute(sc.dpy, sc.xv_port_nr, sc.atom_ct1, &val) != Success) {
      fprintf(stderr, "Couldn't read CRT1 status\n");
   } else sc.CRT1isoff = val ? 0 : 1;

   /* index 0 = OFF, 1 = ON, 2 = LCDA */
   if(c1_index == 0) {
      if(!(sc.currentvbflags & CRT2_ENABLE)) {
	 if(sc.guimode) {
	    showerror("CRT1 can't be switched off, no CRT2 device enabled");
	    setcrt1history();
	    sc.currentmodeindex = setdmhistory();
	 } else {
	    fprintf(stderr, "CRT1 can't be switched off, no CRT2 device enabled\n");
	 }
         return FALSE;
      }
      if(sc.sdflags & SiS_SD_ISMERGEDFB) {
	 if(sc.guimode) {
	    showerror("CRT1 can't be switched off in MergedFB mode");
	    setcrt1history();
	    sc.currentmodeindex = setdmhistory();
	 } else {
	    fprintf(stderr, "CRT1 can't be switched off in MergedFB mode\n");
	 }
         return FALSE;
      }
      if(sc.CRT1isoff) {
	 if(sc.guimode) {
 	    setcrt1history();
	    sc.currentmodeindex = setdmhistory();
	 }
	 return TRUE;
      }
   } else if(c1_index == 1) {
      if((!sc.CRT1isoff) && (!(sc.currentvbflags & CRT1_LCDA))) {
         if(sc.guimode) {
	    setcrt1history();
	    sc.currentmodeindex = setdmhistory();
	 }
	 return TRUE;
      }
   } else {
      if(!(sc.sdflags & SiS_SD_SUPPORTLCDA)) {
         if(sc.guimode) {
	    showerror("Hardware does not support LCD-via-CRT1\n");
	 } else {
	    fprintf(stderr, "Hardware does not support LCD-via-CRT1\n");
	 }
	 return FALSE;
      }
      if(sc.sdflags & SiS_SD_SUPPORTSCALE) {
         newpanelmode = getcrt2lsmenuindex(1);
      }
      if( ((!sc.CRT1isoff) && (sc.currentvbflags & CRT1_LCDA)) &&
          ((!(sc.sdflags & SiS_SD_SUPPORTSCALE)) || (newpanelmode == sc.lcdscaler)) ) {
	 if(sc.guimode) {
	    setcrt1history();
	    sc.currentmodeindex = setdmhistory();
	    setlcdhistory();
	 }
	 return TRUE;
      }
      if((sc.sdflags & SiS_SD_ISMERGEDFB) && (sc.currentvbflags & CRT2_LCD)) {
	 if(sc.guimode) {
	    showerror("Can't switch to LCD-via-CRT1 while CRT2 is LCD");
	    setcrt1history();
	    sc.currentmodeindex = setdmhistory();
	    setlcdhistory();
	 } else {
	    fprintf(stderr, "Can't switch to LCD-via-CRT1 while CRT2 is LCD\n");
	 }
         return FALSE;
      }
      if(sc.sdflags & SiS_SD_SUPPORTSCALE) {
         if(UnlockSiSDirect(1)) {
	    int val;
	    if(XvSetPortAttribute(sc.dpy, sc.xv_port_nr, sc.atom_pmd, newpanelmode) != Success) {
               fprintf(stderr, xverrorstring);
            }
            XFlush(sc.dpy);
	    if(XvGetPortAttribute(sc.dpy, sc.xv_port_nr, sc.atom_pmd, &val) != Success) {
               fprintf(stderr, xverrorstring);
            } else {
	       sc.lcdscaler = val & 0x0f;
	    }
	    UnlockSiSDirect(0);
	 }
      }
   }

   if(UnlockSiSDirect(1)) {
      if(XvSetPortAttribute(sc.dpy, sc.xv_port_nr, sc.atom_ct1, c1_index) != Success) {
         fprintf(stderr, xverrorstring);
      }
      XFlush(sc.dpy);
      UnlockSiSDirect(0);

      updatevbandcrt1();
      if( (myvbflags == sc.currentvbflags)            &&
          ( (!(sc.sdflags & SiS_SD_SUPPORTLCDA))  ||
	    (!(sc.sdflags & SiS_SD_SUPPORTSCALE)) ||
	    (newpanelmode != sc.lcdscaler) ) ) {
         if(sc.guimode) {
	    showerror("Operation failed. Check XFree86 log for error messages");
	 } else {
	    fprintf(stderr, "Operation failed. Check XFree86 log for error messages\n");
	 }
	 return FALSE;
      }
   }

   return TRUE;
}

static gint
getcrt2menuindex(void)
{
   if(sc.guimode)
      return(gtk_option_menu_get_history(GTK_OPTION_MENU(sc.om_omenu)));
   else
      return((gint)sc.givencrt2index);
}

static gint
getcrt2tvstdmenuindex(void)
{
   if(sc.guimode)
      return(gtk_option_menu_get_history(GTK_OPTION_MENU(sc.om_omenut)));
   else
      return((gint)sc.giventvstdindex);
}

static gint
getcrt2tvasmenuindex(void)
{
   if(sc.guimode)
      return(gtk_option_menu_get_history(GTK_OPTION_MENU(sc.om_omenuas)));
   else
      return((gint)sc.giventvasindex);
}

static gint
getcrt2tvosmenuindex(void)
{
   if(sc.guimode)
      return(gtk_option_menu_get_history(GTK_OPTION_MENU(sc.om_omenuo)));
   else
      return((gint)sc.giventvosindex);
}

static Bool
handle_apply_CRT2(void)
{
   int newoverscan, newpanelmode;
   gint om_index = getcrt2menuindex();

   if(om_index == MI_TV_OFF) {
      if(sc.CRT1isoff) {
         if(sc.guimode) {
            showerror("CRT2 can't be switched off while CRT1 is off\n");
	 } else {
	    fprintf(stderr, "CRT2 can't be switched off while CRT1 is off\n");
	 }
	 return FALSE;
      }
      if(sc.sdflags & SiS_SD_ISMERGEDFB) {
         if(sc.guimode) {
	    showerror("CRT2 can't be switched off is MergedFB mode\n");
	 } else {
	    fprintf(stderr, "CRT2 can't be switched off is MergedFB mode\n");
	 }
	 return FALSE;
      }
   }

   sc.newvbflags = sc.currentvbflags & ~(CRT2_ENABLE | TV_INTERFACE | TV_STANDARD);
   switch(om_index) {
      case MI_TV_LCD: 	sc.newvbflags |= CRT2_LCD;
	      		sc.newvbflags &= ~CRT1_LCDA;	      break;
      case MI_TV_SVHS:  sc.newvbflags |= CRT2_TV | TV_SVIDEO; break;
      case MI_TV_CVBS:  sc.newvbflags |= CRT2_TV | TV_AVIDEO; break;
      case MI_TV_CVSV:  sc.newvbflags |= CRT2_TV | TV_SVIDEO | TV_AVIDEO; break;
      case MI_TV_HIVI:  sc.newvbflags |= CRT2_TV | TV_HIVISION; break;
      case MI_TV_525I:  sc.newvbflags |= CRT2_TV | TV_YPBPR | TV_YPBPR525I; break;
      case MI_TV_525P:  sc.newvbflags |= CRT2_TV | TV_YPBPR | TV_YPBPR525P; break;
      case MI_TV_750P:  sc.newvbflags |= CRT2_TV | TV_YPBPR | TV_YPBPR750P; break;
      case MI_TV_1080I: sc.newvbflags |= CRT2_TV | TV_YPBPR | TV_YPBPR1080I; break;
      case MI_TV_TV: 	sc.newvbflags |= CRT2_TV | TV_SVIDEO; break;
      case MI_TV_VGA: 	sc.newvbflags |= CRT2_VGA;            break;
      case MI_TV_SCART: sc.newvbflags |= CRT2_TV | TV_SCART;  break;
   }
   if(sc.newvbflags & TV_YPBPR) {
      sc.newvbflags &= ~(TV_YPBPRAR);
      if(sc.sdflags & SiS_SD_SUPPORTYPBPRAR) {
         gint indext = getcrt2tvasmenuindex();
	 switch(indext) {
	    case 0: sc.newvbflags |= TV_YPBPR43; break;
	    case 1: sc.newvbflags |= TV_YPBPR43LB; break;
	    case 2: sc.newvbflags |= TV_YPBPR169; break;
	 }
      }
   }
   if((sc.sdflags & SiS_SD_SUPPORTSCALE) && (sc.newvbflags & CRT2_LCD)) {
      newpanelmode = getcrt2lsmenuindex(2);
   }
   if((sc.newvbflags & CRT2_TV) && (!(sc.newvbflags & TV_YPBPR))) {
      gint indext = getcrt2tvstdmenuindex();
      sc.newvbflags &= ~(TV_PAL | TV_NTSC | TV_PALM | TV_PALN | TV_NTSCJ);
      if(om_index == MI_TV_SCART) sc.newvbflags |= TV_PAL;
      else if(om_index == MI_TV_HIVI) sc.newvbflags |= TV_PAL;
      else {
         switch(indext) {
            case 0: sc.newvbflags |= TV_PAL;  break;
	    case 1: sc.newvbflags |= TV_NTSC; break;
	    case 2: sc.newvbflags |= (TV_PAL | TV_PALM); break;
	    case 3: sc.newvbflags |= (TV_PAL | TV_PALN); break;
	    case 4: sc.newvbflags |= (TV_NTSC | TV_NTSCJ); break;
         }
      }
      if(sc.currentvbflags & VB_CHRONTEL) {
	 newoverscan = getcrt2tvosmenuindex();
      }
   }

   if( (sc.newvbflags != sc.currentvbflags) ||
       ( (sc.currentvbflags & VB_CHRONTEL) &&
         (sc.newvbflags & CRT2_TV)         &&
	 (newoverscan != sc.choverscanstatus) ) ||
       ( (sc.sdflags & SiS_SD_SUPPORTSCALE) &&
         (sc.newvbflags & CRT2_LCD)         &&
	 (newpanelmode != sc.lcdscaler) ) ) {
      if(UnlockSiSDirect(1)) {
	 if((sc.currentvbflags & VB_CHRONTEL) && (sc.newvbflags & CRT2_TV)) {
	    int val;
	    if(XvSetPortAttribute(sc.dpy, sc.xv_port_nr, sc.atom_ovr, newoverscan) != Success) {
               fprintf(stderr, xverrorstring);
            }
            XFlush(sc.dpy);
	    if(XvGetPortAttribute(sc.dpy, sc.xv_port_nr, sc.atom_ovr, &val) != Success) {
               fprintf(stderr, xverrorstring);
            } else {
	       sc.choverscanstatus = val;
	    }
	 }
	 if((sc.sdflags & SiS_SD_SUPPORTSCALE) && (sc.newvbflags & CRT2_LCD)) {
	    int val;
	    if(XvSetPortAttribute(sc.dpy, sc.xv_port_nr, sc.atom_pmd, newpanelmode) != Success) {
               fprintf(stderr, xverrorstring);
            }
            XFlush(sc.dpy);
	    if(XvGetPortAttribute(sc.dpy, sc.xv_port_nr, sc.atom_pmd, &val) != Success) {
               fprintf(stderr, xverrorstring);
            } else {
	       sc.lcdscaler = val & 0x0f;
	    }
	 }
         if(XvSetPortAttribute(sc.dpy, sc.xv_port_nr, sc.atom_svf, sc.newvbflags) != Success) {
            fprintf(stderr, xverrorstring);
         }
         XFlush(sc.dpy);
	 UnlockSiSDirect(0);

	 updatevbandcrt1();

	 if(sc.guimode) {
	    updatecrt1crt2();
	    updatealltv();
	 }

	 if((sc.sdflags & SiS_SD_SUPPORTSCALE) && (sc.newvbflags & CRT2_LCD)) {
	    if(sc.lcdscaler != newpanelmode) {
	       if(sc.guimode) {
	          showerror("Failed to change LCD scaling mode");
	       } else {
	          fprintf(stderr, "Failed to change LCD scaling mode\n");
	       }
	    }
	 }

	 if((sc.currentvbflags & ~DISPLAY_MODE) != (sc.newvbflags & ~DISPLAY_MODE)) {
	    if(!(sc.newvbflags & CRT2_ENABLE)) {
	       if(sc.guimode) {
	          showerror("CRT2 cannot be switched off");
	       } else {
	          fprintf(stderr, "CRT2 cannot be switched off, check XFree log for error message\n");
	       }
	    } else {
	       if(sc.guimode) {
	          showerror("Current display mode not supported for this CRT2 type");
	       } else {
	          fprintf(stderr, "Current display mode not supported for this CRT2 type\n");
	       }
	    }
	    return FALSE;
	 } else {
	    if(sc.guimode) {
	       gtk_widget_set_sensitive(sc.button4, TRUE);
	    }
	 }
      }
   }
}

/* Bottom buttons */

static Bool
do_apply(void)
{
   gint nb_page = gtk_notebook_get_current_page(GTK_NOTEBOOK(sc.notebook));
   Bool ret = TRUE;

   switch(nb_page) {

   case 0:    /* apply display mode */
      {
      int  dm_index = -1, i;
      Bool dorandr = FALSE, matchsize = FALSE;
      int  checkmask = 0xff;
      int  checkresult = 0;
      int  sizetoset = 0;
      gint randrindex = 0;

      for(i = 0; i < MYMAXMODES; i++) {
         if(gtk_tree_selection_iter_is_selected(sc.dm_select, &sc.dm_iter[i])) {
	    dm_index = i;
	    break;
	 }
      }
      if(dm_index == -1) {
         fprintf(stderr, "Internal error: List iterator not found\n");
	 ret = FALSE;
	 break;
      }

#ifdef HAVE_RANDR
      if(sc.haverandr) {
         randrindex = gtk_option_menu_get_history(GTK_OPTION_MENU(sc.dm_omenu));
         if(randrindex > 0) {
            dorandr = TRUE;
	    checkmask = ~0x02;
	    if(randrindex == 1) matchsize = TRUE;
	    else sizetoset = randrindex - 2;
	 }
	 sc.rrcurrent_size = XRRConfigCurrentConfiguration(sc.sconf, &sc.rrcurrent_rotation);
      }
#endif

      if((checkresult = checkmodeforcrt2(dm_index)) & checkmask) {
         checkresult &= 0xff;
         if(checkresult & 0x01) {
	    showerror("Display mode not supported for current CRT2 type");
	 } else if(checkresult & 0x02) {
	    showerror("Display mode too large for current root window size");
	 } else {
	    char buffer[128];
	    sprintf(buffer, "Failed to verify mode (error code %02x)", checkresult);
	    showerror(buffer);
	 }
	 sc.currentmodeindex = setdmhistory();
	 ret = FALSE;
	 break;
      }

      if(sc.haverandr) {
#ifdef HAVE_RANDR
	 Bool sizefound = FALSE;
   	 sc.rrsizes = XRRConfigSizes(sc.sconf, &sc.rrnsize);
	 if(dorandr) {
	    if(matchsize) {
   	       for(sc.rrsize = 0; sc.rrsize < sc.rrnsize; sc.rrsize++) {
                  if( (sc.rrsizes[sc.rrsize].width == sc.vmmodelines[dm_index]->hdisplay) &&
	              (sc.rrsizes[sc.rrsize].height == sc.vmmodelines[dm_index]->vdisplay) ) {
	             sizefound = TRUE;
	             sc.rrrate = sc.moderate[dm_index];
	             break;
	          }
	       }
	    } else {
	       if( (sc.rrsizes[sizetoset].width >= sc.vmmodelines[dm_index]->hdisplay) &&
	           (sc.rrsizes[sizetoset].height >= sc.vmmodelines[dm_index]->vdisplay) ) {
	          sc.rrsize = sizetoset;
	          sc.rrrate = 0;
	          sizefound = TRUE;
	       }
	    }
         }
	 if((sizefound) && (sc.rrcurrent_size != sc.rrsize)) {
	    Status status = RRSetConfigFailed;
   	    XRRConfigRotations(sc.sconf, &sc.rrcurrent_rotation);
   	    XSelectInput(sc.dpy, sc.root, StructureNotifyMask);
   	    XRRSelectInput(sc.dpy, sc.root, RRScreenChangeNotifyMask);
   	    status = XRRSetScreenConfigAndRate(sc.dpy, sc.sconf, DefaultRootWindow(sc.dpy),
	    	       (SizeID)sc.rrsize, (Rotation)sc.rrcurrent_rotation, sc.rrrate, CurrentTime);
	    if(status != RRSetConfigSuccess) {
	       fprintf(stderr, "XRRSetScreenConfigAndRate failed\n");
	       if(!checkmodeforscreen(dm_index)) {
	          showerror("Display mode too large for current root window size");
	          sc.currentmodeindex = setdmhistory();
	          ret = FALSE;
	       }
	    }
	    updateleaveunchanged();
  	 }
#endif
      }

      if((!dorandr) || (!matchsize)) {
         if(!checkmodeforscreen(dm_index)) {
            showerror("Display mode too large for current root window size");
	    sc.currentmodeindex = setdmhistory();
	    ret = FALSE;
	 }
      }

      if(ret) {
         XFlush(sc.dpy);

         if(!XF86VidModeSwitchToMode(sc.dpy, sc.myscreen, sc.vmmodelines[dm_index])) {
	    showerror("Error calling vidmode extension");
	    ret = FALSE;
         }
         XFlush(sc.dpy);

	 updatevbandcrt1();

	 updatecrt1crt2();

	 updatealltv();

#ifdef HAVE_RANDR
	 if(sc.haverandr) {
	    sc.rrcurrent_size = XRRConfigCurrentConfiguration(sc.sconf, &sc.rrcurrent_rotation);
	 }
#endif

	 if(ret == TRUE) {
	    gtk_widget_set_sensitive(sc.button4, TRUE);
	 }

      }

      break;
      }

   case 1:    /* apply CRT1 settings */
      {
      gint c1_index;

      if((sc.sdflags & SiS_SD_ISDUALHEAD) || (!(sc.currentvbflags & VB_VIDEOBRIDGE))) break;

      c1_index = gtk_option_menu_get_history(GTK_OPTION_MENU(sc.c1_omenu));

      if(!(handle_apply_CRT1(c1_index))) return FALSE;

      updatecrt1crt2();

      gtk_widget_set_sensitive(sc.button4, TRUE);

      break;
      }

   case 2:    /* apply CRT2 settings */

      if((sc.sdflags & SiS_SD_ISDUALHEAD) || (!(sc.currentvbflags & VB_VIDEOBRIDGE))) break;

      if(!(handle_apply_CRT2())) return FALSE;

      break;

   case 3:   /* apply Gamma settings - done directly */
   case 4:   /* apply TV settings - done directly */
   case 5:   /* apply Video settings - done directly */
   default:

     break;

   }
   return ret;
}

void apply_clicked(GtkWidget *widget, gpointer data)
{
   do_apply();
}

void ok_clicked(GtkWidget *widget, gpointer data)
{
   if(do_apply()) {
      cleanupandexit(0);
      gtk_main_quit();
   }
}

void close_clicked(GtkWidget *widget, gpointer data)
{
}

void revert_clicked(GtkWidget *widget, gpointer data)
{
   gint nb_page = gtk_notebook_get_current_page(GTK_NOTEBOOK(sc.notebook));
   int i;
   Bool didit = FALSE;

   switch(nb_page) {

   case 0:    /* revert display mode */
      {
      int checkmask = 0xff;
      if((sc.backupmodeindex != sc.currentmodeindex)
#ifdef HAVE_RANDR
         						 ||
         (sc.rrrotation_backup != sc.rrcurrent_rotation) ||
	 (sc.rrsize_backup != sc.rrcurrent_size)
#endif
						) {
#ifdef HAVE_RANDR
         fprintf(stderr, "b_s %d c_s %d c_m %d\n", sc.rrsize_backup, sc.rrcurrent_size, sc.currentmodeindex);
         if((sc.rrrotation_backup != sc.rrcurrent_rotation) ||
	    (sc.rrsize_backup != sc.rrcurrent_size)) {
	    XSelectInput(sc.dpy, sc.root, StructureNotifyMask);
   	    XRRSelectInput(sc.dpy, sc.root, RRScreenChangeNotifyMask);
   	    XRRSetScreenConfigAndRate(sc.dpy, sc.sconf, DefaultRootWindow(sc.dpy),
	    	       (SizeID)sc.rrsize_backup, (Rotation)sc.rrrotation_backup,
		       sc.moderate[sc.backupmodeindex], CurrentTime);
	    sc.rrcurrent_size = XRRConfigCurrentConfiguration(sc.sconf, &sc.rrcurrent_rotation);
	    checkmask = ~0x02;
	 }
#endif
         if(checkmodeforcrt2(sc.backupmodeindex) & checkmask) {
	    XFlush(sc.dpy);
	    sc.currentvbflags &= ~BMASK;
	    sc.currentvbflags |= (sc.backupvbflags & BMASK);

	    if(UnlockSiSDirect(1)) {
               if(XvSetPortAttribute(sc.dpy, sc.xv_port_nr, sc.atom_svf, sc.currentvbflags) != Success) {
                  fprintf(stderr, xverrorstring);
               }
               XFlush(sc.dpy);
	       UnlockSiSDirect(0);

	       updatevbandcrt1();
	    }
         }

	 XFlush(sc.dpy);

	 if(checkmodeforscreen(sc.backupmodeindex)) {

            if(!XF86VidModeSwitchToMode(sc.dpy, sc.myscreen, sc.vmmodelines[sc.backupmodeindex])) {
               fprintf(stderr, "Failed to set the video mode\n");
            }
            XFlush(sc.dpy);
	    gtk_widget_set_sensitive(sc.button4, FALSE);

	 } else {
	    showerror("Original mode too large for current root window size");
	 }

#ifdef HAVE_RANDR
 	 if(sc.haverandr) {
	    updateleaveunchanged();
	 }
#endif

         updatevbandcrt1();

	 updatecrt1crt2();
      }

      break;
      }

   case 1:    /* revert CRT1 settings */

      if((sc.sdflags & SiS_SD_ISDUALHEAD) || (!(sc.currentvbflags & VB_VIDEOBRIDGE))) break;

      if(sc.sdflags & SiS_SD_SUPPORTSCALE) {
         if(sc.lcdscaler != sc.backuplcdscaler) {
	    if(UnlockSiSDirect(1)) {
	       if(XvSetPortAttribute(sc.dpy, sc.xv_port_nr, sc.atom_pmd, sc.backuplcdscaler) != Success) {
                  fprintf(stderr, xverrorstring);
               }
               XFlush(sc.dpy);
	       sc.lcdscaler = sc.backuplcdscaler;
	       UnlockSiSDirect(0);
	       didit = TRUE;
	    }
	 }
      }

      if( (sc.CRT1isoff != sc.CRT1isoffbackup) ||
          ((sc.currentvbflags & CRT1_LCDA) != (sc.backupvbflags & CRT1_LCDA)) ||
	  (didit) ) {
         if(UnlockSiSDirect(1)) {
	    if(sc.sdflags & SiS_SD_SUPPORTLCDA) {
	       int val = 1;
	       if(sc.CRT1isoffbackup) val = 0;
	       if((sc.currentvbflags & CRT1_LCDA) != (sc.backupvbflags & CRT1_LCDA)) {
	          sc.currentvbflags &= ~BMASK;
	 	  sc.currentvbflags |= (sc.backupvbflags & BMASK);
		  if(XvSetPortAttribute(sc.dpy, sc.xv_port_nr, sc.atom_svf, sc.currentvbflags) != Success) {
                     fprintf(stderr, xverrorstring);
                  }
                  XFlush(sc.dpy);
		  if(sc.currentvbflags & CRT1_LCDA) val = 2;
	       }
	       if(XvSetPortAttribute(sc.dpy, sc.xv_port_nr, sc.atom_ct1, val) != Success) {
                  fprintf(stderr, xverrorstring);
               }
	    } else {
               if(XvSetPortAttribute(sc.dpy, sc.xv_port_nr, sc.atom_ct1, sc.CRT1isoffbackup ? 0 : 1) != Success) {
                  fprintf(stderr, xverrorstring);
               }
	    }
            XFlush(sc.dpy);
	    UnlockSiSDirect(0);
	    didit = TRUE;
         }
      }

      if(didit) {
         updatevbandcrt1();
	 updatecrt1crt2();
	 gtk_widget_set_sensitive(sc.button4, FALSE);
      }
      break;

   case 2:    /* revert CRT2 type */

      if((sc.sdflags & SiS_SD_ISDUALHEAD) || (!(sc.currentvbflags & VB_VIDEOBRIDGE))) break;

      if(sc.sdflags & SiS_SD_SUPPORTSCALE) {
         if(sc.lcdscaler != sc.backuplcdscaler) {
	    if(UnlockSiSDirect(1)) {
	       if(XvSetPortAttribute(sc.dpy, sc.xv_port_nr, sc.atom_pmd, sc.backuplcdscaler) != Success) {
                  fprintf(stderr, xverrorstring);
               }
               XFlush(sc.dpy);
	       sc.lcdscaler = sc.backuplcdscaler;
	       UnlockSiSDirect(0);
	       didit = TRUE;
	    }
	 }
      }

      if( ((sc.currentvbflags & BMASK) != (sc.backupvbflags & BMASK)) ||
          ((sc.currentvbflags & VB_CHRONTEL) && (sc.choverscanstatus != sc.backupoverscan)) ||
	  (didit) ) {

         sc.currentvbflags &= ~BMASK;
	 sc.currentvbflags |= (sc.backupvbflags & BMASK);

	 if(UnlockSiSDirect(1)) {
	    if(sc.currentvbflags & VB_CHRONTEL) {
	       if(XvSetPortAttribute(sc.dpy, sc.xv_port_nr, sc.atom_ovr, sc.backupoverscan) != Success) {
                  fprintf(stderr, xverrorstring);
               }
               XFlush(sc.dpy);
  	    }
            if(XvSetPortAttribute(sc.dpy, sc.xv_port_nr, sc.atom_svf, sc.currentvbflags) != Success) {
               fprintf(stderr, xverrorstring);
            }
            XFlush(sc.dpy);
	    UnlockSiSDirect(0);
	    updatevbandcrt1();
	    didit = TRUE;
         }
      }

      if(didit) {
	 updatecrt1crt2();
	 updatealltv();
	 gtk_widget_set_sensitive(sc.button4, FALSE);
      }
      break;

   case 3:   /* Revert gamma */

      memcpy(&sc.newgamma, &sc.currentgamma, sizeof(XF86VidModeGamma));
      sc.newgamma.red = sc.backup_gred;
      sc.newgamma.green = sc.backup_ggreen;
      sc.newgamma.blue = sc.backup_gblue;
      for(i=0; i<3; i++) {
         sc.gamma_max[i] = sc.gamma_max_b[i];
	 sc.gamma_prescale[i] = sc.gamma_pre_b[i];
      }
      setgamma();
      setgammaspinvalues();
      if(!(sc.sdflags & SiS_SD_ISDHXINERAMA)) {
         if((!(sc.sdflags & SiS_SD_ISDUALHEAD)) || (sc.sdflags & SiS_SD_ISDHSECONDHEAD)) {
            endisablegamma1((sc.backupgammaflags & 0x01));
	 }
	 if((!(sc.sdflags & SiS_SD_ISDUALHEAD)) || (!(sc.sdflags & SiS_SD_ISDHSECONDHEAD))) {
            endisablegamma2((sc.backupgammaflags & 0x02));
	 }
         setgammastatus();
      }
      break;

   default:
      break;

   }
}

/* Generic signal handlers */

gint delete_event(GtkWidget *widget, GdkEvent *event, gpointer data)
{
    cleanupandexit(0);
    return FALSE;
}

/* destroy callback */
void destroy(GtkWidget *widget, gpointer data)
{
    cleanupandexit(0);
    gtk_main_quit();
}

/**********************/
/*      Build GUI     */
/**********************/

static void
buildgui(void)
{
    int i,j,k;
    char label3buffer[128];
    char label5buffer[128];
    GtkWidget *vitable;

    /* create a new window */
    sc.window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
    gtk_widget_realize(sc.window); /* realize here for gdk calls */

    /* set window title */
    gtk_window_set_title(GTK_WINDOW(sc.window), "SiS display control panel");

    /* set the border width of the window. */
    gtk_container_set_border_width(GTK_CONTAINER(sc.window), 8);

    g_signal_connect(G_OBJECT(sc.window), "delete_event", G_CALLBACK(delete_event), NULL);

    sc.table = gtk_table_new(2, 4, FALSE);
    gtk_table_set_row_spacings(GTK_TABLE(sc.table), 10);
    gtk_table_set_col_spacings(GTK_TABLE(sc.table), 10);

    gtk_container_add(GTK_CONTAINER(sc.window), sc.table);

    g_signal_connect(G_OBJECT(sc.window), "destroy", G_CALLBACK(destroy), NULL);

    /* Create and attach notebook */
    sc.notebook = gtk_notebook_new();
    gtk_notebook_set_tab_pos(GTK_NOTEBOOK(sc.notebook), GTK_POS_TOP);
    gtk_table_attach_defaults(GTK_TABLE(sc.table), sc.notebook, 0, 4, 0, 1);
    gtk_notebook_set_scrollable(GTK_NOTEBOOK(sc.notebook), sc.nbscroll ? TRUE : FALSE);

    gtk_widget_show(sc.notebook);

    /* Create and attach notebook pages */

    /* page 0 */
    sc.dm_frame = gtk_frame_new("Display mode");
    gtk_container_set_border_width(GTK_CONTAINER(sc.dm_frame), 8);
    gtk_widget_show(sc.dm_frame);

    if(sc.haverandr) {
#ifdef HAVE_RANDR
       sc.dm_table = gtk_table_new(6, 3, TRUE);
       gtk_table_set_row_spacings(GTK_TABLE(sc.dm_table), 4);
       gtk_table_set_col_spacings(GTK_TABLE(sc.dm_table), 0);
       gtk_container_set_border_width(GTK_CONTAINER(sc.dm_table), 10);
       gtk_container_add(GTK_CONTAINER(sc.dm_frame), sc.dm_table);
       gtk_widget_show(sc.dm_table);

       build_randr_sizes();

       sc.dm_omenu = gtk_option_menu_new();
       sc.dm_menu = gtk_menu_new();

       sc.dm_menuitem[0] = gtk_menu_item_new_with_label(sc.leaveunchangedtext);
       gtk_menu_shell_append(GTK_MENU_SHELL(sc.dm_menu), sc.dm_menuitem[0]);
       gtk_widget_show(sc.dm_menuitem[0]);
       sc.dm_menuitem[1] = gtk_menu_item_new_with_label("Resize to display mode");
       gtk_menu_shell_append(GTK_MENU_SHELL(sc.dm_menu), sc.dm_menuitem[1]);
       gtk_widget_show(sc.dm_menuitem[1]);

       for(i = 0; i < sc.rrnsize; i++) {
          sc.dm_menuitem[i+2] = gtk_menu_item_new_with_label(sc.sizetextptr[i]);
          gtk_menu_shell_append(GTK_MENU_SHELL(sc.dm_menu), sc.dm_menuitem[i+2]);
          gtk_widget_show(sc.dm_menuitem[i+2]);
       }

       gtk_option_menu_set_menu(GTK_OPTION_MENU(sc.dm_omenu), sc.dm_menu);
       gtk_table_attach_defaults(GTK_TABLE(sc.dm_table), sc.dm_omenu, 1, 3, 5, 6);

       sc.dm_size_label = gtk_label_new("Desktop size");
       gtk_table_attach_defaults(GTK_TABLE(sc.dm_table), sc.dm_size_label, 0, 1, 5, 6);
       gtk_widget_show(sc.dm_size_label);

       updateleaveunchanged();

       gtk_widget_show(sc.dm_omenu);
#endif
    }

    sc.dm_liststore = gtk_list_store_new(14, G_TYPE_STRING,   GDK_TYPE_PIXBUF, GDK_TYPE_PIXBUF,
                                             GDK_TYPE_PIXBUF, GDK_TYPE_PIXBUF, GDK_TYPE_PIXBUF,
			  		     GDK_TYPE_PIXBUF, GDK_TYPE_PIXBUF, GDK_TYPE_PIXBUF,
					     GDK_TYPE_PIXBUF, GDK_TYPE_PIXBUF, GDK_TYPE_PIXBUF,
					     GDK_TYPE_PIXBUF, G_TYPE_STRING);

    for(i = 0; i < MYMAXMODES; i++) {
       if(sc.modetextptr[i]) {
          gtk_list_store_append(sc.dm_liststore, &sc.dm_iter[i]);
          gtk_list_store_set(sc.dm_liststore, &sc.dm_iter[i], 0, sc.modetextptr[i], -1);
       }
    }

    setcrt2typesformodelist();

    sc.dm_list = gtk_tree_view_new_with_model(GTK_TREE_MODEL(sc.dm_liststore));

    sc.dm_renderer = gtk_cell_renderer_text_new();
    sc.dm_column = gtk_tree_view_column_new_with_attributes("Display mode", sc.dm_renderer, "text", 0, NULL);
    gtk_tree_view_append_column(GTK_TREE_VIEW(sc.dm_list), sc.dm_column);

    sc.dm_renderer = gtk_cell_renderer_pixbuf_new();
    sc.dm_column = gtk_tree_view_column_new_with_attributes("LCD", sc.dm_renderer, "pixbuf", 1, NULL);
    gtk_tree_view_append_column(GTK_TREE_VIEW(sc.dm_list), sc.dm_column);

    sc.dm_renderer = gtk_cell_renderer_pixbuf_new();
    sc.dm_column = gtk_tree_view_column_new_with_attributes("LCDA", sc.dm_renderer, "pixbuf", 2, NULL);
    gtk_tree_view_append_column(GTK_TREE_VIEW(sc.dm_list), sc.dm_column);

    sc.dm_renderer = gtk_cell_renderer_pixbuf_new();
    sc.dm_column = gtk_tree_view_column_new_with_attributes("TVPAL", sc.dm_renderer, "pixbuf", 3, NULL);
    gtk_tree_view_append_column(GTK_TREE_VIEW(sc.dm_list), sc.dm_column);
    sc.dm_renderer = gtk_cell_renderer_pixbuf_new();
    sc.dm_column = gtk_tree_view_column_new_with_attributes("TVNTSC", sc.dm_renderer, "pixbuf", 4, NULL);
    gtk_tree_view_append_column(GTK_TREE_VIEW(sc.dm_list), sc.dm_column);
    sc.dm_renderer = gtk_cell_renderer_pixbuf_new();
    sc.dm_column = gtk_tree_view_column_new_with_attributes("TVPALM", sc.dm_renderer, "pixbuf", 5, NULL);
    gtk_tree_view_append_column(GTK_TREE_VIEW(sc.dm_list), sc.dm_column);
    sc.dm_renderer = gtk_cell_renderer_pixbuf_new();
    sc.dm_column = gtk_tree_view_column_new_with_attributes("TVPALN", sc.dm_renderer, "pixbuf", 6, NULL);
    gtk_tree_view_append_column(GTK_TREE_VIEW(sc.dm_list), sc.dm_column);
    sc.dm_renderer = gtk_cell_renderer_pixbuf_new();
    sc.dm_column = gtk_tree_view_column_new_with_attributes("TVHIVISION", sc.dm_renderer, "pixbuf", 7, NULL);
    gtk_tree_view_append_column(GTK_TREE_VIEW(sc.dm_list), sc.dm_column);
    sc.dm_renderer = gtk_cell_renderer_pixbuf_new();
    sc.dm_column = gtk_tree_view_column_new_with_attributes("TVY525I", sc.dm_renderer, "pixbuf", 8, NULL);
    gtk_tree_view_append_column(GTK_TREE_VIEW(sc.dm_list), sc.dm_column);
    sc.dm_renderer = gtk_cell_renderer_pixbuf_new();
    sc.dm_column = gtk_tree_view_column_new_with_attributes("TVY525P", sc.dm_renderer, "pixbuf", 9, NULL);
    gtk_tree_view_append_column(GTK_TREE_VIEW(sc.dm_list), sc.dm_column);
    sc.dm_renderer = gtk_cell_renderer_pixbuf_new();
    sc.dm_column = gtk_tree_view_column_new_with_attributes("TVY750P", sc.dm_renderer, "pixbuf", 10, NULL);
    gtk_tree_view_append_column(GTK_TREE_VIEW(sc.dm_list), sc.dm_column);
    sc.dm_renderer = gtk_cell_renderer_pixbuf_new();
    sc.dm_column = gtk_tree_view_column_new_with_attributes("TVY1080I", sc.dm_renderer, "pixbuf", 11, NULL);
    gtk_tree_view_append_column(GTK_TREE_VIEW(sc.dm_list), sc.dm_column);

    sc.dm_renderer = gtk_cell_renderer_pixbuf_new();
    sc.dm_column = gtk_tree_view_column_new_with_attributes("VGA2", sc.dm_renderer, "pixbuf", 12, NULL);
    gtk_tree_view_append_column(GTK_TREE_VIEW(sc.dm_list), sc.dm_column);

    /* This last column is only for keeping gtk to make wrong alignment assumptions */
    sc.dm_renderer = gtk_cell_renderer_text_new();
    sc.dm_column = gtk_tree_view_column_new_with_attributes("pseudo", sc.dm_renderer, "text", 13, NULL);
    gtk_tree_view_append_column(GTK_TREE_VIEW(sc.dm_list), sc.dm_column);

    gtk_tree_view_set_headers_visible(GTK_TREE_VIEW(sc.dm_list), FALSE);

    sc.dm_select = gtk_tree_view_get_selection(GTK_TREE_VIEW(sc.dm_list));
    gtk_tree_selection_set_mode(sc.dm_select, GTK_SELECTION_SINGLE);

    sc.backupmodeindex = sc.currentmodeindex = setdmhistory();

    sc.dm_list_window = gtk_scrolled_window_new(NULL, NULL);
    if(sc.haverandr) {
       gtk_container_set_border_width(GTK_CONTAINER(sc.dm_list_window), 0);
    } else {
       gtk_container_set_border_width(GTK_CONTAINER(sc.dm_list_window), 10);
    }
    gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(sc.dm_list_window), GTK_POLICY_AUTOMATIC,
    									   GTK_POLICY_AUTOMATIC);

    gtk_scrolled_window_add_with_viewport(GTK_SCROLLED_WINDOW(sc.dm_list_window), sc.dm_list);

    if(sc.haverandr) {
       gtk_table_attach_defaults(GTK_TABLE(sc.dm_table), sc.dm_list_window, 0, 3, 0, 5);
    } else {
       gtk_container_add(GTK_CONTAINER(sc.dm_frame), sc.dm_list_window);
    }

    gtk_widget_show(sc.dm_list);
    gtk_widget_show(sc.dm_list_window);

    sc.dm_label = xpm_label_box(icon_mode, "Mode", NULL, 1);
    gtk_notebook_append_page(GTK_NOTEBOOK(sc.notebook), sc.dm_frame, sc.dm_label);

    /* page 1 */
    sc.c1_frame = gtk_frame_new("CRT1 settings");
    gtk_container_set_border_width(GTK_CONTAINER(sc.c1_frame), 8);
    gtk_widget_show(sc.c1_frame);

    i = 5; j = 6; k = 15;
    if(!(sc.sdflags & SiS_SD_ISDUALHEAD)) {
       if(sc.currentvbflags & VB_VIDEOBRIDGE) {
          if((sc.sdflags & SiS_SD_SUPPORTLCDA) && (sc.sdflags & SiS_SD_SUPPORTSCALE)) {
             i = 6; j = 4; k = 5;
	  }
       }
    }
    sc.c1_table = gtk_table_new(i, 3, TRUE);
    gtk_table_set_row_spacings(GTK_TABLE(sc.c1_table), j);
    gtk_table_set_col_spacings(GTK_TABLE(sc.c1_table), 10);

    gtk_container_set_border_width(GTK_CONTAINER(sc.c1_table), k);

    gtk_container_add(GTK_CONTAINER(sc.c1_frame), sc.c1_table);
    gtk_widget_show(sc.c1_table);

    if(!(sc.sdflags & SiS_SD_ISDUALHEAD)) {

       if(sc.currentvbflags & VB_VIDEOBRIDGE) {

          sc.c1_iconon = gdk_pixmap_create_from_xpm_d(sc.window->window, &sc.c1_maskon, NULL, crt1on_xpm);
    	  sc.c1_imageon = gtk_pixmap_new(sc.c1_iconon, sc.c1_maskon);
	  sc.c1_iconoff = gdk_pixmap_create_from_xpm_d(sc.window->window, &sc.c1_maskoff, NULL, crt1off_xpm);
    	  sc.c1_imageoff = gtk_pixmap_new(sc.c1_iconoff, sc.c1_maskoff);
	  sc.c1_iconlcda = gdk_pixmap_create_from_xpm_d(sc.window->window, &sc.c1_masklcda, NULL, crt1lcda_xpm);
    	  sc.c1_imagelcda = gtk_pixmap_new(sc.c1_iconlcda, sc.c1_masklcda);
    	  gtk_table_attach_defaults(GTK_TABLE(sc.c1_table), sc.c1_imageon, 0, 3, 0, 4);
	  gtk_table_attach_defaults(GTK_TABLE(sc.c1_table), sc.c1_imageoff, 0, 3, 0, 4);
	  gtk_table_attach_defaults(GTK_TABLE(sc.c1_table), sc.c1_imagelcda, 0, 3, 0, 4);
    	  gtk_widget_hide(sc.c1_imageon);
	  gtk_widget_hide(sc.c1_imageoff);
	  gtk_widget_hide(sc.c1_imagelcda);

	  if(sc.sdflags & SiS_SD_SUPPORTLCDA) {
	     sc.c1_label_onoff = gtk_label_new("CRT1 Type");
	  } else {
             sc.c1_label_onoff = gtk_label_new("Switch CRT1");
	  }
          gtk_table_attach_defaults(GTK_TABLE(sc.c1_table), sc.c1_label_onoff, 0, 1, 4, 5);
          gtk_widget_show(sc.c1_label_onoff);

          sc.c1_omenu = gtk_option_menu_new();

          sc.c1_menu = gtk_menu_new();

	  sc.c1_menuitemoff = gtk_menu_item_new_with_label("OFF");
	  if(sc.sdflags & SiS_SD_SUPPORTLCDA) {
	     sc.c1_menuitemon   = gtk_menu_item_new_with_label("VGA");
	     sc.c1_menuitemlcda = gtk_menu_item_new_with_label("LCD via CRT1");
	  } else {
	     sc.c1_menuitemon   = gtk_menu_item_new_with_label("ON");
	  }

          gtk_menu_shell_append(GTK_MENU_SHELL(sc.c1_menu), sc.c1_menuitemoff);
          gtk_menu_shell_append(GTK_MENU_SHELL(sc.c1_menu), sc.c1_menuitemon);
	  if(sc.sdflags & SiS_SD_SUPPORTLCDA) {
	     gtk_menu_shell_append(GTK_MENU_SHELL(sc.c1_menu), sc.c1_menuitemlcda);
	  }

          gtk_widget_show(sc.c1_menuitemon);
          gtk_widget_show(sc.c1_menuitemoff);
	  if(sc.sdflags & SiS_SD_SUPPORTLCDA) {
	     gtk_widget_show(sc.c1_menuitemlcda);
	  }

          gtk_option_menu_set_menu(GTK_OPTION_MENU(sc.c1_omenu), sc.c1_menu);

	  if((sc.sdflags & SiS_SD_SUPPORTLCDA) && (sc.sdflags & SiS_SD_SUPPORTSCALE)) {
	     sc.c1_label_lcds = gtk_label_new("LCD scaling");
             gtk_table_attach_defaults(GTK_TABLE(sc.c1_table), sc.c1_label_lcds, 0, 1, 5, 6);

             sc.c1_omenuls = gtk_option_menu_new();
             sc.c1_menuls = gtk_menu_new();
             sc.c1_menuitemlsd = gtk_menu_item_new_with_label("Default");
             gtk_menu_shell_append(GTK_MENU_SHELL(sc.c1_menuls), sc.c1_menuitemlsd);
             gtk_widget_show(sc.c1_menuitemlsd);
	     sc.c1_menuitemlso = gtk_menu_item_new_with_label("On");
             gtk_menu_shell_append(GTK_MENU_SHELL(sc.c1_menuls), sc.c1_menuitemlso);
             gtk_widget_show(sc.c1_menuitemlso);
	     sc.c1_menuitemlsf = gtk_menu_item_new_with_label("Off");
             gtk_menu_shell_append(GTK_MENU_SHELL(sc.c1_menuls), sc.c1_menuitemlsf);
             gtk_widget_show(sc.c1_menuitemlsf);
	     gtk_option_menu_set_menu(GTK_OPTION_MENU(sc.c1_omenuls), sc.c1_menuls);
	     gtk_table_attach_defaults(GTK_TABLE(sc.c1_table), sc.c1_omenuls, 1, 2, 5, 6);

	     if(sc.sdflags & SiS_SD_SUPPORTCENTER) {
	        sc.c1_omenulcc = gtk_option_menu_new();
                sc.c1_menulcc = gtk_menu_new();
                sc.c1_menuitemlccd = gtk_menu_item_new_with_label("Default");
                gtk_menu_shell_append(GTK_MENU_SHELL(sc.c1_menulcc), sc.c1_menuitemlccd);
                gtk_widget_show(sc.c1_menuitemlccd);
	        sc.c1_menuitemlccc = gtk_menu_item_new_with_label("Don't center");
                gtk_menu_shell_append(GTK_MENU_SHELL(sc.c1_menulcc), sc.c1_menuitemlccc);
                gtk_widget_show(sc.c1_menuitemlccc);
	        sc.c1_menuitemlccp = gtk_menu_item_new_with_label("Center");
                gtk_menu_shell_append(GTK_MENU_SHELL(sc.c1_menulcc), sc.c1_menuitemlccp);
                gtk_widget_show(sc.c1_menuitemlccp);
	        gtk_option_menu_set_menu(GTK_OPTION_MENU(sc.c1_omenulcc), sc.c1_menulcc);
	        gtk_table_attach_defaults(GTK_TABLE(sc.c1_table), sc.c1_omenulcc, 2, 3, 5, 6);
	     }
	  }

          setcrt1history(sc.currentvbflags);
	  gtk_table_attach_defaults(GTK_TABLE(sc.c1_table), sc.c1_omenu, 1, 3, 4, 5);

	  if((sc.sdflags & SiS_SD_SUPPORTLCDA) && (sc.sdflags & SiS_SD_SUPPORTSCALE)) {
	     setlcdhistory();
	     showlcdmenu(1);
	  }

          g_signal_connect(G_OBJECT(sc.c1_menuitemon), "activate", G_CALLBACK(c1_menu_changed), NULL);
          g_signal_connect(G_OBJECT(sc.c1_menuitemoff), "activate", G_CALLBACK(c1_menu_changed), NULL);
	  if(sc.sdflags & SiS_SD_SUPPORTLCDA) {
	     g_signal_connect(G_OBJECT(sc.c1_menuitemlcda), "activate", G_CALLBACK(c1_menu_changed), NULL);
	  }

	  if((sc.sdflags & SiS_SD_SUPPORTLCDA) &&
	     (sc.sdflags & SiS_SD_SUPPORTSCALE) && (sc.sdflags & SiS_SD_SUPPORTCENTER))  {
	     g_signal_connect(G_OBJECT(sc.c1_menuitemlsd), "activate", G_CALLBACK(c1_omenuls_changed), NULL);
	     g_signal_connect(G_OBJECT(sc.c1_menuitemlso), "activate", G_CALLBACK(c1_omenuls_changed), NULL);
	     g_signal_connect(G_OBJECT(sc.c1_menuitemlsf), "activate", G_CALLBACK(c1_omenuls_changed), NULL);
	  }

          gtk_widget_show(sc.c1_omenu);

       } else {

          sc.c1_label_nodh = gtk_label_new("No videobridge present");
          gtk_widget_show(sc.c1_label_nodh);
          gtk_table_attach_defaults(GTK_TABLE(sc.c1_table), sc.c1_label_nodh, 0, 3, 2, 3);

       }

    } else {

       sc.c1_label_nodh = gtk_label_new("Not supported in Dual Head mode");
       gtk_widget_show(sc.c1_label_nodh);
       gtk_table_attach_defaults(GTK_TABLE(sc.c1_table), sc.c1_label_nodh, 0, 3, 2, 3);

    }

    sc.c1_label = xpm_label_box(icon_crt1, "CRT1", NULL, 1);
    gtk_notebook_append_page(GTK_NOTEBOOK(sc.notebook), sc.c1_frame, sc.c1_label);

    /* page 2 */
    sc.om_frame = gtk_frame_new("CRT2 settings");
    gtk_container_set_border_width(GTK_CONTAINER(sc.om_frame), 8);
    gtk_widget_show(sc.om_frame);

    sc.om_table = gtk_table_new(6, 3, TRUE);
    gtk_table_set_row_spacings(GTK_TABLE(sc.om_table), 2);
    gtk_table_set_col_spacings(GTK_TABLE(sc.om_table), 10);

    gtk_container_set_border_width(GTK_CONTAINER(sc.om_table), 8);

    gtk_container_add(GTK_CONTAINER(sc.om_frame), sc.om_table);
    gtk_widget_show(sc.om_table);

    if(!(sc.sdflags & SiS_SD_ISDUALHEAD)) {

       if(sc.currentvbflags & VB_VIDEOBRIDGE) {

          sc.om_label_type = gtk_label_new("CRT2 type");
          gtk_table_attach_defaults(GTK_TABLE(sc.om_table), sc.om_label_type, 0, 1, 3, 4);
          gtk_widget_show(sc.om_label_type);

	  sc.om_ilcd = gdk_pixmap_create_from_xpm_d(sc.window->window, &sc.om_mlcd, NULL, crt2lcd_xpm);
    	  sc.om_lcd = gtk_pixmap_new(sc.om_ilcd, sc.om_mlcd);
	  sc.om_itv = gdk_pixmap_create_from_xpm_d(sc.window->window, &sc.om_mtv, NULL, crt2tv_xpm);
    	  sc.om_tv = gtk_pixmap_new(sc.om_itv, sc.om_mtv);
	  sc.om_iypbpr = gdk_pixmap_create_from_xpm_d(sc.window->window, &sc.om_mypbpr, NULL, crt2ypbpr_xpm);
    	  sc.om_ypbpr = gtk_pixmap_new(sc.om_iypbpr, sc.om_mypbpr);
	  sc.om_itvs = gdk_pixmap_create_from_xpm_d(sc.window->window, &sc.om_mtvs, NULL, crt2svideo_xpm);
    	  sc.om_tvs = gtk_pixmap_new(sc.om_itvs, sc.om_mtvs);
	  sc.om_itvc = gdk_pixmap_create_from_xpm_d(sc.window->window, &sc.om_mtvc, NULL, crt2cvbs_xpm);
    	  sc.om_tvc = gtk_pixmap_new(sc.om_itvc, sc.om_mtvc);
	  sc.om_ivga = gdk_pixmap_create_from_xpm_d(sc.window->window, &sc.om_mvga, NULL, crt2vga_xpm);
    	  sc.om_vga = gtk_pixmap_new(sc.om_ivga, sc.om_mvga);
	  gtk_table_attach_defaults(GTK_TABLE(sc.om_table), sc.om_lcd, 0, 3, 0, 3);
	  gtk_table_attach_defaults(GTK_TABLE(sc.om_table), sc.om_tv, 0, 3, 0, 3);
	  gtk_table_attach_defaults(GTK_TABLE(sc.om_table), sc.om_tvs, 0, 3, 0, 3);
	  gtk_table_attach_defaults(GTK_TABLE(sc.om_table), sc.om_tvc, 0, 3, 0, 3);
	  gtk_table_attach_defaults(GTK_TABLE(sc.om_table), sc.om_ypbpr, 0, 3, 0, 3);
	  gtk_table_attach_defaults(GTK_TABLE(sc.om_table), sc.om_vga, 0, 3, 0, 3);
	  gtk_widget_hide(sc.om_lcd);
	  gtk_widget_hide(sc.om_tv);
	  gtk_widget_hide(sc.om_tvs);
	  gtk_widget_hide(sc.om_tvc);
	  gtk_widget_hide(sc.om_ypbpr);
	  gtk_widget_hide(sc.om_vga);

          sc.om_omenu = gtk_option_menu_new();

          sc.om_menu = gtk_menu_new();

          sc.om_menuitemlcd   = gtk_menu_item_new_with_label("LCD / DVI-D");
          sc.om_menuitemsvhs  = gtk_menu_item_new_with_label("TV (S-Video)");
          sc.om_menuitemcvbs  = gtk_menu_item_new_with_label("TV (Composite)");
	  sc.om_menuitemcvsv  = gtk_menu_item_new_with_label("TV (Composite + S-Video)");
          sc.om_menuitemtv    = gtk_menu_item_new_with_label("TV");
	  sc.om_menuitemhivi  = gtk_menu_item_new_with_label("HDTV (HiVision 1080i)");
	  sc.om_menuitemy525i = gtk_menu_item_new_with_label("HDTV (YPbPr 480i)");
	  sc.om_menuitemy525p = gtk_menu_item_new_with_label("HDTV (YPbPr 480p)");
	  sc.om_menuitemy750p = gtk_menu_item_new_with_label("HDTV (YPbPr 720p)");
	  sc.om_menuitemy1080i= gtk_menu_item_new_with_label("HDTV (YPbPr 1080i)");
          sc.om_menuitemvga2  = gtk_menu_item_new_with_label("VGA2 / DVI-A");
	  sc.om_menuitemscart = gtk_menu_item_new_with_label("TV (SCART PAL)");
          sc.om_menuitemoff   = gtk_menu_item_new_with_label("OFF");

          gtk_menu_shell_append(GTK_MENU_SHELL(sc.om_menu), sc.om_menuitemlcd);
          gtk_menu_shell_append(GTK_MENU_SHELL(sc.om_menu), sc.om_menuitemsvhs);
          gtk_menu_shell_append(GTK_MENU_SHELL(sc.om_menu), sc.om_menuitemcvbs);
	  gtk_menu_shell_append(GTK_MENU_SHELL(sc.om_menu), sc.om_menuitemcvsv);
          gtk_menu_shell_append(GTK_MENU_SHELL(sc.om_menu), sc.om_menuitemtv);
	  gtk_menu_shell_append(GTK_MENU_SHELL(sc.om_menu), sc.om_menuitemhivi);
	  gtk_menu_shell_append(GTK_MENU_SHELL(sc.om_menu), sc.om_menuitemy525i);
	  gtk_menu_shell_append(GTK_MENU_SHELL(sc.om_menu), sc.om_menuitemy525p);
	  gtk_menu_shell_append(GTK_MENU_SHELL(sc.om_menu), sc.om_menuitemy750p);
	  gtk_menu_shell_append(GTK_MENU_SHELL(sc.om_menu), sc.om_menuitemy1080i);
          gtk_menu_shell_append(GTK_MENU_SHELL(sc.om_menu), sc.om_menuitemvga2);
	  gtk_menu_shell_append(GTK_MENU_SHELL(sc.om_menu), sc.om_menuitemscart);
          gtk_menu_shell_append(GTK_MENU_SHELL(sc.om_menu), sc.om_menuitemoff);

          showcrt2typemenu(sc.detecteddevices, TRUE);

          gtk_option_menu_set_menu(GTK_OPTION_MENU(sc.om_omenu), sc.om_menu);

          setcrt2typehistory(sc.currentvbflags);

          gtk_table_attach_defaults(GTK_TABLE(sc.om_table), sc.om_omenu, 1, 3, 3, 4);

          sc.om_label_tvtype = gtk_label_new("TV format");
          gtk_table_attach_defaults(GTK_TABLE(sc.om_table), sc.om_label_tvtype, 0, 1, 5, 6);

          sc.om_omenut = gtk_option_menu_new();

          sc.om_menut = gtk_menu_new();

          sc.om_menuitempal = gtk_menu_item_new_with_label("PAL");
          gtk_menu_shell_append(GTK_MENU_SHELL(sc.om_menut), sc.om_menuitempal);
          gtk_widget_show(sc.om_menuitempal);

          sc.om_menuitemntsc = gtk_menu_item_new_with_label("NTSC");
          gtk_menu_shell_append(GTK_MENU_SHELL(sc.om_menut), sc.om_menuitemntsc);
          gtk_widget_show(sc.om_menuitemntsc);

          if(sc.sdflags & SiS_SD_SUPPORTPALMN) {
             sc.om_menuitempalm = gtk_menu_item_new_with_label("PAL-M");
             gtk_menu_shell_append(GTK_MENU_SHELL(sc.om_menut), sc.om_menuitempalm);
             gtk_widget_show(sc.om_menuitempalm);
             sc.om_menuitempaln = gtk_menu_item_new_with_label("PAL-N");
             gtk_menu_shell_append(GTK_MENU_SHELL(sc.om_menut), sc.om_menuitempaln);
             gtk_widget_show(sc.om_menuitempaln);
          }
	  if(sc.sdflags & SiS_SD_SUPPORTNTSCJ) {
             sc.om_menuitemntscj = gtk_menu_item_new_with_label("NTSC-J");
             gtk_menu_shell_append(GTK_MENU_SHELL(sc.om_menut), sc.om_menuitemntscj);
             gtk_widget_show(sc.om_menuitemntscj);
	  }

          gtk_option_menu_set_menu(GTK_OPTION_MENU(sc.om_omenut), sc.om_menut);

          gtk_table_attach_defaults(GTK_TABLE(sc.om_table), sc.om_omenut, 1,
	       (sc.currentvbflags & VB_CHRONTEL) ? 2 : 3, 5, 6);

	  if(sc.sdflags & SiS_SD_SUPPORTYPBPRAR) {

	     sc.om_label_tvas = gtk_label_new("Aspect ratio");
             gtk_table_attach_defaults(GTK_TABLE(sc.om_table), sc.om_label_tvas, 0, 1, 5, 6);

             sc.om_omenuas = gtk_option_menu_new();

             sc.om_menuas = gtk_menu_new();

             sc.om_menuitem43 = gtk_menu_item_new_with_label("4:3");
             gtk_menu_shell_append(GTK_MENU_SHELL(sc.om_menuas), sc.om_menuitem43);
             gtk_widget_show(sc.om_menuitem43);
	     sc.om_menuitem43lb = gtk_menu_item_new_with_label("4:3 Letterbox");
             gtk_menu_shell_append(GTK_MENU_SHELL(sc.om_menuas), sc.om_menuitem43lb);
             gtk_widget_show(sc.om_menuitem43lb);
	     sc.om_menuitem169 = gtk_menu_item_new_with_label("16:9");
             gtk_menu_shell_append(GTK_MENU_SHELL(sc.om_menuas), sc.om_menuitem169);
             gtk_widget_show(sc.om_menuitem169);

	     gtk_option_menu_set_menu(GTK_OPTION_MENU(sc.om_omenuas), sc.om_menuas);
	     gtk_table_attach_defaults(GTK_TABLE(sc.om_table), sc.om_omenuas, 1, 3, 5, 6);

	  }

	  if(sc.currentvbflags & VB_CHRONTEL) {
	     sc.om_omenuo = gtk_option_menu_new();
             sc.om_menuo = gtk_menu_new();

             sc.om_menuitembios = gtk_menu_item_new_with_label("BIOS");
             gtk_menu_shell_append(GTK_MENU_SHELL(sc.om_menuo), sc.om_menuitembios);
             gtk_widget_show(sc.om_menuitembios);

             sc.om_menuitemunder = gtk_menu_item_new_with_label("Underscan");
             gtk_menu_shell_append(GTK_MENU_SHELL(sc.om_menuo), sc.om_menuitemunder);
             gtk_widget_show(sc.om_menuitemunder);

	     sc.om_menuitemover = gtk_menu_item_new_with_label("Overscan");
             gtk_menu_shell_append(GTK_MENU_SHELL(sc.om_menuo), sc.om_menuitemover);
             gtk_widget_show(sc.om_menuitemover);

	     sc.om_menuitemsover = gtk_menu_item_new_with_label("S-Overscan");
             gtk_menu_shell_append(GTK_MENU_SHELL(sc.om_menuo), sc.om_menuitemsover);
             gtk_widget_show(sc.om_menuitemsover);

	     showtvovermenu(sc.currentvbflags);

	     gtk_option_menu_set_menu(GTK_OPTION_MENU(sc.om_omenuo), sc.om_menuo);

             settvoverhistory();

	     gtk_table_attach_defaults(GTK_TABLE(sc.om_table), sc.om_omenuo, 2, 3, 5, 6);
	  }

	  if(sc.sdflags & SiS_SD_SUPPORTSCALE) {
	     sc.om_label_lcds = gtk_label_new("LCD scaling");
             gtk_table_attach_defaults(GTK_TABLE(sc.om_table), sc.om_label_lcds, 0, 1, 5, 6);

             sc.om_omenuls = gtk_option_menu_new();
             sc.om_menuls = gtk_menu_new();
             sc.om_menuitemlsd = gtk_menu_item_new_with_label("Default");
             gtk_menu_shell_append(GTK_MENU_SHELL(sc.om_menuls), sc.om_menuitemlsd);
             gtk_widget_show(sc.om_menuitemlsd);
	     sc.om_menuitemlso = gtk_menu_item_new_with_label("On");
             gtk_menu_shell_append(GTK_MENU_SHELL(sc.om_menuls), sc.om_menuitemlso);
             gtk_widget_show(sc.om_menuitemlso);
	     sc.om_menuitemlsf = gtk_menu_item_new_with_label("Off");
             gtk_menu_shell_append(GTK_MENU_SHELL(sc.om_menuls), sc.om_menuitemlsf);
             gtk_widget_show(sc.om_menuitemlsf);
	     gtk_option_menu_set_menu(GTK_OPTION_MENU(sc.om_omenuls), sc.om_menuls);
	     gtk_table_attach_defaults(GTK_TABLE(sc.om_table), sc.om_omenuls, 1, 2, 5, 6);

	     if(sc.sdflags & SiS_SD_SUPPORTCENTER) {
	        sc.om_omenulcc = gtk_option_menu_new();
                sc.om_menulcc = gtk_menu_new();
                sc.om_menuitemlccd = gtk_menu_item_new_with_label("Default");
                gtk_menu_shell_append(GTK_MENU_SHELL(sc.om_menulcc), sc.om_menuitemlccd);
                gtk_widget_show(sc.om_menuitemlccd);
	        sc.om_menuitemlccc = gtk_menu_item_new_with_label("Don't center");
                gtk_menu_shell_append(GTK_MENU_SHELL(sc.om_menulcc), sc.om_menuitemlccc);
                gtk_widget_show(sc.om_menuitemlccc);
	        sc.om_menuitemlccp = gtk_menu_item_new_with_label("Center");
                gtk_menu_shell_append(GTK_MENU_SHELL(sc.om_menulcc), sc.om_menuitemlccp);
                gtk_widget_show(sc.om_menuitemlccp);
	        gtk_option_menu_set_menu(GTK_OPTION_MENU(sc.om_omenulcc), sc.om_menulcc);
	        gtk_table_attach_defaults(GTK_TABLE(sc.om_table), sc.om_omenulcc, 2, 3, 5, 6);
	     }
	  }

	  showtvtypemenu(sc.currentvbflags);
	  settvtypehistory(sc.currentvbflags);

	  if(sc.sdflags & SiS_SD_SUPPORTYPBPRAR) {
	     showtvasmenu(sc.currentvbflags);
	     settvashistory(sc.currentvbflags);
	  }

	  if(sc.sdflags & SiS_SD_SUPPORTSCALE) {
	     showlcdmenu(2);
	     setlcdhistory();
	  }

          sc.om_check = gtk_check_button_new_with_label("Show detected devices only");
          gtk_table_attach_defaults(GTK_TABLE(sc.om_table), sc.om_check, 1, 3, 4, 5);

          gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(sc.om_check), TRUE);

          g_signal_connect(G_OBJECT(sc.om_check), "toggled", G_CALLBACK (om_check_toggled), NULL);

          g_signal_connect(G_OBJECT(sc.om_menuitemlcd),   "activate", G_CALLBACK(om_menu_changed), NULL);
          g_signal_connect(G_OBJECT(sc.om_menuitemsvhs),  "activate", G_CALLBACK(om_menu_changed), NULL);
          g_signal_connect(G_OBJECT(sc.om_menuitemcvbs),  "activate", G_CALLBACK(om_menu_changed), NULL);
	  g_signal_connect(G_OBJECT(sc.om_menuitemcvsv),  "activate", G_CALLBACK(om_menu_changed), NULL);
          g_signal_connect(G_OBJECT(sc.om_menuitemtv),    "activate", G_CALLBACK(om_menu_changed), NULL);
	  if(sc.sdflags & SiS_SD_SUPPORTHIVISION) {
	     g_signal_connect(G_OBJECT(sc.om_menuitemhivi),  "activate", G_CALLBACK(om_menu_changed), NULL);
	  }
	  if(sc.sdflags & SiS_SD_SUPPORTYPBPR) {
	     g_signal_connect(G_OBJECT(sc.om_menuitemy525i), "activate", G_CALLBACK(om_menu_changed), NULL);
	     g_signal_connect(G_OBJECT(sc.om_menuitemy525p), "activate", G_CALLBACK(om_menu_changed), NULL);
	     g_signal_connect(G_OBJECT(sc.om_menuitemy750p), "activate", G_CALLBACK(om_menu_changed), NULL);
	     g_signal_connect(G_OBJECT(sc.om_menuitemy1080i), "activate", G_CALLBACK(om_menu_changed), NULL);
	  }
          g_signal_connect(G_OBJECT(sc.om_menuitemvga2),  "activate", G_CALLBACK(om_menu_changed), NULL);
	  g_signal_connect(G_OBJECT(sc.om_menuitemscart), "activate", G_CALLBACK(om_menu_changed), NULL);
          g_signal_connect(G_OBJECT(sc.om_menuitemoff),   "activate", G_CALLBACK(om_menu_changed), NULL);

	  if((sc.sdflags & SiS_SD_SUPPORTSCALE) && (sc.sdflags & SiS_SD_SUPPORTCENTER))  {
	     g_signal_connect(G_OBJECT(sc.om_menuitemlsd), "activate", G_CALLBACK(om_omenuls_changed), NULL);
	     g_signal_connect(G_OBJECT(sc.om_menuitemlso), "activate", G_CALLBACK(om_omenuls_changed), NULL);
	     g_signal_connect(G_OBJECT(sc.om_menuitemlsf), "activate", G_CALLBACK(om_omenuls_changed), NULL);
	  }

	  g_signal_connect(G_OBJECT(sc.om_menuitempal),   "activate", G_CALLBACK(om_menut_changed), NULL);
          g_signal_connect(G_OBJECT(sc.om_menuitemntsc),  "activate", G_CALLBACK(om_menut_changed), NULL);
	  if(sc.sdflags & SiS_SD_SUPPORTPALMN) {
	     g_signal_connect(G_OBJECT(sc.om_menuitempalm), "activate", G_CALLBACK(om_menut_changed), NULL);
             g_signal_connect(G_OBJECT(sc.om_menuitempaln), "activate", G_CALLBACK(om_menut_changed), NULL);
	  }
	  if(sc.sdflags & SiS_SD_SUPPORTNTSCJ) {
	     g_signal_connect(G_OBJECT(sc.om_menuitemntscj), "activate", G_CALLBACK(om_menut_changed), NULL);
	  }


          gtk_widget_show(sc.om_omenu);
          gtk_widget_show(sc.om_check);

       } else {

          sc.om_label_nodh = gtk_label_new("No videobridge present");
          gtk_widget_show(sc.om_label_nodh);
          gtk_table_attach_defaults(GTK_TABLE(sc.om_table), sc.om_label_nodh, 0, 3, 2, 3);

       }

    } else {

       sc.om_label_nodh = gtk_label_new("Not supported in Dual Head mode");
       gtk_widget_show(sc.om_label_nodh);
       gtk_table_attach_defaults(GTK_TABLE(sc.om_table), sc.om_label_nodh, 0, 3, 2, 3);

    }

    sc.om_label = xpm_label_box(icon_crt2, "CRT2", NULL, 1);
    gtk_notebook_append_page(GTK_NOTEBOOK(sc.notebook), sc.om_frame, sc.om_label);

    /* page 3 */
    sc.ga_frame = gtk_frame_new("Gamma correction");
    gtk_container_set_border_width(GTK_CONTAINER(sc.ga_frame), 8);
    gtk_widget_show(sc.ga_frame);

    sc.ga_table = gtk_table_new(6, 8, FALSE);
    gtk_table_set_row_spacings(GTK_TABLE(sc.ga_table), 3);
    gtk_table_set_col_spacings(GTK_TABLE(sc.ga_table), 2);

    gtk_container_set_border_width(GTK_CONTAINER(sc.ga_table), 10);

    gtk_container_add(GTK_CONTAINER(sc.ga_frame), sc.ga_table);
    gtk_widget_show(sc.ga_table);

    if(sc.sdflags & SiS_SD_ISDHXINERAMA) {

       char buffer[64];

       sprintf(buffer, "Setting for Screen %d", sc.myscreen);
       sc.ga_label_enable = gtk_label_new(buffer);
       gtk_table_attach_defaults(GTK_TABLE(sc.ga_table), sc.ga_label_enable, 1, 5, 1, 2);
       gtk_widget_show(sc.ga_label_enable);

    } else {

       sc.ga_plabel1 = gtk_label_new(" ");
       sc.ga_plabel2 = gtk_label_new(" ");
       gtk_table_attach_defaults(GTK_TABLE(sc.ga_table), sc.ga_plabel1, 1, 4, 0, 1);
       gtk_table_attach_defaults(GTK_TABLE(sc.ga_table), sc.ga_plabel2, 1, 4, 2, 3);
       gtk_widget_show(sc.ga_plabel1);
       gtk_widget_show(sc.ga_plabel2);

       sc.ga_subtable = gtk_table_new(1, 4, FALSE);
       gtk_table_set_row_spacings(GTK_TABLE(sc.ga_subtable), 0);
       gtk_table_set_col_spacings(GTK_TABLE(sc.ga_subtable), 5);
       gtk_container_set_border_width(GTK_CONTAINER(sc.ga_subtable), 0);

       sc.ga_label_enable = gtk_label_new("Enable for");
       gtk_table_attach_defaults(GTK_TABLE(sc.ga_subtable), sc.ga_label_enable, 0, 1, 0, 1);
       gtk_widget_show(sc.ga_label_enable);

       sc.ga_check1 = gtk_check_button_new_with_label("CRT1");
       gtk_table_attach_defaults(GTK_TABLE(sc.ga_subtable), sc.ga_check1, 2, 3, 0, 1);
       gtk_widget_show(sc.ga_check1);

       sc.ga_check2 = gtk_check_button_new_with_label("CRT2");
       gtk_table_attach_defaults(GTK_TABLE(sc.ga_subtable), sc.ga_check2, 3, 4, 0, 1);
       gtk_widget_show(sc.ga_check2);

       gtk_table_attach_defaults(GTK_TABLE(sc.ga_table), sc.ga_subtable, 1, 4, 1, 2);
       gtk_widget_show(sc.ga_subtable);

       setgammastatus();

       if(sc.sdflags & SiS_SD_ISDUALHEAD) {
          if(sc.sdflags & SiS_SD_ISDHSECONDHEAD) {
             gtk_widget_set_sensitive(sc.ga_check2, FALSE);
          } else {
             gtk_widget_set_sensitive(sc.ga_check1, FALSE);
          }
       }

       if(!(sc.currentvbflags & VB_SISBRIDGE)) {
          gtk_widget_set_sensitive(sc.ga_check2, FALSE);
       }

       if(sc.sdflags & SiS_SD_ISDEPTH8) {
          gtk_widget_set_sensitive(sc.ga_check1, FALSE);
          gtk_widget_set_sensitive(sc.ga_check2, FALSE);
       } else {
          g_signal_connect(G_OBJECT(sc.ga_check1), "toggled", G_CALLBACK (ga_check1_toggled), NULL);
          g_signal_connect(G_OBJECT(sc.ga_check2), "toggled", G_CALLBACK (ga_check2_toggled), NULL);
       }

    }

    sc.ga_icon = gdk_pixmap_create_from_xpm_d(sc.window->window, &sc.ga_mask, NULL, gammaicon_xpm);
    sc.ga_image = gtk_pixmap_new(sc.ga_icon, sc.ga_mask);
    gtk_table_attach_defaults(GTK_TABLE(sc.ga_table), sc.ga_image, 0, 1, 0, 6);
    gtk_widget_show(sc.ga_image);

    sc.ga_label_gamma = gtk_label_new("Gamma");
    gtk_table_attach_defaults(GTK_TABLE(sc.ga_table), sc.ga_label_gamma, 1, 2, 3, 4);
    gtk_widget_show(sc.ga_label_gamma);
#ifdef HAVE_MATH_H
    sc.ga_label_bri = gtk_label_new("Brightness");
    gtk_table_attach_defaults(GTK_TABLE(sc.ga_table), sc.ga_label_bri, 1, 2, 4, 5);
    gtk_widget_show(sc.ga_label_bri);
    sc.ga_label_prebri = gtk_label_new("Pre-Brightness");
    gtk_table_attach_defaults(GTK_TABLE(sc.ga_table), sc.ga_label_prebri, 1, 2, 5, 6);
    gtk_widget_show(sc.ga_label_prebri);
#endif

    sc.ga_label_red = gtk_label_new("R");
    gtk_table_attach_defaults(GTK_TABLE(sc.ga_table), sc.ga_label_red, 2, 3, 3, 4);
    gtk_widget_show(sc.ga_label_red);
    sc.ga_label_green = gtk_label_new("G");
    gtk_table_attach_defaults(GTK_TABLE(sc.ga_table), sc.ga_label_green, 4, 5, 3, 4);
    gtk_widget_show(sc.ga_label_green);
    sc.ga_label_blue = gtk_label_new("B");
    gtk_table_attach_defaults(GTK_TABLE(sc.ga_table), sc.ga_label_blue, 6, 7, 3, 4);
    gtk_widget_show(sc.ga_label_blue);
#ifdef HAVE_MATH_H
    sc.ga_label_red1 = gtk_label_new("R");
    gtk_table_attach_defaults(GTK_TABLE(sc.ga_table), sc.ga_label_red1, 2, 3, 4, 5);
    gtk_widget_show(sc.ga_label_red1);
    sc.ga_label_green1 = gtk_label_new("G");
    gtk_table_attach_defaults(GTK_TABLE(sc.ga_table), sc.ga_label_green1, 4, 5, 4, 5);
    gtk_widget_show(sc.ga_label_green1);
    sc.ga_label_blue1 = gtk_label_new("B");
    gtk_table_attach_defaults(GTK_TABLE(sc.ga_table), sc.ga_label_blue1, 6, 7, 4, 5);
    gtk_widget_show(sc.ga_label_blue1);
    sc.ga_label_red2 = gtk_label_new("R");
    gtk_table_attach_defaults(GTK_TABLE(sc.ga_table), sc.ga_label_red2, 2, 3, 5, 6);
    gtk_widget_show(sc.ga_label_red2);
    sc.ga_label_green2 = gtk_label_new("G");
    gtk_table_attach_defaults(GTK_TABLE(sc.ga_table), sc.ga_label_green2, 4, 5, 5, 6);
    gtk_widget_show(sc.ga_label_green2);
    sc.ga_label_blue2 = gtk_label_new("B");
    gtk_table_attach_defaults(GTK_TABLE(sc.ga_table), sc.ga_label_blue2, 6, 7, 5, 6);
    gtk_widget_show(sc.ga_label_blue2);
#endif

    sc.ga_spin1_adj = (GtkAdjustment *)gtk_adjustment_new(1.0, 0.1, 10.0, 0.1, 1.0, 0.0);
    sc.ga_spin2_adj = (GtkAdjustment *)gtk_adjustment_new(1.0, 0.1, 10.0, 0.1, 1.0, 0.0);
    sc.ga_spin3_adj = (GtkAdjustment *)gtk_adjustment_new(1.0, 0.1, 10.0, 0.1, 1.0, 0.0);
#ifdef HAVE_MATH_H
    sc.ga_spin4_adj = (GtkAdjustment *)gtk_adjustment_new(1.0, 0.1, 10.0, 0.01, 0.1, 0.0);
    sc.ga_spin5_adj = (GtkAdjustment *)gtk_adjustment_new(1.0, 0.1, 10.0, 0.01, 0.1, 0.0);
    sc.ga_spin6_adj = (GtkAdjustment *)gtk_adjustment_new(1.0, 0.1, 10.0, 0.01, 0.1, 0.0);
    sc.ga_spin7_adj = (GtkAdjustment *)gtk_adjustment_new(1.0, 0.1, 10.0, 0.01, 0.1, 0.0);
    sc.ga_spin8_adj = (GtkAdjustment *)gtk_adjustment_new(1.0, 0.1, 10.0, 0.01, 0.1, 0.0);
    sc.ga_spin9_adj = (GtkAdjustment *)gtk_adjustment_new(1.0, 0.1, 10.0, 0.01, 0.1, 0.0);
#endif

    sc.ga_spin1 = gtk_spin_button_new(sc.ga_spin1_adj, 0.1, 3);
    sc.ga_spin2 = gtk_spin_button_new(sc.ga_spin2_adj, 0.1, 3);
    sc.ga_spin3 = gtk_spin_button_new(sc.ga_spin3_adj, 0.1, 3);
#ifdef HAVE_MATH_H
    sc.ga_spin4 = gtk_spin_button_new(sc.ga_spin4_adj, 0.1, 3);
    sc.ga_spin5 = gtk_spin_button_new(sc.ga_spin5_adj, 0.1, 3);
    sc.ga_spin6 = gtk_spin_button_new(sc.ga_spin6_adj, 0.1, 3);
    sc.ga_spin7 = gtk_spin_button_new(sc.ga_spin7_adj, 0.1, 3);
    sc.ga_spin8 = gtk_spin_button_new(sc.ga_spin8_adj, 0.1, 3);
    sc.ga_spin9 = gtk_spin_button_new(sc.ga_spin9_adj, 0.1, 3);
#endif

    gtk_table_attach_defaults(GTK_TABLE(sc.ga_table), sc.ga_spin1, 3, 4, 3, 4);
    gtk_table_attach_defaults(GTK_TABLE(sc.ga_table), sc.ga_spin2, 5, 6, 3, 4);
    gtk_table_attach_defaults(GTK_TABLE(sc.ga_table), sc.ga_spin3, 7, 8, 3, 4);
#ifdef HAVE_MATH_H
    gtk_table_attach_defaults(GTK_TABLE(sc.ga_table), sc.ga_spin4, 3, 4, 4, 5);
    gtk_table_attach_defaults(GTK_TABLE(sc.ga_table), sc.ga_spin5, 5, 6, 4, 5);
    gtk_table_attach_defaults(GTK_TABLE(sc.ga_table), sc.ga_spin6, 7, 8, 4, 5);
    gtk_table_attach_defaults(GTK_TABLE(sc.ga_table), sc.ga_spin7, 3, 4, 5, 6);
    gtk_table_attach_defaults(GTK_TABLE(sc.ga_table), sc.ga_spin8, 5, 6, 5, 6);
    gtk_table_attach_defaults(GTK_TABLE(sc.ga_table), sc.ga_spin9, 7, 8, 5, 6);
#endif

    setgammaspinvalues();

    gtk_widget_show(sc.ga_spin1);
    gtk_widget_show(sc.ga_spin2);
    gtk_widget_show(sc.ga_spin3);
#ifdef HAVE_MATH_H
    gtk_widget_show(sc.ga_spin4);
    gtk_widget_show(sc.ga_spin5);
    gtk_widget_show(sc.ga_spin6);
    gtk_widget_show(sc.ga_spin7);
    gtk_widget_show(sc.ga_spin8);
    gtk_widget_show(sc.ga_spin9);
#endif

    if((!(sc.sdflags & SiS_SD_ISDHXINERAMA)) && (!(sc.sdflags & SiS_SD_ISDEPTH8))) {
       if((!GTK_WIDGET_IS_SENSITIVE(sc.ga_check1)) && (!GTK_WIDGET_IS_SENSITIVE(sc.ga_check2))) {
          gtk_widget_set_sensitive(sc.ga_spin1, FALSE);
          gtk_widget_set_sensitive(sc.ga_spin2, FALSE);
          gtk_widget_set_sensitive(sc.ga_spin3, FALSE);
#ifdef HAVE_MATH_H
          gtk_widget_set_sensitive(sc.ga_spin4, FALSE);
          gtk_widget_set_sensitive(sc.ga_spin5, FALSE);
          gtk_widget_set_sensitive(sc.ga_spin6, FALSE);
	  gtk_widget_set_sensitive(sc.ga_spin7, FALSE);
          gtk_widget_set_sensitive(sc.ga_spin8, FALSE);
          gtk_widget_set_sensitive(sc.ga_spin9, FALSE);
#endif
       }
    }

    g_signal_connect(G_OBJECT(sc.ga_spin1_adj), "value_changed", G_CALLBACK (ga_spin1_changed), NULL);
    g_signal_connect(G_OBJECT(sc.ga_spin2_adj), "value_changed", G_CALLBACK (ga_spin2_changed), NULL);
    g_signal_connect(G_OBJECT(sc.ga_spin3_adj), "value_changed", G_CALLBACK (ga_spin3_changed), NULL);
#ifdef HAVE_MATH_H
    g_signal_connect(G_OBJECT(sc.ga_spin4_adj), "value_changed", G_CALLBACK (ga_spin4_changed), NULL);
    g_signal_connect(G_OBJECT(sc.ga_spin5_adj), "value_changed", G_CALLBACK (ga_spin5_changed), NULL);
    g_signal_connect(G_OBJECT(sc.ga_spin6_adj), "value_changed", G_CALLBACK (ga_spin6_changed), NULL);
    g_signal_connect(G_OBJECT(sc.ga_spin7_adj), "value_changed", G_CALLBACK (ga_spin7_changed), NULL);
    g_signal_connect(G_OBJECT(sc.ga_spin8_adj), "value_changed", G_CALLBACK (ga_spin8_changed), NULL);
    g_signal_connect(G_OBJECT(sc.ga_spin9_adj), "value_changed", G_CALLBACK (ga_spin9_changed), NULL);
#endif

#ifdef HAVE_MATH_H
    sc.ga_drawable = gtk_drawing_area_new();
    gtk_widget_set_size_request(sc.ga_drawable, GACW, GACH);
    sc.ga_vbox = gtk_vbox_new(FALSE, 0);
    sc.ga_hbox = gtk_hbox_new(FALSE, 0);
    gtk_box_pack_start(GTK_BOX(sc.ga_vbox), sc.ga_hbox, FALSE, FALSE, 0);
    gtk_box_pack_end(GTK_BOX(sc.ga_hbox), sc.ga_drawable, FALSE, FALSE, 0);
    gtk_table_attach_defaults(GTK_TABLE(sc.ga_table), sc.ga_vbox, 6, 8, 0, 3);
    gtk_widget_show(sc.ga_drawable);
    gtk_widget_show(sc.ga_hbox);
    gtk_widget_show(sc.ga_vbox);
    g_signal_connect(G_OBJECT(sc.ga_drawable), "expose-event", GTK_SIGNAL_FUNC(ga_picture_exposed), NULL);
#endif

    sc.ga_label = xpm_label_box(icon_gamma, "Gamma", NULL, 1);
    gtk_notebook_append_page(GTK_NOTEBOOK(sc.notebook), sc.ga_frame, sc.ga_label);

    /* page 4 */
    if(vbsupportstv()) {

       sc.tp_box = gtk_hbox_new(FALSE, 3);
       gtk_widget_show(sc.tp_box);

       sc.tp_frame = gtk_frame_new("TV settings");
       gtk_container_set_border_width(GTK_CONTAINER(sc.tp_frame), 8);
       gtk_widget_show(sc.tp_frame);

       gtk_box_pack_start(GTK_BOX(sc.tp_box), sc.tp_frame, FALSE, FALSE, 0);

       if(sc.currentvbflags & VB_SISBRIDGE) {

          sc.tp_table = gtk_table_new(6, 3, TRUE);
          gtk_table_set_row_spacings(GTK_TABLE(sc.tp_table), 0);
          gtk_table_set_col_spacings(GTK_TABLE(sc.tp_table), 13);

          gtk_container_set_border_width(GTK_CONTAINER(sc.tp_table), 4);

          gtk_container_add(GTK_CONTAINER(sc.tp_frame), sc.tp_table);
          gtk_widget_show(sc.tp_table);

          sc.tp_type = 1;

          sc.tp_label_af = gtk_label_new("Antiflicker");
          gtk_widget_show(sc.tp_label_af);
          gtk_table_attach_defaults(GTK_TABLE(sc.tp_table), sc.tp_label_af, 0, 1, 0, 1);

          sc.tp_omenu = gtk_option_menu_new();

          sc.tp_menu = gtk_menu_new();

          sc.tp_menuitem1 = gtk_menu_item_new_with_label("Off");
          sc.tp_menuitem2 = gtk_menu_item_new_with_label("Low");
          sc.tp_menuitem3 = gtk_menu_item_new_with_label("Medium");
          sc.tp_menuitem4 = gtk_menu_item_new_with_label("High");
          sc.tp_menuitem5 = gtk_menu_item_new_with_label("Adaptive");

          gtk_menu_shell_append(GTK_MENU_SHELL(sc.tp_menu), sc.tp_menuitem1);
          gtk_menu_shell_append(GTK_MENU_SHELL(sc.tp_menu), sc.tp_menuitem2);
          gtk_menu_shell_append(GTK_MENU_SHELL(sc.tp_menu), sc.tp_menuitem3);
          gtk_menu_shell_append(GTK_MENU_SHELL(sc.tp_menu), sc.tp_menuitem4);
          gtk_menu_shell_append(GTK_MENU_SHELL(sc.tp_menu), sc.tp_menuitem5);

          gtk_widget_show(sc.tp_menuitem1);
          gtk_widget_show(sc.tp_menuitem2);
          gtk_widget_show(sc.tp_menuitem3);
          gtk_widget_show(sc.tp_menuitem4);
          gtk_widget_show(sc.tp_menuitem5);

          gtk_option_menu_set_menu(GTK_OPTION_MENU(sc.tp_omenu), sc.tp_menu);

          setantiflickerhistory();

          gtk_widget_show(sc.tp_omenu);

          gtk_table_attach_defaults(GTK_TABLE(sc.tp_table), sc.tp_omenu, 1, 3, 0, 1);

          g_signal_connect(G_OBJECT(sc.tp_omenu), "changed", G_CALLBACK(tp_af_changed), NULL);

          sc.tp_label_sa = gtk_label_new("Saturation");
          gtk_widget_show(sc.tp_label_sa);
          gtk_table_attach_defaults(GTK_TABLE(sc.tp_table), sc.tp_label_sa, 0, 1, 1, 2);

          sc.tp_range_sa = gtk_hscale_new_with_range(0, 15, 1);
	  gtk_scale_set_value_pos(GTK_SCALE(sc.tp_range_sa), GTK_POS_RIGHT);
          gtk_widget_show(sc.tp_range_sa);
          setsaturationvalue();
          gtk_table_attach_defaults(GTK_TABLE(sc.tp_table), sc.tp_range_sa, 1, 3, 1, 2);

	  if(sc.currentvbflags & VB_301) {
	     gtk_widget_set_sensitive(sc.tp_range_sa, FALSE);
          } else {
             GtkAdjustment *myadj = gtk_range_get_adjustment(GTK_RANGE(sc.tp_range_sa));
             g_signal_connect(G_OBJECT(myadj), "value_changed", G_CALLBACK(tp_sa_changed), NULL);
          }

          sc.tp_label_ee = gtk_label_new("Edge enhance");
          gtk_widget_show(sc.tp_label_ee);
          gtk_table_attach_defaults(GTK_TABLE(sc.tp_table), sc.tp_label_ee, 0, 1, 2, 3);

          sc.tp_range_ee = gtk_hscale_new_with_range(0, 15, 1);
	  gtk_scale_set_value_pos(GTK_SCALE(sc.tp_range_ee), GTK_POS_RIGHT);
          gtk_widget_show(sc.tp_range_ee);
          setedgeenhancevalue();
          gtk_table_attach_defaults(GTK_TABLE(sc.tp_table), sc.tp_range_ee, 1, 3, 2, 3);

	  if(!(sc.currentvbflags & VB_301)) {
             gtk_widget_set_sensitive(sc.tp_range_ee, FALSE);
          } else {
             GtkAdjustment *myadj = gtk_range_get_adjustment(GTK_RANGE(sc.tp_range_ee));
             g_signal_connect(G_OBJECT(myadj), "value_changed", G_CALLBACK(tp_ee_changed), NULL);
          }

	  sc.tp_label_coc = gtk_label_new("Color (coarse)");
          gtk_widget_show(sc.tp_label_coc);
          gtk_table_attach_defaults(GTK_TABLE(sc.tp_table), sc.tp_label_coc, 0, 1, 3, 4);

          sc.tp_range_coc = gtk_hscale_new_with_range(-120, 120, 1);
	  gtk_scale_set_value_pos(GTK_SCALE(sc.tp_range_coc), GTK_POS_RIGHT);
          gtk_widget_show(sc.tp_range_coc);
          setcolcalibcvalue();
          gtk_table_attach_defaults(GTK_TABLE(sc.tp_table), sc.tp_range_coc, 1, 3, 3, 4);

          {
          GtkAdjustment *myadj = gtk_range_get_adjustment(GTK_RANGE(sc.tp_range_coc));
          g_signal_connect(G_OBJECT(myadj), "value_changed", G_CALLBACK(tp_coc_changed), NULL);
          }

	  sc.tp_label_cof = gtk_label_new("Color (fine)");
          gtk_widget_show(sc.tp_label_cof);
          gtk_table_attach_defaults(GTK_TABLE(sc.tp_table), sc.tp_label_cof, 0, 1, 4, 5);

          sc.tp_range_cof = gtk_hscale_new_with_range(-128, 127, 1);
	  gtk_scale_set_value_pos(GTK_SCALE(sc.tp_range_cof), GTK_POS_RIGHT);
          gtk_widget_show(sc.tp_range_cof);
          setcolcalibfvalue();
          gtk_table_attach_defaults(GTK_TABLE(sc.tp_table), sc.tp_range_cof, 1, 3, 4, 5);

          {
          GtkAdjustment *myadj = gtk_range_get_adjustment(GTK_RANGE(sc.tp_range_cof));
          g_signal_connect(G_OBJECT(myadj), "value_changed", G_CALLBACK(tp_cof_changed), NULL);
          }

	  sc.tp_check_cfi = gtk_check_button_new_with_label("C-Filter");
          gtk_table_attach_defaults(GTK_TABLE(sc.tp_table), sc.tp_check_cfi, 0, 1, 5, 6);
          gtk_widget_show(sc.tp_check_cfi);

          setcfiltervalue();
          gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(sc.tp_check_cfi), TRUE);

	  sc.tp_label_yfi = gtk_label_new("      Y-Filter");
	  gtk_label_set_justify(GTK_LABEL(sc.tp_label_yfi), GTK_JUSTIFY_RIGHT);
          gtk_widget_show(sc.tp_label_yfi);
          gtk_table_attach_defaults(GTK_TABLE(sc.tp_table), sc.tp_label_yfi, 1, 2, 5, 6);

          sc.tp_omenuy = gtk_option_menu_new();

          sc.tp_menuy = gtk_menu_new();

          sc.tp_menuitemy1 = gtk_menu_item_new_with_label("Off");
          sc.tp_menuitemy2 = gtk_menu_item_new_with_label("Default");
          sc.tp_menuitemy3 = gtk_menu_item_new_with_label("Filter 1");
          sc.tp_menuitemy4 = gtk_menu_item_new_with_label("Filter 2");
          sc.tp_menuitemy5 = gtk_menu_item_new_with_label("Filter 3");
	  sc.tp_menuitemy6 = gtk_menu_item_new_with_label("Filter 4");
	  sc.tp_menuitemy7 = gtk_menu_item_new_with_label("Filter 5");
	  sc.tp_menuitemy8 = gtk_menu_item_new_with_label("Filter 6");
	  sc.tp_menuitemy9 = gtk_menu_item_new_with_label("Filter 7");

          gtk_menu_shell_append(GTK_MENU_SHELL(sc.tp_menuy), sc.tp_menuitemy1);
          gtk_menu_shell_append(GTK_MENU_SHELL(sc.tp_menuy), sc.tp_menuitemy2);
          gtk_menu_shell_append(GTK_MENU_SHELL(sc.tp_menuy), sc.tp_menuitemy3);
          gtk_menu_shell_append(GTK_MENU_SHELL(sc.tp_menuy), sc.tp_menuitemy4);
          gtk_menu_shell_append(GTK_MENU_SHELL(sc.tp_menuy), sc.tp_menuitemy5);
	  gtk_menu_shell_append(GTK_MENU_SHELL(sc.tp_menuy), sc.tp_menuitemy6);
	  gtk_menu_shell_append(GTK_MENU_SHELL(sc.tp_menuy), sc.tp_menuitemy7);
	  gtk_menu_shell_append(GTK_MENU_SHELL(sc.tp_menuy), sc.tp_menuitemy8);
	  gtk_menu_shell_append(GTK_MENU_SHELL(sc.tp_menuy), sc.tp_menuitemy9);

          gtk_widget_show(sc.tp_menuitemy1);
          gtk_widget_show(sc.tp_menuitemy2);
          gtk_widget_show(sc.tp_menuitemy3);
          gtk_widget_show(sc.tp_menuitemy4);
          gtk_widget_show(sc.tp_menuitemy5);
	  gtk_widget_show(sc.tp_menuitemy6);
	  gtk_widget_show(sc.tp_menuitemy7);
	  gtk_widget_show(sc.tp_menuitemy8);
	  gtk_widget_show(sc.tp_menuitemy9);

          gtk_option_menu_set_menu(GTK_OPTION_MENU(sc.tp_omenuy), sc.tp_menuy);

          setyfilterhistory();

          gtk_widget_show(sc.tp_omenuy);

          gtk_table_attach_defaults(GTK_TABLE(sc.tp_table), sc.tp_omenuy, 2, 3, 5, 6);

          g_signal_connect(G_OBJECT(sc.tp_omenuy), "changed", G_CALLBACK(tp_yfi_changed), NULL);

       } else if(sc.currentvbflags & VB_CHRONTEL) {

          sc.tp_table = gtk_table_new(5, 3, TRUE);
          gtk_table_set_row_spacings(GTK_TABLE(sc.tp_table), 2);
          gtk_table_set_col_spacings(GTK_TABLE(sc.tp_table), 1);

          gtk_container_set_border_width(GTK_CONTAINER(sc.tp_table), 4);

          gtk_container_add(GTK_CONTAINER(sc.tp_frame), sc.tp_table);
          gtk_widget_show(sc.tp_table);

          sc.tp_type = 2;

          sc.tp_label_co = gtk_label_new("Contrast");
          gtk_widget_show(sc.tp_label_co);
          gtk_table_attach_defaults(GTK_TABLE(sc.tp_table), sc.tp_label_co, 0, 1, 0, 1);

          sc.tp_range_co = gtk_hscale_new_with_range(0, 15, 1);
          gtk_scale_set_value_pos(GTK_SCALE(sc.tp_range_co), GTK_POS_RIGHT);
          gtk_widget_show(sc.tp_range_co);
          setcontrastvalue();
          gtk_table_attach_defaults(GTK_TABLE(sc.tp_table), sc.tp_range_co, 1, 3, 0, 1);

          {
          GtkAdjustment *myadj = gtk_range_get_adjustment(GTK_RANGE(sc.tp_range_co));
          g_signal_connect(G_OBJECT(myadj), "value_changed", G_CALLBACK(tp_co_changed), NULL);
          }

          sc.tp_label_te = gtk_label_new("Text enhance");
          gtk_widget_show(sc.tp_label_te);
          gtk_table_attach_defaults(GTK_TABLE(sc.tp_table), sc.tp_label_te, 0, 1, 1, 2);

          sc.tp_range_te = gtk_hscale_new_with_range(0, 15, 1);
          gtk_scale_set_value_pos(GTK_SCALE(sc.tp_range_te), GTK_POS_RIGHT);
          gtk_widget_show(sc.tp_range_te);
          settextenhancevalue();
          gtk_table_attach_defaults(GTK_TABLE(sc.tp_table), sc.tp_range_te, 1, 3, 1, 2);

          {
          GtkAdjustment *myadj = gtk_range_get_adjustment(GTK_RANGE(sc.tp_range_te));
          g_signal_connect(G_OBJECT(myadj), "value_changed", G_CALLBACK(tp_te_changed), NULL);
          }

          sc.tp_label_cf = gtk_label_new("C-flickerfilter");
          gtk_widget_show(sc.tp_label_cf);
          gtk_table_attach_defaults(GTK_TABLE(sc.tp_table), sc.tp_label_cf, 0, 1, 2, 3);

          sc.tp_range_cf = gtk_hscale_new_with_range(0, 15, 1);
          gtk_scale_set_value_pos(GTK_SCALE(sc.tp_range_cf), GTK_POS_RIGHT);
          gtk_widget_show(sc.tp_range_cf);
          setchromaffvalue();
          gtk_table_attach_defaults(GTK_TABLE(sc.tp_table), sc.tp_range_cf, 1, 3, 2, 3);

          {
          GtkAdjustment *myadj = gtk_range_get_adjustment(GTK_RANGE(sc.tp_range_cf));
          g_signal_connect(G_OBJECT(myadj), "value_changed", G_CALLBACK(tp_cf_changed), NULL);
          }

          sc.tp_label_lf = gtk_label_new("L-flickerfilter");
          gtk_widget_show(sc.tp_label_lf);
          gtk_table_attach_defaults(GTK_TABLE(sc.tp_table), sc.tp_label_lf, 0, 1, 3, 4);

          sc.tp_range_lf = gtk_hscale_new_with_range(0, 15, 1);
          gtk_scale_set_value_pos(GTK_SCALE(sc.tp_range_lf), GTK_POS_RIGHT);
          gtk_widget_show(sc.tp_range_lf);
          setlumaffvalue();
          gtk_table_attach_defaults(GTK_TABLE(sc.tp_table), sc.tp_range_lf, 1, 3, 3, 4);

          {
          GtkAdjustment *myadj = gtk_range_get_adjustment(GTK_RANGE(sc.tp_range_lf));
          g_signal_connect(G_OBJECT(myadj), "value_changed", G_CALLBACK(tp_lf_changed), NULL);
          }

          sc.tp_check_col = gtk_check_button_new_with_label("Color (CVBS)");
          gtk_table_attach_defaults(GTK_TABLE(sc.tp_table), sc.tp_check_col, 0, 3, 4, 5);
          gtk_widget_show(sc.tp_check_col);
          setcvbscolorvalue();
          g_signal_connect(G_OBJECT(sc.tp_check_col), "toggled", G_CALLBACK (tp_check_col_toggled), NULL);

       }

       if(sc.currentvbflags & VB_SISBRIDGE) {
          sc.tv_frame = gtk_frame_new("TV position & size");
       } else {
          sc.tv_frame = gtk_frame_new("TV position");
       }
       gtk_container_set_border_width(GTK_CONTAINER(sc.tv_frame), 8);
       gtk_widget_show(sc.tv_frame);

       gtk_box_pack_start(GTK_BOX(sc.tp_box), sc.tv_frame, FALSE, FALSE, 0);

       sc.tv_table = gtk_table_new(3, 3, TRUE);
       gtk_table_set_row_spacings(GTK_TABLE(sc.tv_table), 8);

       gtk_table_set_col_spacings(GTK_TABLE(sc.tv_table), 4);

       gtk_container_set_border_width(GTK_CONTAINER(sc.tv_table), 4);

       gtk_container_add(GTK_CONTAINER(sc.tv_frame), sc.tv_table);
       gtk_widget_show(sc.tv_table);

       sc.tv_arrow_le = xpm_label_box(icon_tvleft, NULL, NULL, 1);
       sc.tv_arrow_ri = xpm_label_box(icon_tvright, NULL, NULL, 1);
       sc.tv_arrow_up = xpm_label_box(icon_tvup, NULL, NULL, 1);
       sc.tv_arrow_dn = xpm_label_box(icon_tvdown, NULL, NULL, 1);

       sc.tv_button_up = gtk_button_new();
       sc.tv_button_dn = gtk_button_new();
       sc.tv_button_ri = gtk_button_new();
       sc.tv_button_le = gtk_button_new();

       gtk_container_add(GTK_CONTAINER(sc.tv_button_up), sc.tv_arrow_up);
       gtk_container_add(GTK_CONTAINER(sc.tv_button_dn), sc.tv_arrow_dn);
       gtk_container_add(GTK_CONTAINER(sc.tv_button_le), sc.tv_arrow_le);
       gtk_container_add(GTK_CONTAINER(sc.tv_button_ri), sc.tv_arrow_ri);

       sc.tv_button_de = gtk_button_new_with_mnemonic("_Default");

       gtk_widget_show(sc.tv_button_up);
       gtk_widget_show(sc.tv_button_ri);
       gtk_widget_show(sc.tv_button_le);
       gtk_widget_show(sc.tv_button_dn);
       gtk_widget_show(sc.tv_button_de);

       gtk_widget_show(sc.tv_arrow_le);
       gtk_widget_show(sc.tv_arrow_ri);
       gtk_widget_show(sc.tv_arrow_dn);
       gtk_widget_show(sc.tv_arrow_up);

       gtk_table_attach_defaults(GTK_TABLE(sc.tv_table), sc.tv_button_up, 1, 2, 0, 1);
       gtk_table_attach_defaults(GTK_TABLE(sc.tv_table), sc.tv_button_le, 0, 1, 1, 2);
       gtk_table_attach_defaults(GTK_TABLE(sc.tv_table), sc.tv_button_de, 1, 2, 1, 2);
       gtk_table_attach_defaults(GTK_TABLE(sc.tv_table), sc.tv_button_ri, 2, 3, 1, 2);
       gtk_table_attach_defaults(GTK_TABLE(sc.tv_table), sc.tv_button_dn, 1, 2, 2, 3);

       if(!(sc.sdflags & SiS_SD_SUPPORTTVPOS)) {
          gtk_widget_set_sensitive(sc.tv_button_up, FALSE);
          gtk_widget_set_sensitive(sc.tv_button_dn, FALSE);
          gtk_widget_set_sensitive(sc.tv_button_le, FALSE);
          gtk_widget_set_sensitive(sc.tv_button_ri, FALSE);
          gtk_widget_set_sensitive(sc.tv_button_de, FALSE);
       }

       g_signal_connect(G_OBJECT (sc.tv_button_up), "clicked", G_CALLBACK (tv_up_clicked), NULL);
       g_signal_connect(G_OBJECT (sc.tv_button_dn), "clicked", G_CALLBACK (tv_dn_clicked), NULL);
       g_signal_connect(G_OBJECT (sc.tv_button_le), "clicked", G_CALLBACK (tv_le_clicked), NULL);
       g_signal_connect(G_OBJECT (sc.tv_button_ri), "clicked", G_CALLBACK (tv_ri_clicked), NULL);
       g_signal_connect(G_OBJECT (sc.tv_button_de), "clicked", G_CALLBACK (tv_de_clicked), NULL);

       if(sc.currentvbflags & VB_SISBRIDGE) {
          sc.tv_arrow_xp = xpm_label_box(icon_tvxp, NULL, NULL, 1);
	  sc.tv_arrow_xm = xpm_label_box(icon_tvxm, NULL, NULL, 1);
	  sc.tv_arrow_yp = xpm_label_box(icon_tvyp, NULL, NULL, 1);
	  sc.tv_arrow_ym = xpm_label_box(icon_tvym, NULL, NULL, 1);

	  sc.tv_button_xp = gtk_button_new();
          sc.tv_button_xm = gtk_button_new();
	  sc.tv_button_yp = gtk_button_new();
          sc.tv_button_ym = gtk_button_new();

	  gtk_container_set_border_width(GTK_CONTAINER(sc.tv_button_xp), 1);
	  gtk_container_set_border_width(GTK_CONTAINER(sc.tv_button_xm), 1);
	  gtk_container_set_border_width(GTK_CONTAINER(sc.tv_button_yp), 1);
	  gtk_container_set_border_width(GTK_CONTAINER(sc.tv_button_ym), 1);

	  gtk_container_add(GTK_CONTAINER(sc.tv_button_xp), sc.tv_arrow_xp);
          gtk_container_add(GTK_CONTAINER(sc.tv_button_xm), sc.tv_arrow_xm);
	  gtk_container_add(GTK_CONTAINER(sc.tv_button_yp), sc.tv_arrow_yp);
          gtk_container_add(GTK_CONTAINER(sc.tv_button_ym), sc.tv_arrow_ym);

	  gtk_widget_show(sc.tv_button_xp);
          gtk_widget_show(sc.tv_button_xm);
	  gtk_widget_show(sc.tv_button_yp);
          gtk_widget_show(sc.tv_button_ym);

	  gtk_widget_show(sc.tv_arrow_xp);
	  gtk_widget_show(sc.tv_arrow_xm);
          gtk_widget_show(sc.tv_arrow_yp);
	  gtk_widget_show(sc.tv_arrow_ym);

	  gtk_table_attach_defaults(GTK_TABLE(sc.tv_table), sc.tv_button_xm, 0, 1, 2, 3);
          gtk_table_attach_defaults(GTK_TABLE(sc.tv_table), sc.tv_button_xp, 2, 3, 2, 3);
	  gtk_table_attach_defaults(GTK_TABLE(sc.tv_table), sc.tv_button_ym, 0, 1, 0, 1);
          gtk_table_attach_defaults(GTK_TABLE(sc.tv_table), sc.tv_button_yp, 2, 3, 0, 1);

	  g_signal_connect(G_OBJECT (sc.tv_button_xp), "clicked", G_CALLBACK (tv_xp_clicked), NULL);
          g_signal_connect(G_OBJECT (sc.tv_button_xm), "clicked", G_CALLBACK (tv_xm_clicked), NULL);
	  g_signal_connect(G_OBJECT (sc.tv_button_yp), "clicked", G_CALLBACK (tv_yp_clicked), NULL);
          g_signal_connect(G_OBJECT (sc.tv_button_ym), "clicked", G_CALLBACK (tv_ym_clicked), NULL);
       }

       sc.tp_label = xpm_label_box(icon_tvsettings, "TV", NULL, 1);
       gtk_notebook_append_page(GTK_NOTEBOOK(sc.notebook), sc.tp_box, sc.tp_label);

    } else {

       sc.tp_frame = gtk_frame_new("TV settings");
       gtk_container_set_border_width(GTK_CONTAINER(sc.tp_frame), 8);
       gtk_widget_show(sc.tp_frame);

       sc.tv_table = gtk_table_new(3, 3, TRUE);
       gtk_table_set_row_spacings(GTK_TABLE(sc.tv_table), 10);
       gtk_table_set_col_spacings(GTK_TABLE(sc.tv_table), 4);

       gtk_container_set_border_width(GTK_CONTAINER(sc.tv_table), 8);

       gtk_container_add(GTK_CONTAINER(sc.tp_frame), sc.tv_table);
       gtk_widget_show(sc.tv_table);

       sc.tv_label_notv = gtk_label_new("No TV connector present");
       gtk_widget_show(sc.tv_label_notv);
       gtk_table_attach_defaults(GTK_TABLE(sc.tv_table), sc.tv_label_notv, 0, 3, 1, 2);

       sc.tp_label = xpm_label_box(icon_tvsettings, "TV", NULL, 1);
       gtk_notebook_append_page(GTK_NOTEBOOK(sc.notebook), sc.tp_frame, sc.tp_label);

    }

    /* page 5 */
    sc.vi_frame = gtk_frame_new("Video settings (Xv)");
    gtk_container_set_border_width(GTK_CONTAINER(sc.vi_frame), 8);
    gtk_widget_show(sc.vi_frame);

    {
       int rows = 5, colspacing = 10, border = 15;

       if(sc.sdflags & SiS_SD_SUPPORTXVGAMMA1) rows = 6;
#ifdef USEXV
       if(!sc.noxvdemo) {
          colspacing = 5; border = 10;
       }
#endif
       sc.vi_table = gtk_table_new(rows, 3, TRUE);
       gtk_table_set_row_spacings(GTK_TABLE(sc.vi_table), 2);
       gtk_table_set_col_spacings(GTK_TABLE(sc.vi_table), colspacing);
       gtk_container_set_border_width(GTK_CONTAINER(sc.vi_table), border);
       vitable = sc.vi_table;

       gtk_container_add(GTK_CONTAINER(sc.vi_frame), sc.vi_table);
       gtk_widget_show(sc.vi_table);

#ifdef USEXV
       if(!sc.noxvdemo) {
          sc.vi_table2 = gtk_table_new(4, 3, FALSE);
	  gtk_table_set_row_spacings(GTK_TABLE(sc.vi_table), 2);
          gtk_table_set_col_spacings(GTK_TABLE(sc.vi_table), 2);
	  gtk_container_set_border_width(GTK_CONTAINER(sc.vi_table2), 0);
	  gtk_table_attach_defaults(GTK_TABLE(sc.vi_table), sc.vi_table2, 1, 3, 0, 4);
	  gtk_widget_show(sc.vi_table2);
	  vitable = sc.vi_table2;
       }
#endif
    }

    sc.vi_label_co = gtk_label_new("Contrast");
    gtk_widget_show(sc.vi_label_co);
    gtk_table_attach_defaults(GTK_TABLE(vitable), sc.vi_label_co, 0, 1, 0, 1);

    sc.vi_range_co = gtk_hscale_new_with_range(0, 7, 1);
    gtk_scale_set_value_pos(GTK_SCALE(sc.vi_range_co), GTK_POS_RIGHT);
    gtk_widget_show(sc.vi_range_co);
    setvicontrastvalue();
    gtk_table_attach_defaults(GTK_TABLE(vitable), sc.vi_range_co, 1, 3, 0, 1);

    {
       GtkAdjustment *myadj = gtk_range_get_adjustment(GTK_RANGE(sc.vi_range_co));
       g_signal_connect(G_OBJECT(myadj), "value_changed", G_CALLBACK(vi_co_changed), NULL);
    }

    sc.vi_label_br = gtk_label_new("Brightness");
    gtk_widget_show(sc.vi_label_br);
    gtk_table_attach_defaults(GTK_TABLE(vitable), sc.vi_label_br, 0, 1, 1, 2);

    sc.vi_range_br = gtk_hscale_new_with_range(-128, 127, 1);
    gtk_scale_set_value_pos(GTK_SCALE(sc.vi_range_br), GTK_POS_RIGHT);
    gtk_widget_show(sc.vi_range_br);
    setvibrightnessvalue();
    gtk_table_attach_defaults(GTK_TABLE(vitable), sc.vi_range_br, 1, 3, 1, 2);

    {
       GtkAdjustment *myadj = gtk_range_get_adjustment(GTK_RANGE(sc.vi_range_br));
       g_signal_connect(G_OBJECT(myadj), "value_changed", G_CALLBACK(vi_br_changed), NULL);
    }

    if(sc.sdflags & (SiS_SD_IS315SERIES | SiS_SD_IS330SERIES)) {
       sc.vi_label_hu = gtk_label_new("Hue");
       gtk_widget_show(sc.vi_label_hu);
       gtk_table_attach_defaults(GTK_TABLE(vitable), sc.vi_label_hu, 0, 1, 2, 3);

       sc.vi_range_hu = gtk_hscale_new_with_range(-8, 7, 1);
       gtk_scale_set_value_pos(GTK_SCALE(sc.vi_range_hu), GTK_POS_RIGHT);
       gtk_widget_show(sc.vi_range_hu);
       setvihuevalue();
       gtk_table_attach_defaults(GTK_TABLE(vitable), sc.vi_range_hu, 1, 3, 2, 3);

       {
          GtkAdjustment *myadj = gtk_range_get_adjustment(GTK_RANGE(sc.vi_range_hu));
          g_signal_connect(G_OBJECT(myadj), "value_changed", G_CALLBACK(vi_hu_changed), NULL);
       }

       sc.vi_label_sa = gtk_label_new("Saturation");
       gtk_widget_show(sc.vi_label_sa);
       gtk_table_attach_defaults(GTK_TABLE(vitable), sc.vi_label_sa, 0, 1, 3, 4);

       sc.vi_range_sa = gtk_hscale_new_with_range(-7, 7, 1);
       gtk_scale_set_value_pos(GTK_SCALE(sc.vi_range_sa), GTK_POS_RIGHT);
       gtk_widget_show(sc.vi_range_sa);
       setvisaturationvalue();
       gtk_table_attach_defaults(GTK_TABLE(vitable), sc.vi_range_sa, 1, 3, 3, 4);

       {
          GtkAdjustment *myadj = gtk_range_get_adjustment(GTK_RANGE(sc.vi_range_sa));
          g_signal_connect(G_OBJECT(myadj), "value_changed", G_CALLBACK(vi_sa_changed), NULL);
       }
    }

#ifdef USEXV
    if(!sc.noxvdemo) {
       sc.vi_layout = gtk_label_new("  ");
       gtk_table_attach_defaults(GTK_TABLE(sc.vi_table), sc.vi_layout, 0, 1, 0, 4);
       gtk_widget_show(sc.vi_layout);
    }
#endif

    if(sc.sdflags & SiS_SD_SUPPORTXVGAMMA1) {

       sc.vi_check_gamma = gtk_check_button_new_with_label("Gamma (CRT1 only)");
       gtk_table_attach_defaults(GTK_TABLE(sc.vi_table), sc.vi_check_gamma, 0, 1, 4, 5);
       gtk_widget_show(sc.vi_check_gamma);

       sc.vi_ga_table = gtk_table_new(1, 6, FALSE);
       gtk_table_set_row_spacings(GTK_TABLE(sc.vi_ga_table), 1);
       gtk_table_set_col_spacings(GTK_TABLE(sc.vi_ga_table), 1);
       gtk_container_set_border_width(GTK_CONTAINER(sc.vi_ga_table), 1);
       gtk_table_attach_defaults(GTK_TABLE(sc.vi_table), sc.vi_ga_table, 1, 3, 4, 5);
       gtk_widget_show(sc.vi_ga_table);

       sc.vi_label_red = gtk_label_new("R");
       gtk_table_attach_defaults(GTK_TABLE(sc.vi_ga_table), sc.vi_label_red, 0, 1, 0, 1);
       gtk_widget_show(sc.vi_label_red);
       sc.vi_label_green = gtk_label_new("G");
       gtk_table_attach_defaults(GTK_TABLE(sc.vi_ga_table), sc.vi_label_green, 2, 3, 0, 1);
       gtk_widget_show(sc.vi_label_green);
       sc.vi_label_blue = gtk_label_new("B");
       gtk_table_attach_defaults(GTK_TABLE(sc.vi_ga_table), sc.vi_label_blue, 4, 5, 0, 1);
       gtk_widget_show(sc.vi_label_blue);

       sc.vi_spin1_adj = (GtkAdjustment *)gtk_adjustment_new(1.0, 0.1, 10.0, 0.1, 1.0, 0.0);
       sc.vi_spin2_adj = (GtkAdjustment *)gtk_adjustment_new(1.0, 0.1, 10.0, 0.1, 1.0, 0.0);
       sc.vi_spin3_adj = (GtkAdjustment *)gtk_adjustment_new(1.0, 0.1, 10.0, 0.1, 1.0, 0.0);
       sc.vi_spin1 = gtk_spin_button_new(sc.vi_spin1_adj, 0.1, 3);
       sc.vi_spin2 = gtk_spin_button_new(sc.vi_spin2_adj, 0.1, 3);
       sc.vi_spin3 = gtk_spin_button_new(sc.vi_spin3_adj, 0.1, 3);
       gtk_table_attach_defaults(GTK_TABLE(sc.vi_ga_table), sc.vi_spin1, 1, 2, 0, 1);
       gtk_table_attach_defaults(GTK_TABLE(sc.vi_ga_table), sc.vi_spin2, 3, 4, 0, 1);
       gtk_table_attach_defaults(GTK_TABLE(sc.vi_ga_table), sc.vi_spin3, 5, 6, 0, 1);
       setxvgammaspinvalues();
       gtk_widget_show(sc.vi_spin1);
       gtk_widget_show(sc.vi_spin2);
       gtk_widget_show(sc.vi_spin3);

       setxvgammastatus();

       g_signal_connect(G_OBJECT(sc.vi_check_gamma), "toggled", G_CALLBACK (vi_check_gamma_toggled), NULL);
       g_signal_connect(G_OBJECT(sc.vi_spin1_adj), "value_changed", G_CALLBACK (vi_spin1_changed), NULL);
       g_signal_connect(G_OBJECT(sc.vi_spin2_adj), "value_changed", G_CALLBACK (vi_spin2_changed), NULL);
       g_signal_connect(G_OBJECT(sc.vi_spin3_adj), "value_changed", G_CALLBACK (vi_spin3_changed), NULL);
    }

    if(!(sc.sdflags & SiS_SD_SUPPORT2OVL)) {
       int where = 4;
       if(sc.sdflags & SiS_SD_SUPPORTXVGAMMA1) where = 5;
       sc.vi_label_sw = gtk_label_new("Video on");
       gtk_widget_show(sc.vi_label_sw);
       gtk_table_attach_defaults(GTK_TABLE(sc.vi_table), sc.vi_label_sw, 0, 1, where, where+1);

       sc.vi_radio_crt1 = gtk_radio_button_new_with_label(NULL, "CRT1");
       gtk_table_attach_defaults(GTK_TABLE(sc.vi_table), sc.vi_radio_crt1, 1, 2, where, where+1);
       gtk_widget_show(sc.vi_radio_crt1);

       sc.vi_group = gtk_radio_button_get_group(GTK_RADIO_BUTTON(sc.vi_radio_crt1));
       sc.vi_radio_crt2 = gtk_radio_button_new_with_label(sc.vi_group, "CRT2");
       gtk_table_attach_defaults(GTK_TABLE(sc.vi_table), sc.vi_radio_crt2, 2, 3, where, where+1);
       gtk_widget_show(sc.vi_radio_crt2);

       setviswitchcrt();

       g_signal_connect(G_OBJECT(sc.vi_radio_crt1), "toggled", G_CALLBACK(vi_radio_changed), NULL);
    }

    sc.vi_label = xpm_label_box(icon_video, "Video", NULL, 1);
    gtk_notebook_append_page(GTK_NOTEBOOK(sc.notebook), sc.vi_frame, sc.vi_label);

    /* page 6 */
    sc.cf_frame = gtk_frame_new("Current configuration");
    gtk_container_set_border_width(GTK_CONTAINER(sc.cf_frame), 8);
    gtk_widget_show(sc.cf_frame);

    sc.cf_text_window = gtk_scrolled_window_new(NULL, NULL);
    gtk_container_set_border_width(GTK_CONTAINER(sc.cf_text_window), 6);
    gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(sc.cf_text_window), GTK_POLICY_AUTOMATIC, GTK_POLICY_ALWAYS);

    sc.cf_tview = gtk_text_view_new();
    sc.cf_buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(sc.cf_tview));

    gtk_text_view_set_wrap_mode(GTK_TEXT_VIEW(sc.cf_tview), GTK_WRAP_WORD);
    gtk_text_view_set_editable(GTK_TEXT_VIEW(sc.cf_tview), FALSE);
    gtk_text_buffer_set_text(sc.cf_buffer, "", -1);

    gtk_scrolled_window_add_with_viewport(GTK_SCROLLED_WINDOW(sc.cf_text_window), sc.cf_tview);

    gtk_container_add(GTK_CONTAINER(sc.cf_frame), sc.cf_text_window);

    gtk_widget_show(sc.cf_tview);
    gtk_widget_show(sc.cf_text_window);

    sc.black = gtk_text_buffer_create_tag(sc.cf_buffer, "black_foreground", "foreground", "black", NULL);
    sc.grey = gtk_text_buffer_create_tag(sc.cf_buffer, "grey_foreground", "foreground", "grey32", NULL);
    sc.red = gtk_text_buffer_create_tag(sc.cf_buffer, "red_foreground", "foreground", "red", NULL);

    sc.cf_label = xpm_label_box(icon_config, "Current", NULL, 1);
    gtk_notebook_append_page(GTK_NOTEBOOK(sc.notebook), sc.cf_frame, sc.cf_label);

    /* page 7 */
    sc.ab_bgcolor.red = 155 << 8;
    sc.ab_bgcolor.green = 152 << 8;
    sc.ab_bgcolor.blue = 255 << 8;

    sc.ab_frame = gtk_frame_new("Info");
    gtk_container_set_border_width(GTK_CONTAINER(sc.ab_frame), 8);
    gtk_widget_show(sc.ab_frame);

    sc.ab_table = gtk_table_new(6, 1, FALSE);
    gtk_table_set_row_spacings(GTK_TABLE(sc.ab_table), 0);
    gtk_table_set_col_spacings(GTK_TABLE(sc.ab_table), 0);
    gtk_container_set_border_width(GTK_CONTAINER(sc.ab_table), 0);
#if 0
    gtk_widget_show(sc.ab_table);
#endif

    sc.ab_layout = gtk_layout_new(NULL, NULL);
    gtk_widget_modify_fg(GTK_WIDGET(sc.ab_layout), GTK_STATE_NORMAL, &sc.ab_bgcolor);
    g_signal_connect(G_OBJECT(sc.ab_layout), "expose_event", G_CALLBACK(ab_layout_expose), NULL);
    gtk_container_add(GTK_CONTAINER(sc.ab_frame), sc.ab_layout);

    gtk_layout_put(GTK_LAYOUT(sc.ab_layout), sc.ab_table, 0, 0);
    gtk_widget_show(sc.ab_layout);

    sc.ab_iconul = gdk_pixmap_create_from_xpm_d(sc.window->window, &sc.ab_maskul, NULL, upperleft_xpm);
    sc.ab_iconlr = gdk_pixmap_create_from_xpm_d(sc.window->window, &sc.ab_masklr, NULL, lowerright_xpm);
    sc.ab_icon = gdk_pixmap_create_from_xpm_d(sc.window->window, &sc.ab_mask, NULL, sisctrl_xpm);

    sc.ab_label_ab1 = gtk_label_new("SiS Display Control Panel (" SISVERSION ")");
    sc.ab_label_ab2 = gtk_label_new("(C) 2004 Thomas Winischhofer <thomas@winischhofer.net>");
    sc.ab_label_ab4 = gtk_label_new("Licensed under the terms of the GPL");

    {
      char *chip, *vbtype;

      chip = getchiptype();

      switch(sc.currentvbflags & (VB_VIDEOBRIDGE | VB_30xBDH)) {
      case VB_301:    		   vbtype = "; SiS 301 video bridge";     break;
      case (VB_301B | VB_30xBDH):  vbtype = "; SiS 301B-DH video bridge"; break;
      case VB_301B:   		   vbtype = "; SiS 301B video bridge";    break;
      case VB_301C:   		   vbtype = "; SiS 301C video bridge";    break;
      case VB_302B:    		   vbtype = "; SiS 302B video bridge";    break;
      case VB_301LV:  		   vbtype = "; SiS 301LV video bridge";   break;
      case VB_302LV:  		   vbtype = "; SiS 302LV video bridge";   break;
      case VB_302ELV:  		   vbtype = "; SiS 302ELV video bridge";   break;
      case (VB_LVDS | VB_CHRONTEL):vbtype = "; LVDS transmitter; Chrontel TV encoder"; break;
      case VB_LVDS:		   vbtype = "; LVDS transmitter";	  break;
      case VB_CHRONTEL: 	   vbtype = "; Chrontel TV encoder";      break;
      case VB_CONEXANT: 	   vbtype = "; Conexant ???";      	  break;
      default: 			   vbtype = "";
      }

      sprintf(label3buffer, "[SiS %s rev %d%s]", chip, (int)(sc.hwinfo >> 24), vbtype);
    }
    sc.ab_label_ab3 = gtk_label_new(label3buffer);

    sprintf(label5buffer, "[SiS X driver version %lu/%02lu/%02lu-%lu]",
       ((sc.driverversion >> 16) & 0xff) + 2000,
       (sc.driverversion >> 8) & 0xff,
       sc.driverversion & 0xff,
       sc.driverversion >> 24);
    sc.ab_label_ab5 = gtk_label_new(label5buffer);

    gtk_table_attach_defaults(GTK_TABLE(sc.ab_table), sc.ab_label_ab1, 0, 1, 1, 2);
    gtk_table_attach_defaults(GTK_TABLE(sc.ab_table), sc.ab_label_ab2, 0, 1, 2, 3);
    gtk_table_attach_defaults(GTK_TABLE(sc.ab_table), sc.ab_label_ab4, 0, 1, 3, 4);
    gtk_table_attach_defaults(GTK_TABLE(sc.ab_table), sc.ab_label_ab3, 0, 1, 4, 5);
    gtk_table_attach_defaults(GTK_TABLE(sc.ab_table), sc.ab_label_ab5, 0, 1, 5, 6);
    gtk_widget_show(sc.ab_label_ab1);
    gtk_widget_show(sc.ab_label_ab2);
    gtk_widget_show(sc.ab_label_ab4);
    gtk_widget_show(sc.ab_label_ab3);
    gtk_widget_show(sc.ab_label_ab5);

    sc.ab_label = xpm_label_box(icon_about, "Info", NULL, 1);
    gtk_notebook_append_page(GTK_NOTEBOOK(sc.notebook), sc.ab_frame, sc.ab_label);

    /* --- end of notebook --- */

    /* Set current page of notebook */
    gtk_notebook_set_current_page(GTK_NOTEBOOK(sc.notebook), 0);

    g_signal_connect(G_OBJECT(sc.notebook), "switch-page", G_CALLBACK(notebook_changed), NULL);

    /* Create and attach buttons */
    sc.button1 = gtk_button_new();
    sc.button2 = gtk_button_new();
    sc.button3 = gtk_button_new();
    sc.button4 = gtk_button_new();

    sc.button1box = xpm_label_box(icon_ok, "_OK", sc.button1, 3);
    sc.button2box = xpm_label_box(icon_apply, "_Apply", sc.button2, 3);
    sc.button3box = xpm_label_box(icon_close, "_Close", sc.button3, 3);
    sc.button4box = xpm_label_box(icon_undo, "_Reset", sc.button4, 3);

    gtk_widget_show(sc.button1box);
    gtk_widget_show(sc.button2box);
    gtk_widget_show(sc.button3box);
    gtk_widget_show(sc.button4box);

    gtk_widget_set_sensitive(sc.button4, FALSE);

    gtk_container_add(GTK_CONTAINER(sc.button1), sc.button1box);
    gtk_container_add(GTK_CONTAINER(sc.button2), sc.button2box);
    gtk_container_add(GTK_CONTAINER(sc.button3), sc.button3box);
    gtk_container_add(GTK_CONTAINER(sc.button4), sc.button4box);

    g_signal_connect(G_OBJECT (sc.button1), "clicked", G_CALLBACK (ok_clicked), NULL);
    g_signal_connect(G_OBJECT (sc.button2), "clicked", G_CALLBACK (apply_clicked), NULL);
    g_signal_connect(G_OBJECT (sc.button3), "clicked", G_CALLBACK (close_clicked), NULL);
    g_signal_connect(G_OBJECT (sc.button4), "clicked", G_CALLBACK (revert_clicked), NULL);

    g_signal_connect_swapped(G_OBJECT(sc.button3), "clicked", G_CALLBACK (gtk_widget_destroy), sc.window);

    gtk_table_attach_defaults(GTK_TABLE(sc.table), sc.button1, 0, 1, 1, 2);
    gtk_table_attach_defaults(GTK_TABLE(sc.table), sc.button2, 1, 2, 1, 2);
    gtk_table_attach_defaults(GTK_TABLE(sc.table), sc.button3, 2, 3, 1, 2);
    gtk_table_attach_defaults(GTK_TABLE(sc.table), sc.button4, 3, 4, 1, 2);

    gtk_widget_show(sc.button1);
    gtk_widget_show(sc.button2);
    gtk_widget_show(sc.button3);
    gtk_widget_show(sc.button4);

    /* Show the table */
    gtk_widget_show(sc.table);

    /* Set window icons */
    {
       GdkPixbuf *icon16_pixbuf, *icon32_pixbuf, *icon48_pixbuf, *icon64_pixbuf;
       GList *icon_list = NULL;

       icon16_pixbuf = gdk_pixbuf_new_from_xpm_data((const char **)sisctrl_icon16);
       icon32_pixbuf = gdk_pixbuf_new_from_xpm_data((const char **)sisctrl_icon32);
       icon48_pixbuf = gdk_pixbuf_new_from_xpm_data((const char **)sisctrl_icon48);
       icon64_pixbuf = gdk_pixbuf_new_from_xpm_data((const char **)sisctrl_icon64);

       icon_list = g_list_append(icon_list, icon16_pixbuf);
       icon_list = g_list_append(icon_list, icon32_pixbuf);
       icon_list = g_list_append(icon_list, icon48_pixbuf);
       icon_list = g_list_append(icon_list, icon64_pixbuf);

       gtk_window_set_icon_list((GtkWindow *)sc.window, icon_list);
       gtk_window_set_default_icon_list(icon_list);

       g_list_free(icon_list);
    }

    /* Show the window */
    gtk_widget_show(sc.window);

#ifdef USEXV
    if(!sc.noxvdemo) {
       gdk_flush();
       makexvwindow();
       g_signal_connect(G_OBJECT(sc.vi_layout), "configure_event", G_CALLBACK(vi_layout_configure), NULL);
       g_signal_connect(G_OBJECT(sc.vi_layout), "expose_event", G_CALLBACK(vi_layout_expose), NULL);
       g_signal_connect(G_OBJECT(sc.window), "configure_event", G_CALLBACK(window_configure), NULL);
       /* g_signal_connect(G_OBJECT(sc.window), "expose_event", G_CALLBACK(window_expose), NULL); */
       g_signal_connect(G_OBJECT(sc.window), "visibility_notify_event", G_CALLBACK(window_visibility), NULL);
       gtk_widget_add_events(sc.window, gtk_widget_get_events(sc.window) | GDK_VISIBILITY_NOTIFY_MASK);
       sc.xvinit = TRUE;
       /* Now show it once, otherwise - strangely- at the first time,
        * Xv will be re-initialized on every XVShmPutImage; seems to
        * relate to gtk internals
        */
       resizexvwindow();
       showxvimage();
       hidexvimage();
    }
#endif
}

static void
writeoutstatus(void)
{
    fprintf(stdout, "MERGEDFB MODE: %s\n", (sc.sdflags & SiS_SD_ISMERGEDFB) ? "ON" : "OFF");
    fprintf(stdout, "DUALHEAD MODE: %s\n", (sc.sdflags & SiS_SD_ISDUALHEAD) ? "ON" : "OFF");
    if(sc.sdflags & SiS_SD_ISDUALHEAD) {
       fprintf(stdout, "XINERAMA: %s\n", (sc.sdflags & SiS_SD_ISDHXINERAMA) ? "ON" : "OFF");
       if(!(sc.sdflags & SiS_SD_ISDHXINERAMA)) {
          fprintf(stdout, "THIS HEAD: %s\n", (sc.sdflags & SiS_SD_ISDHSECONDHEAD) ? "CRT1" : "CRT2");
       }
    }

    fprintf(stdout, "CRT1: %s\n", (sc.currentvbflags & CRT1_LCDA) ?
       			"LCD" : (sc.CRT1isoff ? "OFF" : "ON"));

    fprintf(stdout, "CRT2 CURRENT: ");
    switch(sc.currentvbflags & CRT2_ENABLE) {
       case CRT2_LCD: fprintf(stdout, "LCD\n"); break;
       case CRT2_VGA: fprintf(stdout, "VGA\n"); break;
       case CRT2_TV:
	  if(sc.currentvbflags & VB_SISBRIDGE) {
	     if((sc.currentvbflags & (TV_SVIDEO | TV_AVIDEO)) == (TV_SVIDEO | TV_AVIDEO))
	        fprintf(stdout, "SVIDEO+COMPOSITE\n");
	     else if(sc.currentvbflags & TV_SVIDEO)
	        fprintf(stdout, "SVIDEO\n");
	     else if(sc.currentvbflags & TV_AVIDEO)
		fprintf(stdout, "COMPOSITE\n");
	     else if(sc.currentvbflags & TV_SCART)
		fprintf(stdout, "SCART\n");
	     else if(sc.currentvbflags & TV_HIVISION)
		fprintf(stdout, "HIVISION\n");
	     else if(sc.currentvbflags & TV_YPBPR) {
		fprintf(stdout, "YPBPR");
		if(sc.currentvbflags & TV_YPBPR525I)
		   fprintf(stdout, "480I [");
		else if(sc.currentvbflags & TV_YPBPR525P)
		   fprintf(stdout, "480P [");
		else if(sc.currentvbflags & TV_YPBPR750P)
		   fprintf(stdout, "720P [");
		else
		   fprintf(stdout, "1080I [");
		if(sc.sdflags & SiS_SD_SUPPORTYPBPRAR) {
		   if((sc.currentvbflags & TV_YPBPRAR) == TV_YPBPR43)
		      fprintf(stdout, "4:3]\n");
		   else if((sc.currentvbflags & TV_YPBPRAR) == TV_YPBPR43LB)
		      fprintf(stdout, "4:3LB]\n");
		   else if((sc.currentvbflags & TV_YPBPRAR) == TV_YPBPR169)
		      fprintf(stdout, "16:9]\n");
		   else
		      fprintf(stdout, "UNKNOWN ASPECT]\n");
		} else fprintf(stdout, "UNKNOWN ASPECT]\n");
	     } else
		fprintf(stdout, "UNKNOWN\n");
	  } else {
	     fprintf(stdout, "TV\n");
	  }
	  break;
       default: fprintf(stdout, "OFF\n");
    }
    fprintf(stdout, "CRT2 DETECTED: ");
    if(sc.detecteddevices & CRT2_LCD) fprintf(stdout, "LCD ");
    if(sc.detecteddevices & CRT2_VGA) fprintf(stdout, "VGA ");
    if(sc.detecteddevices & CRT2_TV) {
       if(sc.currentvbflags & VB_SISBRIDGE) {
	  if(sc.detecteddevices & TV_SVIDEO) fprintf(stdout, "TV-SVIDEO ");
	  if(sc.detecteddevices & TV_AVIDEO) fprintf(stdout, "TV-COMPOSITE ");
	  if(sc.detecteddevices & TV_SCART) fprintf(stdout, "TV-SCART ");
	  if(sc.detecteddevices & TV_HIVISION) fprintf(stdout, "TV-HIVISION ");
	  if(sc.detecteddevices & TV_YPBPR) fprintf(stdout, "TV-YPBPR ");
       } else {
          fprintf(stdout, "TV");
       }
    }
    fprintf(stdout, "\n");

    if((sc.currentvbflags & CRT2_TV) && (!(sc.currentvbflags & (TV_YPBPR | TV_HIVISION)))) {
       fprintf(stdout, "TVSTANDARD: ");
       if(sc.currentvbflags & TV_NTSC) {
	  if(sc.currentvbflags & TV_NTSCJ)
	     fprintf(stdout, "NTSC-J\n");
	  else
	     fprintf(stdout, "NTSC\n");
       } else {
	  if(sc.currentvbflags & TV_PALM)
	     fprintf(stdout, "PAL-M\n");
          if(sc.currentvbflags & TV_PALN)
	     fprintf(stdout, "PAL-N\n");
	  else
	     fprintf(stdout, "PAL\n");
       }
       if(sc.currentvbflags & VB_CHRONTEL) {
	  fprintf(stdout, "TVOVERSCAN: ");
	  switch(sc.choverscanstatus) {
	     case 0: fprintf(stdout, "BIOS\n"); break;
	     case 1: fprintf(stdout, "UNDERSCAN\n"); break;
	     case 2: fprintf(stdout, "OVERSCAN\n"); break;
	     case 3: fprintf(stdout, "SUPEROVERSCAN\n"); break;
	     default: fprintf(stdout, "UNKNOWN\n");
	  }
       }
    }

    fprintf(stdout, "GAMMA CORRECTION: ");
    if(sc.sdflags & SiS_SD_ISDHXINERAMA) {
       fprintf(stdout, "UNKNOWN");
    } else if(sc.sdflags & SiS_SD_ISDUALHEAD) {
       if(sc.sdflags & SiS_SD_ISDHSECONDHEAD) {
	  if(sc.gammaflags & 0x01) fprintf(stdout, "CRT1");
       } else {
	  if(sc.gammaflags & 0x02) fprintf(stdout, "CRT2");
       }
    } else {
       if(sc.gammaflags & 0x01) fprintf(stdout, "CRT1 ");
       if(sc.currentvbflags & VB_SISBRIDGE) {
          if(sc.gammaflags & 0x02) fprintf(stdout, "CRT2");
       }
    }
    fprintf(stdout, "\n");
    fprintf(stdout, "GAMMA CURRENT: %.3f %.3f %.3f\n",
      	    sc.currentgamma.red, sc.currentgamma.green, sc.currentgamma.blue);
    fprintf(stdout, "GAMMA BRIGHTNESS CURRENT: %d %d %d\n",
            (int)(sc.gamma_max[0] * 1000),
	    (int)(sc.gamma_max[1] * 1000),
	    (int)(sc.gamma_max[2] * 1000));
    fprintf(stdout, "GAMMA PRE-BRIGHTNESS CURRENT: %d %d %d\n",
    	    (int)(sc.gamma_prescale[0] * 1000),
	    (int)(sc.gamma_prescale[1] * 1000),
	    (int)(sc.gamma_prescale[2] * 1000));

    if(sc.currentvbflags & CRT2_TV) {
       if(sc.currentvbflags & VB_SISBRIDGE) {
          int temp = getxvval(sc.atom_taf);
          fprintf(stdout, "TV ANTIFLICKER: ");
	  switch(temp) {
	     case 0: fprintf(stdout, "OFF\n"); break;
	     case 1: fprintf(stdout, "LOW\n"); break;
	     case 2: fprintf(stdout, "MED\n"); break;
	     case 3: fprintf(stdout, "HIGH\n"); break;
	     case 4: fprintf(stdout, "ADAPTIVE\n"); break;
	  }
	  fprintf(stdout, "TV SATURATION: %d\n", getxvval(sc.atom_tsa));
	  if(sc.currentvbflags & VB_301) {
	     fprintf(stdout, "TV EDGEENHANCE: %d\n", getxvval(sc.atom_tee));
	  }
	  fprintf(stdout, "TV COLOR CALIBRATION COARSE: %d\n", getxvval(sc.atom_coc));
	  fprintf(stdout, "TV COLOR CALIBRATION FINE: %d\n", getxvval(sc.atom_cof));
	  fprintf(stdout, "TV C-FILTER: %s\n", getxvval(sc.atom_cfi) ? "ON" : "OFF");
	  fprintf(stdout, "TV Y-FILTER: %d\n", getxvval(sc.atom_yfi));
       } else if(sc.currentvbflags & VB_CHRONTEL) {
	  fprintf(stdout, "TV CONTRAST: %d\n", getxvval(sc.atom_tco));
	  fprintf(stdout, "TV EDGEENHANCE: %d\n", getxvval(sc.atom_tte));
	  fprintf(stdout, "TV CHROMAFLICKERFILTER: %d\n", getxvval(sc.atom_tcf));
	  fprintf(stdout, "TV LUMAFLICKERFILTER: %d\n", getxvval(sc.atom_tlf));
	  fprintf(stdout, "TV CVBS COLOR: %s\n", getxvval(sc.atom_ccc) ? "ON" : "OFF");
       }
       if(sc.sdflags & SiS_SD_SUPPORTTVPOS) {
	  fprintf(stdout, "TV X POSITION: %d\n", getxvval(sc.atom_tvx));
	  fprintf(stdout, "TV Y POSITION: %d\n", getxvval(sc.atom_tvy));
       }
       if(sc.currentvbflags & VB_SISBRIDGE) {
	  fprintf(stdout, "TV X SCALE: %d\n", getxvval(sc.atom_txs));
	  fprintf(stdout, "TV Y SCALE: %d\n", getxvval(sc.atom_tys));
       }
    }

    fprintf(stdout, "VIDEO BRIGHTNESS: %d\n", getxvval(sc.atom_bri));
    fprintf(stdout, "VIDEO CONTRAST: %d\n", getxvval(sc.atom_con));
    if(sc.sdflags & (SiS_SD_IS315SERIES | SiS_SD_IS330SERIES)) {
       fprintf(stdout, "VIDEO SATURATION: %d\n", getxvval(sc.atom_sat));
       fprintf(stdout, "VIDEO HUE: %d\n", getxvval(sc.atom_hue));
    }
    if(sc.sdflags & SiS_SD_SUPPORTXVGAMMA1) {
       fprintf(stdout, "VIDEO CRT1 GAMMA ENABLE: %s\n", (sc.gammaflags & 0x04) ? "ON" : "OFF");
       if(sc.gammaflags & 0x04) {
          fprintf(stdout, "VIDEO CRT1 GAMMA: %.3f %.3f %.3f\n", sc.xvgammared, sc.xvgammagreen, sc.xvgammablue);
       }
    }
    if(!(sc.sdflags & SiS_SD_SUPPORT2OVL)) {
       fprintf(stdout, "VIDEO ON CRT2: %s\n", getxvval(sc.atom_swc) ? "YES" : "NO");
    }
}


/***********************************************/
/*                   main()                    */
/***********************************************/

int
main(int argc, char *argv[] )
{
    char *displayname = NULL;
    int i, force = 0;
    int MajorVersion, MinorVersion;
    int EventBase, ErrorBase;
    Bool screengiven = FALSE;
    Bool printstatus = FALSE;
    Bool setgammafromdriver = FALSE;
    Bool setcrt1 = FALSE, setcrt2 = FALSE, settvstd = FALSE, settvos = FALSE;
    Bool setxvcrt = FALSE, settvxpos = FALSE, settvypos = FALSE, settvas = FALSE;
    Bool settvxscale = FALSE, settvyscale = FALSE, doexit = FALSE;
    int tvxposvalue, tvyposvalue, tvxscalevalue, tvyscalevalue;
    int setcrt1index, xvcrtindex;
    char *missingparm = "Missing -%s argument, try -help\n";
    char *illegalparm = "Illegal -%s argument, try -help\n";

    char buffer[256];

    int num_adaptors = 0, num_attr = 0;
    XvAdaptorInfo *adaptor_info;
    unsigned int xv_major, xv_minor;
    unsigned int xv_req_base, xv_event_base, xv_error_base;
    int rr_event_base, rr_error_base;
    int xv_val;
    XvAttribute *attributes;

    sc.myscreen = -1;
    sc.lockfile = -1;
    sc.window = NULL;
    sc.tp_type = 0;
    sc.xv_port_nr = -1;
    sc.attr_name = NULL;
    sc.vmmodecount = 0;
    sc.nbscroll = FALSE;
    sc.guimode = FALSE;
#ifdef HAVE_RANDR
    sc.sconf = NULL;
#endif
#ifdef USEXV
    sc.xvinit = sc.xvshown = sc.xvattached = sc.xvraised = sc.xvwasbusy = FALSE;
    sc.myxvimage = NULL;
    sc.xvcontext = NULL;
    sc.xvwin = 0;
    sc.noxvdemo = FALSE;
#endif

    sc.gamma_max[0] = sc.gamma_max[1] = sc.gamma_max[2] = 1.0;
    sc.gamma_prescale[0] = sc.gamma_prescale[1] = sc.gamma_prescale[2] = 1.0;
    sc.ga_drawable = NULL;

    sc.pixbuf_lcd = sc.pixbuf_lcda = sc.pixbuf_tvpal = sc.pixbuf_tvpalm = NULL;
    sc.pixbuf_tvpaln = sc.pixbuf_tvntsc = sc.pixbuf_tvhivi = sc.pixbuf_tv525i = NULL;
    sc.pixbuf_tv525p = sc.pixbuf_tv750p = sc.pixbuf_tv1080i = sc.pixbuf_vga2 = NULL;

    /* handle gtk-specific command line options */
    gtk_init(&argc, &argv);

    /* handle our command line options */

    for(i = 1; i < argc; i++) {
	char *arg = argv[i];

	if(arg[0] == '-') {
	   if(strcmp(arg, "-display") == 0) {
		if(++i >= argc) {
		   fprintf(stderr, missingparm, "display");
		   exit(1);
		}
		displayname = argv[i];
		continue;
	    } else if(strcmp(arg, "-screen") == 0) {
		if(++i >= argc) {
		   fprintf(stderr, missingparm, "screen");
		   exit(1);
		}
		sc.myscreen = atoi(argv[i]);
		screengiven = TRUE;
		continue;
	    } else if(strcmp(arg, "-force") == 0) {
	        force = 1;
	    } else if((strncmp(arg, "-h", 2) == 0) || (strcmp(arg, "-?") == 0)) {
	        fprintf(stdout, "%s [-display host:dpy] [-screen screen] [-force] [-status]\n\t"
				"[-setgammabrightness|-sg] [ [-setcrt1 mode] | [-setcrt2 mode]\n\t"
				"[-settvstandard standard] [-settvoverscan mode] [settvsapect mode] ]\n\t"
				"[-setxvcrt num] [-settvxpos x] [-settvypos y] [-settvxscale x]\n\t"
				"[-settvyscale y] [-small] "
#ifdef USEXV
				"[-noxvdemo] "
#endif
				"[-v|-version] [-h|-help|-?]\n", argv[0]);
		fprintf(stdout, "display: target display in host:dpy format\n");
		fprintf(stdout, "screen: screen number (used only if Xinerama is detected)\n");
		fprintf(stdout, "force: to delete lock file (and ignore other eventually running sessions)\n");
		fprintf(stdout, "small: enable scrolling in main menu, eventually allowing smaller window\n");
#ifdef USEXV
		fprintf(stdout, "noxvdemo: disable xv demo graphics in Xv settings tab\n");
#endif
		fprintf(stdout, "version: print sisctrl version and exit\n");
		fprintf(stdout, "setgammabrightness: set stored gamma brightness from driver and exit\n");
		fprintf(stdout, "-- Specifying one of the following options will disable the GUI:\n");
		fprintf(stdout, "status: print current display status and exit\n");
		fprintf(stdout, "setcrt1: set CRT1 mode (\"on\"=\"vga\", \"off\", \"lcd\")\n");
		fprintf(stdout, "setcrt2: set CRT2 mode (\"lcd\", \"tv\", \"svideo\", \"cvbs\", \"svideo+cvbs\",\n\t\"hivision\", \"ypbpr1080i\", \"ypbpr720p\", \"ypbpr480p\", \"ypbpr480i\", \"vga\",\n\t\"off\")\n");
		fprintf(stdout, "settvstandard: set TV standard (\"pal\", \"ntsc\", \"palm\", \"paln\", \"ntscj\")\n");
		fprintf(stdout, "settvoverscan: set TV overscan (\"bios\", \"overscan\", \"underscan\", \"soverscan\")\n");
		fprintf(stdout, "settvaspect: set YPbPr TV aspect ratio (\"4:3\", \"4:3lb\", \"16:9\")\n");
		fprintf(stdout, "setxvcrt: set video overlay CRT number (1 or 2)\n");
		fprintf(stdout, "settvxpos: set TV x position (-32 <= x <= 32)\n");
		fprintf(stdout, "settvypos: set TV y position (-32 <= y <= 32)\n");
		fprintf(stdout, "settvxscale: set TV x scale (-16 <= x <= 16)\n");
		fprintf(stdout, "settvyscale: set TV y scale (-4 <= y <= 3)\n");
		exit(0);
	    } else if(strncmp(arg, "-v", 2) == 0) {
	        fprintf(stdout, "sisctrl version " SISVERSION "\n");
		exit(0);
	    } else if(strcmp(arg, "-status") == 0) {
	        printstatus = TRUE;
	    } else if((strcmp(arg, "-setgammabrightness") == 0) || (strncmp(arg, "-sg", 3) == 0)) {
	        setgammafromdriver = TRUE;
	    } else if(strcmp(arg, "-setcrt1") == 0) {
	        if(++i >= argc) {
		   fprintf(stderr, missingparm, "setcrt1");
		   exit(1);
		}
		if((strcmp(argv[i], "on") == 0) ||
		   (strcmp(argv[i], "vga") == 0) ||
		   (strcmp(argv[i], "true") == 0) ||
		   (strcmp(argv[i], "yes") == 0)) {
		   setcrt1index = 1;
		   setcrt1 = TRUE;
		} else if((strcmp(argv[i], "off") == 0) ||
		          (strcmp(argv[i], "false") == 0) ||
		          (strcmp(argv[i], "no") == 0)) {
		   setcrt1index = 0;
		   setcrt1 = TRUE;
		} else if((strcmp(argv[i], "lcd") == 0) ||
		          (strcmp(argv[i], "lcda") == 0) ||
		          (strcmp(argv[i], "lcd-a") == 0)) {
		   setcrt1index = 2;
		   setcrt1 = TRUE;
		} else {
		   fprintf(stderr, illegalparm, "setcrt1");
		   exit(1);
		}
	    } else if(strcmp(arg, "-setcrt2") == 0) {
	        if(++i >= argc) {
		   fprintf(stderr, missingparm, "setcrt2");
		   exit(1);
		}
		if((strcmp(argv[i], "lcd") == 0) ||
		   (strcmp(argv[i], "plasma") == 0) ||
		   (strcmp(argv[i], "dvi-d") == 0)) {
		   sc.givencrt2index = MI_TV_LCD;
		   setcrt2 = TRUE;
		} else if((strcmp(argv[i], "svideo") == 0) ||
		          (strcmp(argv[i], "svhs") == 0)) {
		   sc.givencrt2index = MI_TV_SVHS;
		   setcrt2 = TRUE;
		} else if((strcmp(argv[i], "composite") == 0) ||
		          (strcmp(argv[i], "cvbs") == 0)) {
		   sc.givencrt2index = MI_TV_CVBS;
		   setcrt2 = TRUE;
		} else if((strcmp(argv[i], "svideo+composite") == 0) ||
			  (strcmp(argv[i], "composite+svideo") == 0) ||
			  (strcmp(argv[i], "svideo+cvbs") == 0) ||
			  (strcmp(argv[i], "cvbs+svideo") == 0)) {
		   sc.givencrt2index = MI_TV_CVSV;
		   setcrt2 = TRUE;
		} else if((strcmp(argv[i], "hivision") == 0)) {
		   sc.givencrt2index = MI_TV_HIVI;
		   setcrt2 = TRUE;
		} else if((strcmp(argv[i], "ypbpr480i") == 0)) {
		   sc.givencrt2index = MI_TV_525I;
		   setcrt2 = TRUE;
		} else if((strcmp(argv[i], "ypbpr480p") == 0)) {
		   sc.givencrt2index = MI_TV_525P;
		   setcrt2 = TRUE;
		} else if((strcmp(argv[i], "ypbpr720p") == 0)) {
		   sc.givencrt2index = MI_TV_750P;
		   setcrt2 = TRUE;
		} else if((strcmp(argv[i], "ypbpr1080i") == 0)) {
		   sc.givencrt2index = MI_TV_1080I;
		   setcrt2 = TRUE;
		} else if((strcmp(argv[i], "tv") == 0)) {
		   sc.givencrt2index = MI_TV_TV;
		   setcrt2 = TRUE;
		} else if((strcmp(argv[i], "vga") == 0) ||
		          (strcmp(argv[i], "vga2") == 0)) {
		   sc.givencrt2index = MI_TV_VGA;
		   setcrt2 = TRUE;
		} else if((strcmp(argv[i], "scart") == 0)) {
		   sc.givencrt2index = MI_TV_SCART;
		   setcrt2 = TRUE;
		} else if((strcmp(argv[i], "off") == 0)) {
		   sc.givencrt2index = MI_TV_OFF;
		   setcrt2 = TRUE;
		} else {
		   fprintf(stderr, illegalparm, "setcrt2");
		   exit(1);
		}
  	     } else if((strcmp(arg, "-settvstd") == 0) ||
	               (strcmp(arg, "-settvstandard") == 0)) {
	        if(++i >= argc) {
		   fprintf(stderr, missingparm, "settvstandard");
		   exit(1);
		}
		if(strcmp(argv[i], "pal") == 0) {
		   sc.giventvstdindex = 0;
		   settvstd = TRUE;
		} else if(strcmp(argv[i], "ntsc") == 0) {
		   sc.giventvstdindex = 1;
		   settvstd = TRUE;
		} else if((strcmp(argv[i], "pal-m") == 0) ||
		          (strcmp(argv[i], "palm") == 0)) {
		   sc.giventvstdindex = 2;
		   settvstd = TRUE;
		} else if((strcmp(argv[i], "pal-n") == 0) ||
		          (strcmp(argv[i], "paln") == 0)) {
		   sc.giventvstdindex = 3;
		   settvstd = TRUE;
		} else if((strcmp(argv[i], "ntsc-j") == 0) ||
		          (strcmp(argv[i], "ntscj") == 0)) {
		   sc.giventvstdindex = 4;
		   settvstd = TRUE;
		} else {
		   fprintf(stderr, illegalparm, "settvstandard");
		   exit(1);
		}
            } else if(strcmp(arg, "-settvoverscan") == 0) {
	        if(++i >= argc) {
		   fprintf(stderr, missingparm, "settvoverscan");
		   exit(1);
		}
		if(strcmp(argv[i], "bios") == 0) {
		   sc.giventvosindex = 0;
		   settvos = TRUE;
		} else if((strcmp(argv[i], "underscan") == 0) ||
		          (strcmp(argv[i], "under") == 0)) {
		   sc.giventvosindex = 1;
		   settvos = TRUE;
		} else if((strcmp(argv[i], "overscan") == 0) ||
		          (strcmp(argv[i], "over") == 0)) {
		   sc.giventvosindex = 2;
		   settvos = TRUE;
		} else if((strcmp(argv[i], "soverscan") == 0) ||
		          (strcmp(argv[i], "sover") == 0)) {
		   sc.giventvosindex = 3;
		   settvos = TRUE;
		} else {
		   fprintf(stderr, illegalparm, "settvoverscan");
		   exit(1);
		}
	    } else if(strcmp(arg, "-settvaspect") == 0) {
	        if(++i >= argc) {
		   fprintf(stderr, missingparm, "settvaspect");
		   exit(1);
		}
		if(strcmp(argv[i], "4:3") == 0) {
		   sc.giventvasindex = 0;
		   settvas = TRUE;
		} else if(strcmp(argv[i], "4:3lb") == 0) {
		   sc.giventvasindex = 1;
		   settvas = TRUE;
		} else if(strcmp(argv[i], "16:9") == 0) {
		   sc.giventvasindex = 2;
		   settvas = TRUE;
		} else {
		   fprintf(stderr, illegalparm, "settvaspect");
		   exit(1);
		}
	    } else if(strcmp(arg, "-setxvcrt") == 0) {
	        if(++i >= argc) {
		   fprintf(stderr, missingparm, "setxvcrt");
		   exit(1);
		}
		xvcrtindex = atoi(argv[i]);
		if((xvcrtindex != 1) && (xvcrtindex != 2)) {
		   fprintf(stderr, illegalparm, "setxvcrt");
		   exit(1);
		} else {
		   setxvcrt = TRUE;
		}
	    } else if((strcmp(arg, "-settvxposition") == 0) ||
	              (strcmp(arg, "-settvxpos") == 0)) {
	    	if(++i >= argc) {
		   fprintf(stderr, missingparm, "settvxpos");
		   exit(1);
		}
		tvxposvalue = atoi(argv[i]);
		if(tvxposvalue >= -32 && tvxposvalue <= 32) {
	           settvxpos = TRUE;
		} else {
		   fprintf(stderr, illegalparm, "settvxpos");
		   exit(1);
		}
	    } else if((strcmp(arg, "-settvyposition") == 0) ||
	              (strcmp(arg, "-settvypos") == 0)) {
	        if(++i >= argc) {
		   fprintf(stderr, missingparm, "settvypos");
		   exit(1);
		}
		tvyposvalue = atoi(argv[i]);
		if(tvyposvalue >= -32 && tvyposvalue <= 32) {
	           settvypos = TRUE;
		} else {
		   fprintf(stderr, illegalparm, "settvypos");
		   exit(1);
		}
	    } else if(strcmp(arg, "-settvxscale") == 0) {
	        if(++i >= argc) {
		   fprintf(stderr, missingparm, "settvxscale");
		   exit(1);
		}
		tvxscalevalue = atoi(argv[i]);
		if(tvxscalevalue >= -16 && tvxscalevalue <= 16) {
	           settvxscale = TRUE;
		} else {
		   fprintf(stderr, illegalparm, "settvxscale");
		   exit(1);
		}
	    } else if(strcmp(arg, "-settvyscale") == 0) {
	        if(++i >= argc) {
		   fprintf(stderr, missingparm, "settvyscale");
		   exit(1);
		}
		tvyscalevalue = atoi(argv[i]);
		if(tvyscalevalue >= -4 && tvyscalevalue <= 3) {
	           settvyscale = TRUE;
		} else {
		   fprintf(stderr, illegalparm, "settvyscale");
		   exit(1);
		}
 	    } else if(strcmp(arg, "-small") == 0) {
	        sc.nbscroll = TRUE;
#ifdef USEXV
	    } else if(strcmp(arg, "-noxvdemo") == 0) {
	        sc.noxvdemo = TRUE;
#endif
	    } else {
	        fprintf(stderr, "Unknown parameter: %s, try -help\n", arg);
	    }
        }
    }

    if(!(sc.dpy = XOpenDisplay(displayname))) {
       fprintf(stderr, "%s:  unable to open display \"%s\"\n",
		argv[0], XDisplayName(displayname));
       exit(1);
    }

    strcpy(sc.filenamebuffer, LOCKFILENAME);
    strncat(sc.filenamebuffer, XDisplayName(displayname), 2);

    if(sc.myscreen == -1) sc.myscreen = DefaultScreen(sc.dpy);

    for(i = 0; i < MYMAXMODES; i++) {
       sc.modetextptr[i] = NULL;
       sc.sizetextptr[i] = NULL;
    }

    if(!printstatus) {

       if((sc.lockfile = open(sc.filenamebuffer, O_RDONLY)) == -1) {
          if(errno != ENOENT) {
             fprintf(stderr, "Error accessing lock file %s\n", sc.filenamebuffer);
	     sc.lockfile = -1;
	     cleanupandexit(3);
          }
       } else {
          close(sc.lockfile);
          if(force) {
             remove(sc.filenamebuffer);
          } else {
             sc.lockfile = -1;
	     if(argc <= 1) {
	        if(showerror_pregui("SiSCtrl found a lockfile which indicates that another session "
	                            "is already running. SiSCtrl should only be run once at a time.\n\n"
		  		    "If no other session is running and the lockfile is left from "
		 		    "a previous session which had crashed, click YES to start it anyway, "
				    "or use the -force parameter\n\n"
				    "Start sisctrl anyway?", TRUE, TRUE)) {
	           remove(sc.filenamebuffer);
	        } else {
	           cleanupandexit(3);
	        }
	     } else {
                fprintf(stderr, "Another session is already running. Exiting...\n");
	        fprintf(stderr, "(If no other session is running, try starting with -force\n");
	        fprintf(stderr, "or remove the lockfile %s)\n", sc.filenamebuffer);
                cleanupandexit(3);
	     }
          }
       }

       if((sc.lockfile = open(sc.filenamebuffer, O_RDWR|O_CREAT|O_EXCL|O_TRUNC, S_IRWXU)) == -1) {
	   fprintf(stderr, "Warning: Could not create lock file; another session might be active\n");
	   fprintf(stderr, "         If no other session is running, try starting with -force\n");
	   fprintf(stderr, "         or remove the lockfile %s\n", sc.filenamebuffer);
	   sc.lockfile = -1;
       }

       signal(SIGTERM, signal_handler);
       signal(SIGHUP, signal_handler);
       signal(SIGINT, signal_handler);

    }

    /* Now query the Xv extension */
    if(!XvQueryExtension(sc.dpy, &xv_major, &xv_minor, &xv_req_base,
		      &xv_event_base, &xv_error_base) == Success) {
       fprintf(stderr, "Unable to query xv extension version\n");
       cleanupandexit(3);
    }

    XvQueryAdaptors(sc.dpy, DefaultRootWindow(sc.dpy), &num_adaptors, &adaptor_info);
    {
       Bool found = FALSE;
       if(num_adaptors >= 1) {
          for(i = 0; i < num_adaptors; i++) {
             if(!strcmp(adaptor_info[i].name, "SIS 300/315/330 series Video Overlay")) {
       	        found = TRUE;
	        sc.xv_port_nr = adaptor_info[i].base_id;
             }
	  }
       }
       XvFreeAdaptorInfo(adaptor_info);
       if((!found) || (sc.xv_port_nr == -1)) {
          fprintf(stderr, "SiS video adaptor not found\n");
	  fprintf(stderr, "This program is for SiS 300, 315 and 330 series only\n");
	  fprintf(stderr, "and requires the Xv (XVideo) extension.\n");
          cleanupandexit(3);
       }
    }

    attributes = XvQueryPortAttributes(sc.dpy, sc.xv_port_nr, &num_attr);
    if((!attributes) || num_attr == 0) {
       fprintf(stderr, "No Xv attributes found\n");
       if(attributes) XFree(attributes);
       cleanupandexit(3);
    }

    /* Check existance of requisc.red Xv attributes */
    if((!(sc.atom_con = searchandmakeatom("XV_CONTRAST", num_attr, attributes)))            ||
       (!(sc.atom_bri = searchandmakeatom("XV_BRIGHTNESS", num_attr, attributes)))          ||
       (!(sc.atom_col = searchandmakeatom("XV_COLORKEY", num_attr, attributes)))          ||
       (!(sc.atom_qvf = searchandmakeatom("XV_QUERYVBFLAGS", num_attr, attributes)))        ||
       (!(sc.atom_tvx = searchandmakeatom("XV_TVXPOSITION", num_attr, attributes)))         ||
       (!(sc.atom_tvy = searchandmakeatom("XV_TVYPOSITION", num_attr, attributes)))         ||
       (!(sc.atom_usd = searchandmakeatom("XV_SD_UNLOCKSISDIRECT", num_attr, attributes)))  ||
       (!(sc.atom_qvv = searchandmakeatom("XV_SD_QUERYVBFLAGSVERSION", num_attr, attributes)))   ||
       (!(sc.atom_gsf = searchandmakeatom("XV_SD_GETSDFLAGS", num_attr, attributes)))       ||
       (!(sc.atom_qdd = searchandmakeatom("XV_SD_QUERYDETECTEDDEVICES", num_attr, attributes)))  ||
       (!(sc.atom_ct1 = searchandmakeatom("XV_SD_CRT1STATUS", num_attr, attributes)))       ||
       (!(sc.atom_cmd = searchandmakeatom("XV_SD_CHECKMODEINDEXFORCRT2", num_attr, attributes))) ||
       (!(sc.atom_cmdr= searchandmakeatom("XV_SD_RESULTCHECKMODEINDEXFORCRT2", num_attr, attributes))) ||
       (!(sc.atom_svf = searchandmakeatom("XV_SD_SETVBFLAGS", num_attr, attributes)))       ||
       (!(sc.atom_taf = searchandmakeatom("XV_SD_SISANTIFLICKER", num_attr, attributes)))   ||
       (!(sc.atom_tsa = searchandmakeatom("XV_SD_SISSATURATION", num_attr, attributes)))    ||
       (!(sc.atom_tee = searchandmakeatom("XV_SD_SISEDGEENHANCE", num_attr, attributes)))   ||
       (!(sc.atom_coc = searchandmakeatom("XV_SD_SISCOLCALIBC", num_attr, attributes)))     ||
       (!(sc.atom_cof = searchandmakeatom("XV_SD_SISCOLCALIBF", num_attr, attributes)))     ||
       (!(sc.atom_cfi = searchandmakeatom("XV_SD_SISCFILTER", num_attr, attributes)))       ||
       (!(sc.atom_yfi = searchandmakeatom("XV_SD_SISYFILTER", num_attr, attributes)))       ||
       (!(sc.atom_tco = searchandmakeatom("XV_SD_CHCONTRAST", num_attr, attributes)))       ||
       (!(sc.atom_tte = searchandmakeatom("XV_SD_CHTEXTENHANCE", num_attr, attributes)))    ||
       (!(sc.atom_tcf = searchandmakeatom("XV_SD_CHCHROMAFLICKERFILTER", num_attr, attributes))) ||
       (!(sc.atom_tlf = searchandmakeatom("XV_SD_CHLUMAFLICKERFILTER", num_attr, attributes)))   ||
       (!(sc.atom_ccc = searchandmakeatom("XV_SD_CHCVBSCOLOR", num_attr, attributes)))      ||
       (!(sc.atom_ovr = searchandmakeatom("XV_SD_CHOVERSCAN", num_attr, attributes)))       ||
       (!(sc.atom_sga = searchandmakeatom("XV_SD_ENABLEGAMMA", num_attr, attributes)))      ||
       (!(sc.atom_gdv = searchandmakeatom("XV_SD_GETDRIVERVERSION", num_attr, attributes))) ||
       (!(sc.atom_ghi = searchandmakeatom("XV_SD_GETHARDWAREINFO", num_attr, attributes)))  ||
       (!(sc.atom_gbi = searchandmakeatom("XV_SD_GETBUSID", num_attr, attributes)))         ||
       (!(sc.atom_txs = searchandmakeatom("XV_SD_TVXSCALE", num_attr, attributes)))         ||
       (!(sc.atom_tys = searchandmakeatom("XV_SD_TVYSCALE", num_attr, attributes)))) {
       XFree(attributes);
       cleanupandexit(3);
    }

    if(XvGetPortAttribute(sc.dpy, sc.xv_port_nr, sc.atom_usd, &xv_val) != Success) {
       fprintf(stderr, "Fatal: Couldn't read SiS Direct lock status\n");
       XFree(attributes);
       cleanupandexit(3);
    } else {
       if(xv_val) {
          if(!showerror_pregui("SiSCtrl found the SiS Direct Interface in unlocked state.\n\n"
	  	 	       "This might mean that another application is currently using"
			       "it, possibly leading to interface congestions.\n\n"
			       "Start sisctrl anyway?", TRUE, TRUE)) {
             XFree(attributes);
             cleanupandexit(3);
	  }
       }
    }

    if(XvGetPortAttribute(sc.dpy, sc.xv_port_nr, sc.atom_gsf, &xv_val) != Success) {
       fprintf(stderr, "Fatal: Couldn't read SiS Direct flags\n");
       XFree(attributes);
       cleanupandexit(3);
    }
    sc.sdflags = (unsigned long)xv_val;

    sc.HaveGss = FALSE;
    if(!(sc.sdflags & SiS_SD_ISDHXINERAMA)) {
       if((sc.atom_gss = searchandmakeatom("XV_SD_GETSCREENSIZE", num_attr, attributes))) {
          sc.HaveGss = TRUE;
       }
    }

    sc.HaveBri = FALSE;
    if((sc.sdflags & SiS_SD_ISDHXINERAMA) && (sc.myscreen != DefaultScreen(sc.dpy))) {
       if(((sc.atom_brr = searchandmakeatom("XV_SD_STOREDGAMMABRIR2", num_attr, attributes))) &&
          ((sc.atom_brg = searchandmakeatom("XV_SD_STOREDGAMMABRIG2", num_attr, attributes))) &&
          ((sc.atom_brb = searchandmakeatom("XV_SD_STOREDGAMMABRIB2", num_attr, attributes))) &&
          ((sc.atom_pbr = searchandmakeatom("XV_SD_STOREDGAMMAPBRIR2", num_attr, attributes))) &&
          ((sc.atom_pbg = searchandmakeatom("XV_SD_STOREDGAMMAPBRIG2", num_attr, attributes))) &&
          ((sc.atom_pbb = searchandmakeatom("XV_SD_STOREDGAMMAPBRIB2", num_attr, attributes)))) {
          sc.HaveBri = TRUE;
       }
    } else {
       if(((sc.atom_brr = searchandmakeatom("XV_SD_STOREDGAMMABRIR", num_attr, attributes))) &&
          ((sc.atom_brg = searchandmakeatom("XV_SD_STOREDGAMMABRIG", num_attr, attributes))) &&
          ((sc.atom_brb = searchandmakeatom("XV_SD_STOREDGAMMABRIB", num_attr, attributes))) &&
          ((sc.atom_pbr = searchandmakeatom("XV_SD_STOREDGAMMAPBRIR", num_attr, attributes))) &&
          ((sc.atom_pbg = searchandmakeatom("XV_SD_STOREDGAMMAPBRIG", num_attr, attributes))) &&
          ((sc.atom_pbb = searchandmakeatom("XV_SD_STOREDGAMMAPBRIB", num_attr, attributes)))) {
          sc.HaveBri = TRUE;
       }
    }

    /* For security reasons, we only allow us to run if the X server allows it */
    if(!(sc.sdflags & SiS_SD_ENABLED)) {
       if(argc <= 1) {
          showerror_pregui("SiSCtrl interface disabled.\n\n"
	  		   "To enable the SiSCtrl interface, place\n\n"
			   "\tOption \"EnableSiSCtrl\" \"true\"\n\n"
			   " in the Device section of your XF86Config(-4) file.\n", FALSE, FALSE);
       } else {
          fprintf(stderr, "Fatal: SiSCtrl interface is disabled\n");
       }
       XFree(attributes);
       cleanupandexit(3);
    }

    /* Check for hardware specific Xv attributes */
    if(sc.sdflags & (SiS_SD_IS315SERIES | SiS_SD_IS330SERIES)) {
       if((!(sc.atom_sat = searchandmakeatom("XV_SATURATION", num_attr, attributes))) ||
          (!(sc.atom_hue = searchandmakeatom("XV_HUE", num_attr, attributes)))) {
	  XFree(attributes);
	  cleanupandexit(3);
       }
    }

    if(!(sc.sdflags & SiS_SD_SUPPORT2OVL)) {
       if(!(sc.atom_swc = searchandmakeatom("XV_SWITCHCRT", num_attr, attributes))) {
          XFree(attributes);
          cleanupandexit(3);
       }
    }

    if(sc.sdflags & SiS_SD_SUPPORTXVGAMMA1) {
       if( (!(sc.atom_vgr = searchandmakeatom("XV_GAMMA_RED", num_attr, attributes)))   ||
           (!(sc.atom_vgg = searchandmakeatom("XV_GAMMA_GREEN", num_attr, attributes))) ||
	   (!(sc.atom_vgb = searchandmakeatom("XV_GAMMA_BLUE", num_attr, attributes))) ) {
          XFree(attributes);
          cleanupandexit(3);
       }
    }

    if(sc.sdflags & SiS_SD_SUPPORTSCALE) {
       if(!(sc.atom_pmd = searchandmakeatom("XV_SD_PANELMODE", num_attr, attributes))) {
          sc.sdflags &= ~SiS_SD_SUPPORTSCALE;
       }
    }

    XFree(attributes);

    if(XvGetPortAttribute(sc.dpy, sc.xv_port_nr, sc.atom_qvv, &xv_val) != Success) {
       fprintf(stderr, "Fatal: Couldn't read vbflags version\n");
       cleanupandexit(3);
    }

    if(xv_val != 1) {
       fprintf(stderr, "Fatal: Invalid vbflags version (%d)\n", xv_val);
       cleanupandexit(3);
    }

    if(XvGetPortAttribute(sc.dpy, sc.xv_port_nr, sc.atom_qvf, &xv_val) != Success) {
       fprintf(stderr, "Fatal: Couldn't read vbflags\n");
       cleanupandexit(3);
    }
    sc.currentvbflags = sc.backupvbflags = (unsigned long)xv_val;

    if(XvGetPortAttribute(sc.dpy, sc.xv_port_nr, sc.atom_qdd, &xv_val) != Success) {
       fprintf(stderr, "Fatal: Couldn't read detected devices\n");
       cleanupandexit(3);
    }
    sc.detecteddevices = (unsigned long)xv_val;

    if(XvGetPortAttribute(sc.dpy, sc.xv_port_nr, sc.atom_ct1, &xv_val) != Success) {
       fprintf(stderr, "Fatal: Couldn't read CRT1 status\n");
       cleanupandexit(3);
    }
    sc.CRT1isoff = sc.CRT1isoffbackup = xv_val ? 0 : 1;

    if(XvGetPortAttribute(sc.dpy, sc.xv_port_nr, sc.atom_sga, &xv_val) != Success) {
       fprintf(stderr, "Fatal: Couldn't read gamma enable/disable\n");
       cleanupandexit(3);
    }
    xv_val &= 0x07;
    if(!(sc.sdflags & SiS_SD_SUPPORTXVGAMMA1)) xv_val &= 0x03;
    sc.gammaflags = sc.backupgammaflags = xv_val;

    if(XvGetPortAttribute(sc.dpy, sc.xv_port_nr, sc.atom_ovr, &xv_val) != Success) {
       fprintf(stderr, "Fatal: Couldn't read overscan setting\n");
       cleanupandexit(3);
    }
    sc.choverscanstatus = sc.backupoverscan = xv_val & 0x03;

    sc.lcdscaler = sc.backuplcdscaler = 0;
    if(sc.sdflags & SiS_SD_SUPPORTSCALE) {
       if(XvGetPortAttribute(sc.dpy, sc.xv_port_nr, sc.atom_pmd, &xv_val) != Success) {
          fprintf(stderr, "Fatal: Couldn't read panel mode setting\n");
          cleanupandexit(3);
       }
       sc.lcdscaler = sc.backuplcdscaler = xv_val & 0x0f;
    }

    if(XvGetPortAttribute(sc.dpy, sc.xv_port_nr, sc.atom_ghi, &xv_val) != Success) {
       fprintf(stderr, "Fatal: Couldn't read SiS hardware info\n");
       cleanupandexit(3);
    }
    sc.hwinfo = (unsigned long)xv_val;

    if(XvGetPortAttribute(sc.dpy, sc.xv_port_nr, sc.atom_gbi, &xv_val) != Success) {
       fprintf(stderr, "Fatal: Couldn't read busID info\n");
       cleanupandexit(3);
    }
    sc.busid = (unsigned long)xv_val;

    if(XvGetPortAttribute(sc.dpy, sc.xv_port_nr, sc.atom_gdv, &xv_val) != Success) {
       fprintf(stderr, "Fatal: Couldn't read SiS driver version\n");
       cleanupandexit(3);
    }
    sc.driverversion = (unsigned long)xv_val;

    if((sc.driverversion & 0x00ffffff) < 0x030c04) {  /* 2003/12/04 */
       fprintf(stderr, "Fatal: This version of SiSCtrl is too new for this version of the SiS XFree86 driver\n");
       cleanupandexit(3);
    }

    if(sc.sdflags & SiS_SD_ISDHXINERAMA) {
       if(!screengiven) {
          fprintf(stdout, "Xinerama detected. You might want to use the -screen parameter\n");
	  fprintf(stdout, "to choose for which screen you want to the settings to apply.\n");
       }
    } else sc.myscreen = DefaultScreen(sc.dpy);

    /* Now query the XVidModExtension */
    if(!XF86VidModeQueryVersion(sc.dpy, &MajorVersion, &MinorVersion)) {
       fprintf(stderr, "Fatal: Unable to query vidmod extension version\n");
       cleanupandexit(2);
    }

    if(!XF86VidModeQueryExtension(sc.dpy, &EventBase, &ErrorBase)) {
       fprintf(stderr, "Fatal: Unable to query vidmod extension information\n");
       cleanupandexit(2);
    }

    /* Fail if the extension version in the server is too old */
    if(MajorVersion < MINMAJOR || (MajorVersion == MINMAJOR && MinorVersion < MINMINOR)) {
       fprintf(stderr, "Fatal: Found VidModeExtension version (%d.%d)\n", MajorVersion, MinorVersion);
       fprintf(stderr, "Minimum required version for this program is %d.%d\n", MINMAJOR, MINMINOR);
       cleanupandexit(2);
    }

    if(!XF86VidModeGetAllModeLines(sc.dpy, sc.myscreen, &sc.vmmodecount, &sc.vmmodelines)) {
       fprintf(stderr, "Fatal: Failed to query modelines from vidmode extension\n");
       cleanupandexit(2);
    }

    if(!XF86VidModeGetGamma(sc.dpy, sc.myscreen, &sc.currentgamma)) {
       fprintf(stderr, "Fatal: Failed to query gamma correction\n");
       cleanupandexit (2);
    }
    memcpy(&sc.newgamma, &sc.currentgamma, sizeof(XF86VidModeGamma));
    sc.backup_gred = sc.currentgamma.red;
    sc.backup_ggreen = sc.currentgamma.green;
    sc.backup_gblue = sc.currentgamma.blue;

    if(sc.HaveBri) {
       sc.gamma_max[0] = sc.gamma_max_b[0] = (float)getxvval(sc.atom_brr) / 1000;
       sc.gamma_max[1] = sc.gamma_max_b[1] = (float)getxvval(sc.atom_brg) / 1000;
       sc.gamma_max[2] = sc.gamma_max_b[2] = (float)getxvval(sc.atom_brb) / 1000;
       sc.gamma_prescale[0] = sc.gamma_pre_b[0] = (float)getxvval(sc.atom_pbr) / 1000;
       sc.gamma_prescale[1] = sc.gamma_pre_b[1] = (float)getxvval(sc.atom_pbg) / 1000;
       sc.gamma_prescale[2] = sc.gamma_pre_b[2] = (float)getxvval(sc.atom_pbb) / 1000;
    }

    if(sc.sdflags & SiS_SD_SUPPORTXVGAMMA1) {
       sc.xvgammared = (float)(getxvval(sc.atom_vgr) / 1000);
       sc.xvgammagreen = (float)(getxvval(sc.atom_vgg) / 1000);
       sc.xvgammablue = (float)(getxvval(sc.atom_vgb) / 1000);
    }

   sc.haverandr = FALSE;
#ifdef HAVE_RANDR
   if(XRRQueryExtension(sc.dpy, &rr_event_base, &rr_error_base)) sc.haverandr = TRUE;
   if(sc.haverandr) {
      XRRQueryVersion(sc.dpy, &MajorVersion, &MinorVersion);
      if((MajorVersion < RRMINMAJOR) ||
        ((MajorVersion == RRMINMAJOR) && (MinorVersion < RRMINMINOR))) {
         fprintf(stderr, "Fatal: Found RANDR Extension version (%d.%d)\n", MajorVersion, MinorVersion);
         fprintf(stderr, "Minimum required version for this program is %d.%d\n", RRMINMAJOR, RRMINMINOR);
         sc.haverandr = FALSE;
      } else {
         sc.root = RootWindow(sc.dpy, sc.myscreen);
         sc.sconf = XRRGetScreenInfo(sc.dpy, sc.root);
         if(!sc.sconf) {
            fprintf(stderr, "Failed to get screen info from RandR\n");
            sc.haverandr = FALSE;
         } else {
            sc.rrcurrent_size = sc.rrsize_backup =
	         XRRConfigCurrentConfiguration(sc.sconf, &sc.rrcurrent_rotation);
	    sc.rrrotation_backup = sc.rrcurrent_rotation;
         }
      }
   }
#endif

    if(printstatus) {
       writeoutstatus();
       cleanupandexit(0);
       exit(0);
    }

    if(sc.HaveBri) {
       memcpy(&sc.newgamma, &sc.currentgamma, sizeof(XF86VidModeGamma));
       setgamma();
    } else {
       fprintf(stderr, "This version of the SiS XFree86 driver does not support storing gamma\n");
    }

    if(setcrt1) {
       if((!(sc.sdflags & SiS_SD_ISDUALHEAD)) && (sc.currentvbflags & VB_VIDEOBRIDGE)) {
          if(!(handle_apply_CRT1(setcrt1index))) {
	     fprintf(stderr, "setcrt1 failed, check XFree log for error messages\n");
	  }
       } else {
          fprintf(stderr, "setcrt1 not allowed in dual head mode or if no video bridge preset\n");
       }
       if(setcrt2 || settvstd || settvos) {
          fprintf(stderr, "-setcrt1 and -set[crt2|tvstandard|tvoverscan|tvaspect] cannot be handled at the same time\n");
       }
    } else if(setcrt2 || settvstd || settvos) {
       if((!(sc.sdflags & SiS_SD_ISDUALHEAD)) && (sc.currentvbflags & VB_VIDEOBRIDGE)) {
          if(setcrt2) {
             /* Cannot set LCD if no LCD detected */
             if(sc.givencrt2index == MI_TV_LCD) {
	        if(!(sc.detecteddevices & CRT2_LCD)) {
	           fprintf(stderr, "Cannot set CRT2 type to LCD; no LCD detected\n");
	           setcrt2 = FALSE;
		   doexit = TRUE;
	        }
	     }
	     /* VGA only on 301/301B/301C/302B */
	     if(sc.givencrt2index == MI_TV_VGA) {
	        if(!(vbsupportsvga2())) {
		   fprintf(stderr, "Hardware does not support CRT2 VGA output\n");
		   setcrt2 = FALSE;
		   doexit = TRUE;
	        }
	     }
	     if(sc.givencrt2index == MI_TV_HIVI) {
	        if(!(sc.sdflags & SiS_SD_SUPPORTHIVISION)) {
		   fprintf(stderr, "Hardware does not support HiVision output\n");
		   setcrt2 = FALSE;
		   doexit = TRUE;
		} else settvstd = FALSE;
	     }
	     if((sc.givencrt2index == MI_TV_525I) ||
	        (sc.givencrt2index == MI_TV_525P) ||
		(sc.givencrt2index == MI_TV_750P) ||
		(sc.givencrt2index == MI_TV_1080I)) {
		if(!(sc.sdflags & SiS_SD_SUPPORTYPBPR)) {
		   fprintf(stderr, "Hardware does not support YPbPr output\n");
		   setcrt2 = FALSE;
		   doexit = TRUE;
		} else settvstd = FALSE;
	     }

	     if(vbsupportstv()) {
	        if(sc.currentvbflags & VB_SISBRIDGE) {
		   if(sc.givencrt2index == MI_TV_TV) {
	              if(sc.detecteddevices & TV_AVIDEO)
		         sc.givencrt2index = MI_TV_CVBS;
		      else
	                 sc.givencrt2index = MI_TV_SVHS;
		      fprintf(stderr, "TV plug type not specified, default is detected or svideo\n");
	           }
	           if(sc.givencrt2index == MI_TV_SCART) {
	              if(!(vbsupportsscart())) {
		         fprintf(stderr, "This video bridge does not support SCART output\n");
		         setcrt2 = FALSE;
		         doexit = TRUE;
		      } else {
		         sc.giventvstdindex = 0;
		         settvstd = TRUE;
		      }
	           }
	        } else if(sc.currentvbflags & VB_CHRONTEL) {
	           if((sc.givencrt2index == MI_TV_SVHS) ||
	              (sc.givencrt2index == MI_TV_CVBS) ||
		      (sc.givencrt2index == MI_TV_SCART)) {
		      sc.givencrt2index = MI_TV_TV;
	           }
		}
	     } else if((sc.givencrt2index == MI_TV_SVHS) ||
	               (sc.givencrt2index == MI_TV_CVBS) ||
		       (sc.givencrt2index == MI_TV_SCART) ||
		       (sc.givencrt2index == MI_TV_TV)) {
		fprintf(stderr, "This hardware does not support TV output\n");
		setcrt2 = FALSE;
		doexit = TRUE;
	     }
          }
	  if(settvstd) {
	     if(vbsupportstv()) {
	        if(!(sc.sdflags & SiS_SD_SUPPORTPALMN)) {
	           if((sc.giventvstdindex == 2) ||
		      (sc.giventvstdindex == 3)) {
		      fprintf(stderr, "Hardware does not support PALM/PALN, using PAL\n");
		      sc.giventvstdindex = 0;
		   }
	        }
	        if(!(sc.sdflags & SiS_SD_SUPPORTNTSCJ)) {
	           if(sc.giventvstdindex == 4) {
		      fprintf(stderr, "Hardware does not support NTSC-J, using NTSC\n");
		      sc.giventvstdindex = 1;
		   }
	        }
	     } else {
	        fprintf(stderr, "This hardware does not support TV output\n");
	        settvstd = FALSE;
		doexit = TRUE;
	     }
	  }
	  if(!setcrt2) {
	     sc.givencrt2index = getcrt2typeindex(sc.currentvbflags);
          }
          if(!settvstd) {
	     sc.giventvstdindex = gettvtypeindex(sc.currentvbflags);
          }
	  if(!settvas) {
	     sc.giventvasindex = getcrt2tvasmenuindex();
	  }
	  if(settvos) {
	     if(!(sc.currentvbflags & VB_CHRONTEL)) {
	        fprintf(stderr, "Overscan only supported on Chrontel TV encoders\n");
		settvos = FALSE;
		doexit = TRUE;
	     } else {
	        if(sc.giventvosindex == 3) {
		   if(!(sc.sdflags & SiS_SD_SUPPORTSOVER)) {
		      fprintf(stderr, "soverscan not supported on this hardware\n");
		      sc.giventvosindex = 2;
		   } else {
		      if(sc.giventvstdindex != 0) {
		         fprintf(stderr, "soverscan only supported for PAL\n");
			 sc.giventvosindex = 2;
		      }
		   }
		}
	     }
	  }
          if(!settvos) {
	     sc.giventvosindex = sc.choverscanstatus;
          }
	  if(setcrt2 || settvstd || settvos || settvas) {
             if(!(handle_apply_CRT2())) {
	        fprintf(stderr, "set[crt2|tvstandard|tvoverscan] failed, check XFree log for error messages\n");
             }
	  }
       } else {
          fprintf(stderr, "set[crt2|tvstandard|tvoverscan|tvaspect] not supported in dual head mode "
	  			"or if no video bridge preset\n");
       }
    }

    if(setxvcrt) {
       if(sc.sdflags & SiS_SD_SUPPORT2OVL) {
          fprintf(stderr, "setxvcrt not supported on this hardware\n");
	  doexit = TRUE;
       } else {
          setxvval(sc.atom_swc, xvcrtindex - 1);
       }
    }

    if((settvxpos || settvypos) && (!(sc.sdflags & SiS_SD_SUPPORTTVPOS))) {
       fprintf(stderr, "This hardware does not support setting the TV position\n");
       settvxpos = settvypos = FALSE;
       doexit = TRUE;
    }

    if((settvxscale || settvyscale) && (!(sc.currentvbflags & VB_SISTVBRIDGE))) {
       fprintf(stderr, "This hardware does not support setting the TV scale\n");
       settvxscale = settvyscale = FALSE;
       doexit = TRUE;
    }

    if(settvxpos) {
       setxvval(sc.atom_tvx, tvxposvalue);
    }

    if(settvypos) {
       setxvval(sc.atom_tvy, tvyposvalue);
    }

    if(settvxscale) {
       setxvval(sc.atom_txs, tvxscalevalue);
    }

    if(settvyscale) {
       setxvval(sc.atom_tys, tvyscalevalue);
    }

    if(setgammafromdriver || setcrt1 || setcrt2 || settvstd ||
       settvos || setxvcrt || settvxpos || settvypos ||
       settvxscale || settvyscale || doexit) {
       cleanupandexit(0);
       exit(0);
    }

    /* Build the mode list */

    for(i = 0; (i < sc.vmmodecount) && (i < MYMAXMODES); i++) {
       if(sc.sdflags & SiS_SD_ISMERGEDFB) {
          sprintf(buffer, "%dx%d",
		sc.vmmodelines[i]->hdisplay,
		sc.vmmodelines[i]->vdisplay);
	  sc.moderate[i] = 0;
       } else {
          int refresh = calcrefresh(sc.vmmodelines[i]);
	  sc.moderate[i] = refresh;
	  if(sc.sdflags & SiS_SD_ISDUALHEAD) {
	     if(sc.sdflags & SiS_SD_ISDHXINERAMA) {
		sprintf(buffer, "Screen %d: %dx%d (%d Hz%s%s)",
		   sc.myscreen,
		   sc.vmmodelines[i]->hdisplay,
		   sc.vmmodelines[i]->vdisplay,
		   refresh,
		   (sc.vmmodelines[i]->flags & V_INTERLACE) ? " I" : "",
		   (sc.vmmodelines[i]->flags & V_DBLSCAN) ? " D" : "");
	     } else {
	        sprintf(buffer, "CRT%d%s: %dx%d (%d Hz%s%s)",
		   (sc.sdflags & SiS_SD_ISDHSECONDHEAD) ? 1 : 2,
		   ((sc.sdflags & SiS_SD_SUPPORTLCDA) && (sc.sdflags & SiS_SD_ISDHSECONDHEAD)) ?
		   		"/VGA" : "",
		   sc.vmmodelines[i]->hdisplay,
		   sc.vmmodelines[i]->vdisplay,
		   refresh,
		   (sc.vmmodelines[i]->flags & V_INTERLACE) ? " I" : "",
		   (sc.vmmodelines[i]->flags & V_DBLSCAN) ? " D" : "");
	     }
	  } else {
             sprintf(buffer, "%dx%d (CRT1%s: %d Hz%s%s)",
		sc.vmmodelines[i]->hdisplay,
		sc.vmmodelines[i]->vdisplay,
		(sc.sdflags & SiS_SD_SUPPORTLCDA) ? "/VGA" : "",
		refresh,
		(sc.vmmodelines[i]->flags & V_INTERLACE) ? " I" : "",
		(sc.vmmodelines[i]->flags & V_DBLSCAN) ? " D" : "");
          }
       }
       if(!(sc.modetextptr[i] = malloc(strlen(buffer)+1))) cleanupandexit(3);
       strcpy(sc.modetextptr[i], buffer);
    }

    rebuildmodefieldlist();

    /* Now init gtk and build the gui */

    sc.guimode = TRUE;

    buildgui();

    gtk_main ();

    return 0;
}

